/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.layout.Layout;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.FontUIResource;
import javax.swing.plaf.metal.*; // Java Look & Feel


/**
 * The class Application administrates the outer window where all windows of a Hoverball application
 * are displayed.
 * <p>
 * All Hoverball windows are represented as <i>internal frames</i> (class JInternalFrame).
 * Only the application window is a <i>top-level frame</i>
 * (class JFrame) that shows the Hoverball windows on its <a href="#desktop">desktop</a>
 * (class JDesktopPane).
 * Please note:
 *
 * <menu>
 * <li>The standard windows (of the
 *     <a href="Simulator.html">Simulator</a>
 *     and the
 *     <a href="Controller.html">Controller</a>
 *     together with Screen, as well as the
 *     <a href="Hovlet.html">Hovlet</a> browser)
 *     are automatically put on the desktop.
 *     Only self-defined windows that are to be shown in the application window, must be
 *     added to the desktop by
 *     <p>
 *     <tt>Application.desktop.add(</tt>...<tt>);</tt>
 *     <p></li>
 * <li>The position and size of the application window can be determined by the methods
 *     <p>
 *     <tt>Application.setBounds(</tt>...<tt>);</tt>
 *     </p>
 *     If you display only one single window in a Hoverball application
 *     (e.g. by launching only the Simulator),
 *     the application window can be adjusted directly to the dimensions of the second window by
 *     <p>
 *     <tt>Application.setBounds();</tt>
 *     </p>
 *     That is to say, the application window inherits the attributes of the inner window.
 *     <p></li>
 * <li>The application window opens automatically as soon as a standard window gets visible. If required,
 *     it can be opened "manually" by
 *     <p>
 *     <tt>Application.show();</tt>
 *     </p>
 *     <p></li>
 * <li>The automatic showing mechanism can be suppressed with
 *     <p>
 *     <tt>Application.hide();</tt>
 *     </p>
 *     at the beginning of the application.
 *     <p></li>
 * </menu>
 */
public abstract class Application
{
   private static final String TITLE = "Hoverball"; // "Hoverball "+About.VERSION;

   // ---- AWEFUL BAD WORKAROUND FOR SYNC CONNECTING TO HOVERBALL SERVER ------

   static String servertext_with_hash = null;

   static java.util.Vector<Browser.ConnectionActionListener> listeners = new java.util.Vector<Browser.ConnectionActionListener>();

	/**
	 * Connects the Hoverball application to the specified address.
	 *
	 * @param address
	 *            network address
	 * @return <b><code>true</code></b>, if connection succeeded
	 */
   public static void connect (String address) {
	  SwingUtilities.invokeLater(new Runnable() { // ouch!! :-(
	     public void run () { 
	        new Thread(new Runnable() {
		       public void run () { try { Thread.sleep(200); } catch(InterruptedException e) {}
      servertext_with_hash = null;
      if (address == null || address.equals("")) return;
      for (Browser.ConnectionActionListener listener : listeners) listener.connect(address);
		       }
	        }).start();
		 }
	  });
   }

	/**
	 * Disconnects the Hoverball application.
	 */
   public static void disconnect () {
      SwingUtilities.invokeLater(new Runnable() {
		 public void run () {                     
      for (Browser.ConnectionActionListener listener : listeners) listener.disconnect();
		 }
      });
   }


   // ---- Hoverball Bounds ---------------------------------------------------

   private static final int MINIMUM_WIDTH  = 640;
   private static final int MINIMUM_HEIGHT = 480;

   private static String title = "";
   private static Rectangle bounds = new Rectangle();
   private static Dimension screensize = Toolkit.getDefaultToolkit().getScreenSize();

   /**
    * Returns the application window bounds.
    *
    * @return the bounds
    */
   public static Rectangle getBounds ()
   {
      return (bounds == null)? null : new Rectangle(bounds);
   }

   /**
    * Adjusts the application window to a single displayed window.
    * <p>
    * When the application window is opened, it assumes the attributes of the first visible window
    * in the list of the dektop.
    */
   public static void setBounds ()
   {
      Application.bounds = null;
   }

   /**
    * Defines the application window bounds.
    *
    * @param ratio ratio frame bounds / full screen (between 0 and 1)
    */
   public static void setBounds (double ratio)  // 0..1
   {
      Dimension screen = screensize;
      setBounds(new Dimension((int)(screen.width*ratio),(int)(screen.height*ratio)));
   }

   /**
    * Defines the application window bounds.
    *
    * @param size frame size
    */
   public static void setBounds (Dimension size)
   {
      setBounds(Layout.center(new Rectangle(screensize),size));
   }

   /**
    * Defines the application window bounds.
    *
    * @param bounds frame bounds
    */
   public static void setBounds (Rectangle bounds)
   {
      Rectangle r = bounds.union(Layout.center(bounds,new Dimension(MINIMUM_WIDTH,MINIMUM_HEIGHT)));
      Application.bounds = new Rectangle(Layout.fit(new Rectangle(screensize),r));
   }

   /**
    * Returns the application window title.
    *
    * @return the title
    */
   public static String getTitle ()
   {
      return ""+title;
   }

   /**
    * Defines the application window title.
    *
    * @param title title
    */
   public static void setTitle (String title)
   {
      if (title == null || title.length() == 0) {
         Application.title = "";
         frame.setTitle(TITLE);
      } else {
         Application.title = title;
         frame.setTitle(title);
//         frame.setTitle(title+" - "+TITLE);
      }
   }


   // -------------------------------------------------------------------------

   /**
    * Desktop of the application window
    */
   public static final JDesktopPane desktop = new JDesktopPane();

   /**
    * Frame of the application window
    */
   public static final JFrame frame = new JFrame(TITLE);
   
   private static int FRAME_SUPPLEMENT_L = 4;
   private static int FRAME_SUPPLEMENT_R = 4;
   private static int FRAME_SUPPLEMENT_T = 32;
   private static int FRAME_SUPPLEMENT_B = 4;

   private static boolean opened = false;

   static {
	   /*
	   for ( java.util.Enumeration e = UIManager.getDefaults().keys(); e.hasMoreElements(); )
	   {
	       Object key   = e.nextElement();
	       Object value = UIManager.get( key );

	       if ( value instanceof Font )
	       {
	           Font f = (Font) value;

	           UIManager.put( key, new FontUIResource( f.getName(), f.getStyle(), 14 ) );
	       }
	   }
	   */

	  //if (Applet.applet == null)
      {
      setBounds(Layout.getApplicationBounds());

      Layout.loadResources(frame);

      try { // Fix Ocean Theme
         Class ocean = Class.forName("javax.swing.plaf.metal.OceanTheme");
         if (ocean != null)
            MetalLookAndFeel.setCurrentTheme((MetalTheme)ocean.newInstance());
      } catch(Exception e) {}

      //javax.swing.UIManager.put("Label.font", new javax.swing.plaf.FontUIResource(new Font("Dialog", Font.PLAIN, 20)));
      //javax.swing.UIManager.put("Button.font", new javax.swing.plaf.FontUIResource(new Font("Dialog", Font.BOLD, 20)));
      //javax.swing.UIManager.put("TextField.font", new javax.swing.plaf.FontUIResource(new Font("Dialog", Font.PLAIN, 20)));
      /*
      for (Entry<Object, Object> entry : javax.swing.UIManager.getDefaults().entrySet()) {
    	    Object key = entry.getKey();
    	    Object value = javax.swing.UIManager.get(key);
    	    if (value != null && value instanceof javax.swing.plaf.FontUIResource) {
    	        javax.swing.plaf.FontUIResource fr=(javax.swing.plaf.FontUIResource)value;
    	        javax.swing.plaf.FontUIResource f = new javax.swing.plaf.FontUIResource("Noto Sans",fr.getStyle(),fr.getSize()+4); // Font.BOLD,16); //(fr.getFamily(), fr.getStyle(), 20);
    	        javax.swing.UIManager.put(key, f);
    	    }
    	}
      */
      desktop.setBackground(Layout.colorDesktop);

      if (Layout.icon_png != null) frame.setIconImage(new ImageIcon(Layout.icon_png).getImage());
      frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
      frame.getContentPane().add(desktop);

      frame.addWindowListener(new WindowAdapter() {
         public void windowClosed (WindowEvent e) { Application.exit(); }
      });
      }
   }

   /**
    * Closes the application window and quits the program.
    */
   public static void exit ()                                   // exit -> dispose
   {
      synchronized(desktop) { if(exited) return; exited = true; }

      JInternalFrame[] frames = desktop.getAllFrames();
      for (int i=0; i<frames.length; ++i) frames[i].dispose();  // dispose -> close
      frame.dispose();
      
      if (Simulator.headless_simulator != null) Simulator.headless_simulator.close();
   }
   private static boolean exited = false;

   /**
    * Shows the application window (explicit call not necessary).
    */
   public static void show ()
   {
      if (opened) return;
      opened = true;

      if (bounds != null) {
         frame.setBounds(bounds);
         frame.setVisible(true); // .show(); [deprecated, since 1.5]
      } else {
         JInternalFrame[] frames = desktop.getAllFrames();
         JInternalFrame f = null;
         for (int i=0; i<frames.length; ++i) if (frames[i].isVisible()) { f = frames[i]; break; }
         if (f == null) return;

         Rectangle b = f.getBounds();  // expand bounds (toplevel frame border)
         b.x -= FRAME_SUPPLEMENT_L;
         b.y -= FRAME_SUPPLEMENT_T;
         b.width  += FRAME_SUPPLEMENT_L + FRAME_SUPPLEMENT_R;
         b.height += FRAME_SUPPLEMENT_T + FRAME_SUPPLEMENT_B;

         frame.setBounds(Layout.fit(new Rectangle(screensize),b));
         frame.setResizable(f.isResizable());
         frame.setVisible(true); // .show(); [deprecated, since 1.5]

         f.setFrameIcon(null);
         f.setBounds(desktop.getBounds());
         f.setMaximizable(true);
         try { f.setMaximum(true); } catch(java.beans.PropertyVetoException e) {}
         f.addInternalFrameListener(new InternalFrameAdapter() {
            public void internalFrameClosed(InternalFrameEvent e) { Application.exit(); }
         });
         desktop.setSelectedFrame(f);
      }
   }

   /**
    * Hides the application window. If this method is called at the beginning, the Application window will not be shown automatically
    * and stays hidden.
    */
   public static void hide ()
   {
	  Layout.auto_show = false;;

      if (!opened) return;
      opened = false;

	  frame.setVisible(false);
   }
}
