/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.layout.*;
import hoverball.net.Address;

import java.awt.*;
import javax.swing.*;

// ---- The Hovlet ---------------------------------------------- [Hovlet] ----

/**
		 * The class Hovlet (= "small Hoverball program") represents the base
		 * class for all programming elements that can connect to the Simulator.
		 * <p>
		 * Hovlets can be interlaced into each other. Each hovlet has a
		 * super-hovlet <a href="#parent">parent</a> and a list of sub-hovlets
		 * <a href="#hovlets">hovlets</a>, that can be added by the method <a
		 * href="#add(hoverball.Hovlet)">add(...)</a>. The connectability
		 * upward and downward is to be defined in its constructor. A hovlet
		 * without super-hovlet is called <i>root-hovlet</i>.
		 * <p>
		 * The resulting hovlet trees are displayed in so-called <i>browsers</i>
		 * that show certain hovlets on request. The method <a
		 * href="#show()">show()</a> opens a new browser for the full hovlet
		 * tree (unless this has already happened) which displays the initially
		 * calling hovlet. (The looks of the browser can be defined by the
		 * root-hovlet using the method <a
		 * href="#browse(javax.swing.JInternalFrame)">browse(...)</a>.)
		 * Otherwise the hovlet is simply shown in the corresponding browser.
		 * <p>
		 * <div align="center"> <img src="../resources/hovlet.png" width=400
		 * height=342></img> </div>
		 * <p>
		 * A Hovlet has the following properties: <menu>
		 * <li>It holds a title. See <a href="#title">title</a>.
		 * <p>
		 * </li>
		 * <li>It can define private menus which are shown in the browser
		 * beside the default menu <i>Hovlet</i>. See <a
		 * href="#menubar">menubar</a>.
		 * <p>
		 * </li>
		 * <li>It has an icon panel displayed beside the hovlet selection bar.
		 * See <a href="#icon">icon</a>.
		 * <p>
		 * </li>
		 * <li>It owns a main panel which is displayed in the center of the
		 * browser and can freely be defined. You can use the paint(...)-method
		 * of the main panel, plotting by default the hovlet banner. See <a
		 * href="#main">main</a>.
		 * <p>
		 * </li>
		 * <li>It is able to indicate a status line in the browser. See <a
		 * href="#status(java.lang.String)">status(...)</a>.
		 * <p>
		 * </li>
		 * </menu>
		 *
		 * A hovlet must be able to connect to the Hoverball Simulator. For
		 * this, it has the methods <a
		 * href="#connect(java.lang.String)">connect(...)</a> and <a
		 * href="#disconnect()">disconnect()</a> at its disposal, but the
		 * connecting itself is not yet implemented.
		 */
public abstract class Hovlet {
	// ---- Browser Access

	final boolean upward;
	final boolean downward;

	Hovlet parent_ = null; // internal Hovlet-Structure
	Hovlet[] hovlets_ = new Hovlet[0];

	Browser browser = null; // link to Browser (null = not fixed!)
	Point location = null; // "soft" Browser property for location

	static final int DISCONNECTED = 0; // Connection States...
	static final int CONNECTING = 1;
	static final int CONNECTED = 2;
	static final int DISCONNECTING = 3;

	int state = DISCONNECTED; // current Connection-State
	String servertext = ""; // buffer for Server-TextField
	String status = null; // Status Line

	// ---- Thread Group

//APPLET THREAD BUG//
   static ThreadGroup group = Thread.currentThread().getThreadGroup(); //new ThreadGroup("hovlets");

	// ---- Default Panel

	private static class DefaultPanel extends JPanel {
		private final String TITLE;
		private final String CLASS;

		DefaultPanel(Hovlet hovlet) {
			super(new BorderLayout());
			setBackground((background != null) ? background : getBackground()
					.darker());
			setForeground((foreground != null) ? foreground : Color.black);

			TITLE = hovlet.title;
			String CLASS_ = "." + hovlet.getClass().getName();
			CLASS = "class " + CLASS_.substring(CLASS_.lastIndexOf('.')+1);
		}

		public void paintComponent(Graphics g) {
			super.paintComponent(g);
			Layout.drawBanner(g, getSize(), TITLE, CLASS, true);
		}
	}

	// ---- public -------------------------------------------------------------

	/**
			 * Default background color for all new hovlets
			 */
public static Color background = null;

	/**
			 * Default foreground color for all new hovlets
			 */
public static Color foreground = null;

	/**
			 * Height of the hovlet banner (in pixels)
			 */
public static final int BANNER = Layout.BANNER;

	/**
			 * Super-hovlet (<b><code>null</code></b>, if root)
			 */
public Hovlet parent = parent_; // transparent Hovlet-Structure

	/**
			 * Sub-hovlets
			 */
public Hovlet[] hovlets = hovlets_;

	/**
			 * Title of the hovlet
			 */
public final String title; // Hovlet's properties...

	/**
			 * Menu bar of the hovlet
			 */
public final JMenuBar menubar;

	/**
			 * Icon panel of the hovlet
			 */
public final JPanel icon;

	/**
			 * Main panel of the hovlet
			 */
public final JPanel main;

	// ---- Constructor --------------------------------------------------------

	/**
			 * Creates an empty hovlet.
			 *
			 * @param title
			 *            title of the hovlet
			 */
public Hovlet(String title) {
		this(title, true, true);
	}

	/**
			 * Creates an empty hovlet with the indicated connectabilities.
			 *
			 * @param title
			 *            title of the hovlet
			 * @param upward
			 *            connectable upward?
			 * @param downward
			 *            connectable downward?
			 */
public Hovlet(String title, boolean upward, boolean downward) {
		this.title = title;

		this.upward = upward;
		this.downward = downward;

		main = new DefaultPanel(this);
		icon = new JPanel(new BorderLayout());

		menubar = new JMenuBar();
		menubar.add(new JMenu("Hovlet"));
	}

	private boolean closed = false;

	/**
			 * Closes the hovlet.
			 */
public void close() {
////for(Hovlet i=this; i.parent!=null; i=i.parent) System.out.print("  ");
////System.out.println("closing: "+title+"...");
		if (closed)
			return;
		closed = true;

		for (int i = 0; i < hovlets_.length; ++i)
			hovlets_[i].close();

		disconnect();
////for(Hovlet i=this; i.parent!=null; i=i.parent) System.out.print("  ");
////System.out.println("closed : "+title+".");
	}

	/**
			 * Returns the title of the hovlet with its path.
			 *
			 * @param separator
			 *            separating character string for the path
			 * @return title with path
			 */
public String title(String separator) {
		if (parent_ == null || separator == null)
			return title;
		else
			return parent_.title(separator) + separator + title;
	}

	/**
			 * Adds a hovlet to this hovlet.
			 * <p>
			 * If the connection of the hovlets is not valid, or if one of them
			 * is already connected or displayed, connection cannot succeed.
			 *
			 * @param hovlet
			 *            hovlet
			 * @return <b><code>true</code></b>, if connection succeeded <b><code>false</code></b>,
			 *         if connection failed
			 */
public boolean add(Hovlet hovlet) {
		if ((browser != null || !downward)
				|| (hovlet.browser != null || !hovlet.upward)
				|| hovlet.parent_ != null)
			return false; // addition failed.

		Hovlet[] array = hovlets_;
		hovlets_ = new Hovlet[array.length + 1];
		System.arraycopy(array, 0, hovlets_, 0, array.length);
		hovlets_[array.length] = hovlet;

		hovlet.parent_ = this;

		hovlet.parent = hovlet.parent_;
		hovlets = hovlets_;

		return true;
	}

	/**
			 * Shows the hovlet.
			 * <p>
			 * If this is the first call of show() by all hovlets within the
			 * hovlet tree, a new browser is opened showing the hovlet.
			 * Otherwise the hovlet is simply shown.
			 */
public synchronized void show() {
		if (browser == null) // if not yet shown:
			new Browser(this); // open Browser and setup browser-links!

		browser.show(this);
	}

	/**
			 * Displays a text in the status line of the browser.
			 * <p>
			 * The argument <b><code>null</code></b> displays the default
			 * status of the browser.
			 *
			 * @param status
			 *            status
			 */
public void status(String status) {
		this.status = status;
		if (browser != null)
			browser.status(this, null);
	}

	/**
			 * Demands the browser position.
			 * <p>
			 * If the hovlet is a root-hovlet, the browser will be positioned on
			 * opening at the indicated place.
			 *
			 * @param x
			 *            <i>x</i>-coordinate of the required browser position
			 * @param y
			 *            <i>y</i>-coordinate of the required browser position
			 */
public void locate(int x, int y) {
		locate(new Point(x, y));
	}

	/**
			 * Demands the browser position.
			 * <p>
			 * If the hovlet is a root-hovlet, the browser will be positioned on
			 * opening at the indicated place.
			 *
			 * @param location
			 *            required browser position
			 */
public void locate(Point location) {
		if (location == null)
			this.location = null;
		else
			this.location = new Point(location.x, location.y);
	}

	/**
			 * Alters the looks of the browser if the hovlet is the root-hovlet.
			 * <p>
			 * If a hovlet tree has to be shown in a browser, it is the
			 * root-hovlet that determines its looks by modifying the frame of
			 * the browser. The browser itself is not yet visible.
			 * <p>
			 * This method is empty and can be overwritten.
			 *
			 * @param browser
			 *            frame of the browser
			 */
public void browse(JInternalFrame browser) {}

	// ---- Connection Commands ------------------------------------------------

	/**
			 * Connects the hovlet to the specified address.
			 *
			 * @param address
			 *            network address
			 * @return <b><code>true</code></b>, if connection succeeded
			 */
public synchronized boolean connect(String address) {
        if (state == CONNECTED)
			disconnect(); // disconnect first!
		if (state != DISCONNECTED)
			return false;
		servertext = (address == null)? "" : "" + address;
		if (servertext.startsWith(""+Address.PORT)) servertext = "localhost" + servertext; // ":1234" >> "localhost:1234"
		
		state = CONNECTING;
		if (browser != null)
			browser.refresh(this);

////for(Hovlet i=this; i.parent!=null; i=i.parent) System.out.print("  ");
////System.out.println("Hovlet.connecting: "+title+"...");
      // "AWEFUL BAD WORKAROUND"
      if (Application.servertext_with_hash != null) servertext = Application.servertext_with_hash;

		boolean success = connecting(servertext);

      // "AWEFUL BAD WORKAROUND"
      if (Application.servertext_with_hash != null) servertext = Application.servertext_with_hash;
////for(Hovlet i=this; i.parent!=null; i=i.parent) System.out.print("  ");
////System.out.println("Hovlet.connected : "+title+". "+success);

		if (success)
			state = CONNECTED;
		else
			state = DISCONNECTED; // CONNECTION_FAILED;
		if (browser != null)
			browser.refresh(this);

		return success;
	}

	/**
			 * Disconnects the hovlet.
			 */
public synchronized void disconnect() {
		if (state != CONNECTED)
			return;

		state = DISCONNECTING;
		if (browser != null)
			browser.refresh(this);

		disconnecting();

		state = DISCONNECTED;
		if (browser != null)
			browser.refresh(this);
	}

	// ---- Virtual Methods (to override!) -------------------------------------

	/**
			 * Void for the implementation of connecting.
			 * <p>
			 * The method should return <b><code>true</code></b> if the
			 * connection succeeded.
			 * <p>
			 * The present content of this method successively connects all
			 * sub-hovlets to the indicated address and returns <b><code>true</code></b>
			 * if at least one connection succeeded. If required, you can access
			 * this code by <b>super</b>.connecting(...)
			 *
			 * @param address
			 *            network address
			 * @return <b><code>true</code></b>, if connection succeeded
			 */
protected boolean connecting(String address) {
		boolean success = false;
		for (int i = 0; i < hovlets_.length; ++i)
			success |= hovlets_[i].connect(address);
		return success;
	}

	/**
			 * Void for the implementation of disconnecting.
			 * <p>
			 * The present content of this method disconnects sucessively all
			 * sub-hovlets. If required, you can access this code by <b>super</b>.connecting(...)
			 */
protected void disconnecting() {
		for (int i = 0; i < hovlets_.length; ++i)
			hovlets_[i].disconnect();
	}
}
