/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.simulator.Constants;
import hoverball.debug.Text;
import hoverball.layout.Layout;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;


// ---- Human -----------------------------------------------------------------


/**
 * The Human unit is a particular Java unit.
 * It transforms key events into actions which are instantly sent to the Simulator...
 * - in short: you can play Hoverball yourself thanks to the Human unit!
 * <p>
 * For this purpose, you need a key controls that can be defined by the class
 * <a href="Human.Controls.html">Human.Controls</a>.
 * <p>
 * The key input is enabled if the main panel of the human hovlet or
 * the Controller's screen is active.
 * (Of course, the latter works only if the
 * Human unit and the Controller are running on the same Java instance.)
 * <p>
 * <hr>
 * <p>
 * The class Human can also be executed as an application:
 * <p>
 * <tt>java hoverball.Human </tt>[<i>team</i> [<i>name</i> [<i>color</i>] [<i>host</i>][<tt>:</tt><i>port</i>][<tt>#</tt><i>hash</i>] ]]
 * <p>
 * <menu>
 * <li>(without) - Opens a Human unit.</li>
 * <li><i>name</i> - Opens a Human unit with the indicated player name.</li>
 * <li><i>team</i> <i>name</i> - Opens a Human unit with the indicated team and player name.</li>
 * <li>... <i>color</i> - color, e.g <code>FFC800</code> for orange.</li>
 * <li>... <i>host</i><tt>:</tt><i>port</i><tt>#</tt><i>hash</i>  - network address to be connected to.</li>
 * </menu>
 */
public class Human extends Unit
{
   // ---- static Launcher ----------------------------------------------------

   /**
    * Executes a Human unit (application!).
    */
   public static void main (String[] args)
   {
      String title = "Hoverball";
      String team = Layout.firstUpperCase((args.length > 0)? args[0] : "Human");
      String name = Layout.firstUpperCase((args.length > 1)? args[1] : team);
      int color = Layout.hashColor(team);
      
      if (args.length > 2)
    	 try {
    	    color = Integer.parseInt(args[2],16); args[2] = null;
    	 } catch(java.lang.NumberFormatException e) {}
      
      /*
      Application.setTitle(name);
      Application.setBounds();

      Team humans = new Humans(team);
      Human human = new Human(null,name,color); //,(-name.hashCode()) & 0xFFFFFF,Human.Controls.CRSR);

      humans.add(human);
      human.show();

      if (args.length > 2)
         Application.connect(args[args.length-1]); // host
      */
      
      String host = null;
      if (args.length > 2)
          host = args[args.length-1]; // host

      Session session = (host == null)? new Session(title) : new Session(title,host);

      Team humans = new Humans(team);
      Human human = new Human(null,name,color); //,(-name.hashCode()) & 0xFFFFFF,Human.Controls.CRSR);

      humans.add(human);
      session.add(humans,true);

      session.controller.follow(human);

      human.show();
      session.controller.show();
      session.controller.viewer.show();

// Stay "Initialized." -- for other players to sign in...
//      
//      session.controller.state(1);
//      session.controller.state(2);
   }

   private static class Humans extends Team
   {
      public Humans (String name) { super(name,null,null,null); }
   }


   // ---- static -------------------------------------------------------------

   /**
    * KeyListener object for all Human units
    */
   public static final KeyListener listener = new KeyboardListener();

   /**
    * Assigns a controls to a unit.
    * <p>
    * The assignment makes the unit send corresponding actions on keypress.
    * By this method, Human units automatically assign their controls on call.
    * <p>
    * If <code>controls</code> equals <b><code>null</code></b>, the assignment
    * for this unit is canceled.
    *
    * @param unit unit
    * @param controls controls
    */
   public static final void assign (Unit unit, Human.Controls controls)
   {
      ((KeyboardListener)Human.listener).assign(unit,controls);
   }

   /**
    * Imitates the release of all keys.
    * <p>
    * This method should be called if an AWT component, on which the
    * human KeyListener is enabled, loses its focus.
    */
   public static final void untype ()
   {
      ((KeyboardListener)Human.listener).untype();
   }


   // ---- Hovlet Design... ---------------------------------------------------

   private static final int padx = Layout.padx;
   private static final int tabx = Layout.tabx;
   private static final Dimension iconsize = new Dimension(20,11);

   private class Display extends JPanel  // -----------------------[Banner]----
   {
      final String CONTROLS = "Controls:";
      final String SEPARATOR = "  -  ";
      final String[] MOVE = new String[] {
         "Move forward",
         "Move back",
         "Turn left",
         "Turn right",
         "Shoot ball",
         "Attract ball"
      };

      final Human.Controls controls;

      boolean active = false;

      final Dimension icon = Human.iconsize;
      final int padx = Human.padx;
      final int tabx = Human.tabx;
      int tab2 = 0;
      int tab3 = 0;
      final int pady = 16;
      final int taby = 18;
      final int insy = 8;

      public boolean isFocusTraversable() { return true; }

      public Display (Human.Controls controls)
      {
         this.controls = controls;

         setOpaque(false);  // make main's banner visible!

         addKeyListener(Human.listener);
         addFocusListener(new FocusAdapter() {
            public void focusGained (FocusEvent e) { active = true;  repaint(); }
            public void focusLost (FocusEvent e)   { active = false; repaint(); Human.untype(); }
         });
         addMouseListener(new MouseAdapter() {
            public void mousePressed (MouseEvent e) { requestFocus(); }
         });

         FontMetrics mp = main.getFontMetrics(Layout.fontMain);
         FontMetrics mb = main.getFontMetrics(Layout.fontMainBold);

         tab2 = mp.stringWidth(CONTROLS)+14;
         if (controls == null) tab3 = 0;
         else for (int i=0; i<MOVE.length; ++i)
            tab3 = Math.max(tab3,mb.stringWidth(controls.getKeyText(i)));
      }

      public void paintComponent (Graphics g) { super.paintComponent(g);
         Dimension size = getSize();

         boolean active = this.active || ((KeyboardListener)Human.listener).screen_active;

         Color cb = (active && controls != null)? Layout.colorBackgroundActive : getBackground();
         Color cf = (active && controls != null)? Layout.colorForegroundActive : getForeground();

         if (active && controls != null) {
            g.setColor(cb);
            g.fillRect(0,Layout.BANNER+pady-insy,size.width,5*taby+icon.height+2*insy);
         }
         g.setColor(cf);

         Layout.drawIcon(g,padx,Layout.BANNER+pady,icon.width,icon.height,new Color(connect.color));

         Font FB = Layout.fontMainBold;
         Font FP = Layout.fontMain;

         int y = Layout.BANNER+pady+icon.height-1;
         g.setFont(FP);
         g.drawString(CONTROLS,padx+tabx,y);

         if (controls == null) {
            g.setFont(FB);
            g.drawString("null",padx+tabx+tab2,y);
         } else for (int i=0; i<MOVE.length; ++i) {
            g.setFont(FB);
            g.drawString(controls.getKeyText(i),padx+tabx+tab2,y+i*taby);
            g.setFont(FP);
            g.drawString(SEPARATOR+MOVE[i],padx+tabx+tab2+tab3,y+i*taby);
         }
      }
   }


   // ---- Constructor --------------------------------------------------------

   /**
    * Opens a Human unit with CRSR controls.
    */
   public Human ()
   {
      this("Human");
   }

   /**
    * Opens a Human unit with CRSR controls.
    *
    * @param name unit name
    */
   public Human (String name)
   {
      this(name,name);
   }

   /**
    * Opens a Human unit with CRSR controls.
    *
    * @param team team name
    * @param name unit name
    */
   public Human (String team, String name)
   {
      this(team,name,hashColor(name));
   }

   /**
    * Opens a Human unit with CRSR controls.
    *
    * @param name unit name
    * @param color color
    */
   public Human (String name, int color)
   {
      this(name,name,color);
   }

   /**
    * Opens a Human unit with CRSR controls.
    *
    * @param team team name
    * @param name unit name
    * @param color color
    */
   public Human (String team, String name, int color)
   {
      this(team,name,color,Human.Controls.CRSR);
   }

   /**
    * Opens a Human unit with indicated controls.
    *
    * @param team team name
    * @param name unit name
    * @param color color
    * @param controls controls
    */
   public Human (String team, String name, int color,
                 Human.Controls controls)
   {
      super(team,name,color);

      if (controls != null) Human.assign(this,controls);
      main.add(new Display(controls),"Center");
   }

   /**
    * Empty.
    */
   protected void loop()
   {
      while(look()) {
         for (int i=0; i<pucks.length; ++i)
            if(pucks[i].what == Puck.UNIT && pucks[i].X != null) {
               Puck unit = pucks[i];
               if (unit == self)
                  debug(new Text(unit.X.c,format_(energy),2.2),-1);
               else if (unit.message.startsWith("  ") && unit.message.endsWith("  ")) // Humans send their names! :-)
                  debug(new Text(unit.X.c,unit.message,2.2),-1);
            }

         action("  "+title+"  "); // Trick: with "  "'s around! :-)
      }
   }

   private static String format_(double x)
   {
      int X = (int)(x*100);
      int a = X/100;
      int b = (X/10)%10;
      int c = X%10;
      return ""+a+"."+b+""+c;
   }


// ---- Human Controls --------------------------------------------------------

/**
 * The class Human.Controls defines controls for Human units.
 * For now, only key controls are possible.
 */
public static class Controls
{
   /**
    * Default key controls: cursor keys + control + shift
    */
   public static final Human.Controls CRSR =
      new Human.Controls(KeyEvent.VK_UP,
                         KeyEvent.VK_DOWN,
                         KeyEvent.VK_LEFT,
                         KeyEvent.VK_RIGHT,
                         KeyEvent.VK_SHIFT,
                         KeyEvent.VK_CONTROL);

   static final int NONE = 0;
   static final int FORWARD = 1;
   static final int BACK = 2;
   static final int LEFT = 3;
   static final int RIGHT = 4;
   static final int POSITIVE = 8;
   static final int NEGATIVE = 9;

   int forward = 0;
   int back = 0;
   int left = 0;
   int right = 0;
   int positive = 0;
   int negative = 0;

   private String controls = "\0\0\0\0\0\0";

   /**
    * Defines a key controls.
    * <p>
    * The arguments correspond to the KeyEvent codes.
    *
    * @param forward  move forward
    * @param back     move backward
    * @param left     turn left
    * @param right    turn right
    * @param positive polarize positive
    * @param negative polarize negative
    */
   public Controls (int forward, int back, int left, int right, int positive, int negative)
   {
      this.forward = forward;
      this.back = back;
      this.left = left;
      this.right = right;
      this.positive = positive;
      this.negative = negative;
   }

   /**
    * Defines a key controls.
    * <p>
    * The first 6 characters of the controls string define the keys for:
    * <ul>
    * <li>move forward
    * <li>move backward
    * <li>turn left
    * <li>turn right
    * <li>polarize positive
    * <li>polarize negative
    * </ul>
    *
    * @param controls controls string
    */
   public Controls (String controls)
   {
      if (controls == null) return;

      this.controls = (controls.toUpperCase()+this.controls).substring(0,6);
   }

   /**
    * Returns a string representation of each key controls.
    * <p>
    * The argument <code>control</code> means:
    * <ul>
    * <li>0 = move forward
    * <li>1 = move backward
    * <li>2 = turn left
    * <li>3 = turn right
    * <li>4 = polarize positive
    * <li>5 = polarize negative
    * </ul>
    *
    * @param control required controls
    */
   public String getKeyText (int control) // 0: forward, 1: back, 2: left, ...
   {
      if (control == 0) { if (forward  != 0) return KeyEvent.getKeyText(forward ); else return ""+controls.charAt(0); }
      if (control == 1) { if (back     != 0) return KeyEvent.getKeyText(back    ); else return ""+controls.charAt(1); }
      if (control == 2) { if (left     != 0) return KeyEvent.getKeyText(left    ); else return ""+controls.charAt(2); }
      if (control == 3) { if (right    != 0) return KeyEvent.getKeyText(right   ); else return ""+controls.charAt(3); }
      if (control == 4) { if (positive != 0) return KeyEvent.getKeyText(positive); else return ""+controls.charAt(4); }
      if (control == 5) { if (negative != 0) return KeyEvent.getKeyText(negative); else return ""+controls.charAt(5); }
      return null;
   }

   // ----

   int check (KeyEvent e)
   {
      String T = KeyEvent.getKeyText(e.getKeyCode()).toUpperCase();         // Probleme bei einem Human.Controls("...")
      String S = (T.length() == 1)? T : (""+e.getKeyChar()).toUpperCase();  // für Standard-Tasten (A..Z,0..9)
      char s = S.charAt(0);                                                 // und Shift/Control (-> CRSR!) vermeiden

      if (forward  == 0 && s == controls.charAt(0)) forward  = e.getKeyCode();
      if (back     == 0 && s == controls.charAt(1)) back = e.getKeyCode();
      if (left     == 0 && s == controls.charAt(2)) left = e.getKeyCode();
      if (right    == 0 && s == controls.charAt(3)) right = e.getKeyCode();
      if (positive == 0 && s == controls.charAt(4)) positive = e.getKeyCode();
      if (negative == 0 && s == controls.charAt(5)) negative = e.getKeyCode();

      if (e.getKeyCode() == forward ) return FORWARD;
      if (e.getKeyCode() == back    ) return BACK;
      if (e.getKeyCode() == left    ) return LEFT;
      if (e.getKeyCode() == right   ) return RIGHT;
      if (e.getKeyCode() == positive) return POSITIVE;
      if (e.getKeyCode() == negative) return NEGATIVE;
      return NONE;
   }
}

}


// ---- Human's static KeyListener --------------------------------------------


class KeyboardListener extends KeyAdapter
{
   // ---- Display activated when Screen is activated -------------------------

   public boolean screen_active = false;
   public void screen_active (boolean active) {
      screen_active = active;
      for (Enumeration<Human> e = assign.keys(); e.hasMoreElements(); )
         ((Human)e.nextElement()).main.repaint();
   }

   // ---- KeyboardListener --------------------------- [KeyboardListener] ----

   private Hashtable assign = new Hashtable();  // unit -> controls
   private Vector pressed = new Vector();

   public void assign (Unit unit, Human.Controls controls)
   {
      if (controls != null)  // assign
         assign.put(unit,controls);
      else {                 // remove
         if (!assign.containsKey(unit)) return;
         assign.remove(unit);
         action(unit,null);
      }
   }

   public void untype ()
   {
      pressed.removeAllElements();
      for (Enumeration<Unit> i = assign.keys(); i.hasMoreElements(); )
         action((Unit)i.nextElement(),null);
   }

   // ---- Listen ...

   public void keyPressed (KeyEvent e)
   {
      if (pressed.contains(""+e.getKeyCode())) return;
      pressed.addElement(""+e.getKeyCode());

      for (Enumeration<Unit> i = assign.keys(); i.hasMoreElements(); ) {
         Unit unit = (Unit)i.nextElement();
         Human.Controls controls = (Human.Controls)assign.get(unit);

         if (controls.check(e) == Human.Controls.NONE) continue;
         action(unit,controls);
      }
   }

   public void keyReleased (KeyEvent e)
   {
      if (!pressed.contains(""+e.getKeyCode())) return;
      pressed.removeElement(""+e.getKeyCode());

      for (Enumeration<Unit> i = assign.keys(); i.hasMoreElements(); ) {
         Unit unit = (Unit)i.nextElement();
         Human.Controls controls = (Human.Controls)assign.get(unit);

         if (controls.check(e) == Human.Controls.NONE) continue;
         action(unit,controls);
      }
   }

   // ---- Action!

   private static final double TURN = 0.5;

   private void action (Unit unit, Human.Controls controls)
   {
      double Qmin = 1.0/40 *unit.option(Constants.getKey(Constants.UNIT_CHARGE_MIN));
      double Qmax = 1.0    *unit.option(Constants.getKey(Constants.UNIT_CHARGE_MAX));
      double Fmin = 1.0 *unit.option(Constants.getKey(Constants.UNIT_ENGINE_MIN));
      double Fmax = 1.0 *unit.option(Constants.getKey(Constants.UNIT_ENGINE_MAX));

      double P = 0;
      double L = 0;
      double R = 0;

      if (controls != null) {
         if (pressed.contains(""+controls.forward))  { L += Fmax;      R += Fmax;      }
         if (pressed.contains(""+controls.back))     { L += Fmin;      R += Fmin;      }
         if (pressed.contains(""+controls.left))     { L += Fmin*TURN; R += Fmax*TURN; }
         if (pressed.contains(""+controls.right))    { L += Fmax*TURN; R += Fmin*TURN; }
         if (pressed.contains(""+controls.negative)) { P = Qmin; }
         if (pressed.contains(""+controls.positive)) { P = Qmax; } // hier umgekehrt, damit er bei gleichzeitigem Drücken schießt
      }
/*
      if (L >  unit.engine_max) { R -= (L - unit.engine_max); L =  unit.engine_max; }
      if (R >  unit.engine_max) { L -= (R - unit.engine_max); R =  unit.engine_max; }
      if (L < -unit.engine_max) { R -= (L + unit.engine_max); L = -unit.engine_max; }
      if (R < -unit.engine_max) { L -= (R + unit.engine_max); R = -unit.engine_max; }
*/
      unit.action(P,L,R);
   }
}
