/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.simulator.Constants;

import hoverball.serial.*;  // CS.Set
import hoverball.layout.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;


// ---- Hoverball Option Window -------------------------------- [Options] ----


class Options
{
   // ---- interface Options.Listener -----------------------------------------

   static interface Listener
   {
      public void listen (Object message);
   }


   // ---- class TextfieldListener --------------------------------------------

   private class TextfieldListener
      implements ActionListener, DocumentListener
   {
      private final int i;
      public TextfieldListener (int i) { this.i = i; }

      public void actionPerformed (ActionEvent e) { apply(i); }

      //      public void textValueChanged (TextEvent e) { change(i); }
      public void insertUpdate(DocumentEvent e) { change(i); }
      public void removeUpdate(DocumentEvent e) { change(i); }
      public void changedUpdate(DocumentEvent e) { change(i); }
   }

   // ---- class Options ------------------------------------------------------

   public final JPanel panel;

   private boolean connected = false;
   private boolean updated = false;

   private Options.Listener listener;
   private Constants constants;

   private final JScrollPane scroll;
   private final JButton defaults;
   private final JLabel label;
   private final JButton reset;
   private final JButton apply;

   private JTextField[] textfield = new JTextField[Constants.CONSTANTS];
   private boolean[] changed = new boolean[Constants.CONSTANTS];


   // ---- Constructors -------------------------------------------------------

   public Options (Constants constants, Options.Listener listener)
   {
      this.constants = constants;
      this.listener = listener;

      panel = new JPanel();

      // ---- Constant Panel

      JPanel table = new JPanel(new GridBagLayout());

      String lastClass = null;

      for (int i=0; i<Constants.CONSTANTS; ++i) {
         String A = Constants.getExplanation(i);
         String B = "";

         int s = A.indexOf(Constants.UNIT_SEPARATOR);
         if (s >= 0) {
            B = A.substring(s+1).trim();
            A = A.substring(0,s).trim();
         }

         JLabel unit = new JLabel(B);
         unit.setFont(Layout.fontItalic);

         textfield[i] = new JTextField(6);
         TextfieldListener l = new TextfieldListener(i);
         textfield[i].getDocument().addDocumentListener(l);
         textfield[i].addActionListener(l);
         textfield[i].setEnabled(false);
         textfield[i].setFont(Layout.fontPlain);

         JLabel key = new JLabel(Constants.getKey(i));
         key.setFont(Layout.fontFixed);

         final int padx = 8;
         final int pady = 1;

         String thisClass = key.getText().substring(0,key.getText().indexOf('.'));
         if (!thisClass.equals(lastClass)) {
            if (lastClass != null)
               table.add(new JSeparator(JSeparator.HORIZONTAL),
                         new Constraints(0,i*2,4,1,1.0,0.0,new Insets(pady*2,0,pady*2,0)));
            lastClass = thisClass;
         }

         table.add(key,
                   new Constraints(0,i*2+1,1,1,0.0,0.0,new Insets(pady,padx*2,pady,padx*8)));
         table.add(textfield[i],
                   new Constraints(1,i*2+1,1,1,0.0,0.0,new Insets(pady,0,pady,padx)));
         table.add(unit,
                   new Constraints(2,i*2+1,1,1,0.0,0.0,new Insets(pady,0,pady,padx*2)));

         key.setToolTipText(A);
      }

      JPanel table_ = new JPanel(new GridBagLayout());
      table_.add(table,
                 new Constraints(0,0,1,1,0.0,0.0,new Insets(5,0,5,0)));

      scroll = new JScrollPane(table_,JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

      JPanel buttons = new JPanel(new GridBagLayout());

      buttons.add(defaults = new MyButton("Defaults",false,70),
                  new Constraints(0,0,1,1,0.0,0.0,new Insets(0,0,0,8)));
      buttons.add(label  = new JLabel(), // { public boolean isFocusTraversable() { return true; } },
                  new Constraints(1,0,1,1,1.0,0.0,new Insets(0,0,0,8)));
      label.setFont(Layout.fontPlain);
      buttons.add(reset = new MyButton("Get",false,60),
                  new Constraints(3,0,1,1,0.0,0.0,new Insets(0,0,0,4)));
      buttons.add(apply = new MyButton("Set",false,60),
                  new Constraints(4,0,1,1,0.0,0.0,null));

      defaults.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { defaults(); }
      });

      reset.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { read(); }
      });

      apply.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { apply(); }
      });

      defaults.setEnabled(false);
      reset.setEnabled(false);
      apply.setEnabled(false);

      panel.setLayout(new GridBagLayout());

      panel.add(scroll,
            new Constraints(0,0,1,1,1.0,1.0,new Insets(0,0,0,0)));
      panel.add(buttons,
            new Constraints(0,1,1,1,1.0,0.0,new Insets(4,8,2,8)));
   }

   // ---- public access...

   public void set (final String key, final String value)
   {
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      int i = constants.indexOf(key);
      if (i < 0) return;

      textfield[i].setText(read(i));
      change(i);

      update();
         }
      });
   }

   public void connected (final boolean connected)  // must be within GUI thread!
   {                                                // called by Controller.connected()/disconnected()
      final Options this_ = this;

      if (this_.connected == connected) return;
      this_.connected = connected;

      for (int i=0; i<textfield.length; ++i) textfield[i].setEnabled(connected);

      updated = false;
      if (connected) read();  // read & update
      else {
         for (int i=0; i<textfield.length; ++i) textfield[i].setText("");
         update();            // clear & update
      }
   }


   // ---- private methods...

   private void change (int i)
   {
      boolean change = (!read(i).equals(textfield[i].getText()));
      if (change != changed[i]) textfield[i].setFont(change? 
            Layout.fontBold : 
            Layout.fontPlain);
      changed[i] = change;

      updated = false;
      update();
   }

   private String read (int i)
   {
      if (!connected) return "";

      String value = (String)constants.get(Constants.getKey(i));
      if (value != null) return value;
      else return Constants.getDefault(i);
   }

   // ---- Actions

   private void read ()
   {
      for (int i=0; i<Constants.CONSTANTS; ++i) {
         textfield[i].setText(read(i));
         change(i);
      }
      update();
   }

   private void defaults ()
   {
      for (int i=0; i<textfield.length; ++i) {
         textfield[i].setText(Constants.getDefault(i));
         change(i);
      }
      update();
   }

   private void apply ()
   {
      for (int i=0; i<changed.length; ++i) apply(i);
   }

   private void apply (int i)
   {
      if (!changed[i]) return;

      String text = textfield[i].getText().trim();
      if (text.equals("")) text = null;

      if (connected) listener.listen(new CS.Set(Constants.getKey(i),text));
   }
   
   // ---- Layout

   private void update ()
   {
      if (updated) return;

      int counter = 0;       // update Apply-Button
      for (int i=0; i<changed.length; ++i) {
         if (changed[i]) ++ counter;
      }
      apply.setEnabled(connected && counter != 0);
      reset.setEnabled(connected && counter != 0);

      int not_defaults = 0;  // update Defaults-Button
      for (int i=0; i<textfield.length; ++i) {
         String text = textfield[i].getText().trim();
         if (!(text.equals("") || text.equals(Constants.getDefault(i)))) ++ not_defaults;
      }
      defaults.setEnabled(connected && not_defaults != 0);

      label.setText((!connected)? "" :  // update Label
                    (counter == 0)? "Options up to date." :
                                    "Options changed:  "+counter);
      updated = true;
   }
}

