/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.math.*;
import hoverball.serial.DiskInfo;  // import WHATs, putString


// ---- Hoverball Object Info ------------------------------------- [Puck] ----


/**
 * The class Puck memorizes all information about a puck that can be perceived or
 * deduced by a unit.
 * This can be:
 *
 * <ul>
 * <li>object identity composed of the identification <code>(<i>what t n</i>)</code>,</li>
 * <li>object message, if the object is a unit,
 * <li>position and looking direction in the form of a orthonormal matrix,</li>
 * <li>radius,</li>
 * <li>mass.</li>
 * </ul>
 *
 * The nodes are also represented by the class Puck, although actually
 * they are no real Hoverball pucks.
 * <p>
 * With the class Puck you have at your disposal various methods that allow to
 * organize and compare objects.
 */
public class Puck
{
   /**
    * type constant: node
    */
   public static final int NODE = DiskInfo.NODE;

   /**
    * type constant: ball
    */
   public static final int BALL = DiskInfo.BALL;

   /**
    * type constant: unit
    */
   public static final int UNIT = DiskInfo.UNIT;


   /**
    * object identity (as string)
    */
   public final String id;  // "what t n"

   /**
    * object type
    */
   public final int what;

   /**
    * object team number
    */
   public final int t;

   /**
    * object puck number (= player number if unit)
    */
   public final int n;

   /**
    * object radius
    */
   public final double r;

   /**
    * object mass
    */
   public final double m;

   /**
    * object position (as orthonormal matrix)
    */
   public Matrix X;

   /**
    * object message (if unit)
    */
   public String message;


   /**
    * Creates the identity (as string).
    *
    * @param what object type
    * @param t team number
    * @param n puck number
    * @return the identity
    */
   public static String id (int what, int t, int n)
   {
      return DiskInfo.putWhat(what)+" "+t+" "+n;
   }


   /**
    * Creates a puck equal to another.
    *
    * @param puck copied puck
    */
   public Puck (Puck puck)
   {
      this(puck == null? 0 : puck.what,
           puck == null? 0 : puck.t,
           puck == null? 0 : puck.n,
           puck == null? new Matrix() : puck.X,
           puck == null? 0 : puck.r,
           puck == null? 0 : puck.m,
           puck == null? null : puck.message);
   }

   /**
    * Creates a puck with the specified identity,
    * all other values are null.
    *
    * @param what object type
    * @param t team number
    * @param n puck number
    */
   public Puck (int what, int t, int n)
   {
      this(what,t,n,new Matrix(),0,0,null);
   }

   /**
    * Creates a puck with the specified values.
    *
    * @param what object type
    * @param t team number
    * @param n puck number
    * @param X orientierted position
    * @param r radius
    * @param m mass
    * @param message message
    */
   public Puck (int what, int t, int n, Matrix X, double r, double m, String message)
   {
      this.what = what;
      this.t = t;
      this.n = n;
      this.X = (X==null)? null : new Matrix(X);
      this.r = r;
      this.m = m;
      this.message = (message==null)? null : new String(message);

      this.id = id(what,t,n);
   }


   // ---- system functions

   /**
    * Compares with a puck.
    * <p>
    * Two puck objects are equal iff their identities are equal.
    *
    * @param puck puck
    * @return the equality
    */
   public boolean equals (Object puck)
   {
      if (puck == null) return false;
      if (! (puck instanceof Puck)) return false;

      Puck d = (Puck)puck;
      return equals(d.what,d.t,d.n);
   }

   /**
    * Compares the type.
    *
    * @param what object type
    * @return the equality
    */
   public boolean equals (int what)
   {
      return this.what == what;
   }

   /**
    * Compares type and team.
    *
    * @param what object type
    * @param t team number
    * @return the equality
    */
   public boolean equals (int what, int t)
   {
      return this.what == what && this.t == t;
   }

   /**
    * Compares the identity.
    *
    * @param what object type
    * @param t team number
    * @param n puck number
    * @return the equality
    */
   public boolean equals (int what, int t, int n)
   {
      return this.what == what && this.t == t && this.n == n;
   }

   /**
    * Compares the identity (as string).
    *
    * @param id identity string
    * @return the equality
    */
   public boolean equals (String id)
   {
      if (id == null) return false;
      else return this.id.equals(id.trim());
   }

   /**
    * Returns the hash code.
    *
    * @return the hash code
    */
   public int hashCode ()
   {
      return (what & 0xFF << 16) + (t & 0xFF << 8) + (n & 0xFF);
   }


   /**
    * string representation.
    *
    * @return the string
    */
   public String toString()
   {
      return getClass().getName()
         +"[(" + id + ")"
         +" X=" + (X == null? ""+X : ""+Sphere.euler(X))
         +" r=" + format(r)
         +" m=" + format(m)
         +((message == null)? "" : " " + DiskInfo.putString(message))
         +"]";
   }

   static String format (double d)
   {
      if (Complex.format == null) return ""+d;
      else return Complex.format.format(d);
   }
}
