/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.serial.*;
import hoverball.layout.*;
import hoverball.math.*;
import hoverball.net.Address;
import hoverball.debug.*;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;

import java.util.Enumeration;
import java.util.Hashtable;


// ---- Hoverball Screen ---------------------------------------- [Screen] ----


class Screen extends JPanel
{
   // ---- interface Screen.Listener ------------------------------------------

   static interface Listener                                   // if Screen wants to send
   {                                                           // messages to Simulator
      public void listen (Object message);                     // (not necessary!)
   }

   // ---- class Screen -------------------------------------------------------

   private static final double SPHERE_RATIO    = 0.87;          // wie groß ist Sphere im Screen? (@zoom=1)
   private static final double MOUSE_SENSITIVE = 500;           // wieviel Pixel entspricht einmal rum?
   private static final double FOLLOW_HALFWAY  = 0.1;           // "Halbwegszeit" in Sekunden
   private static final double FOLLOW_FRONT    = 0.075*Math.PI; // wie weit vor Units (in sphär.Winkel)
   private static final double FOLLOW_TOLERANCE= 0.2;           // Zeitabstand, innerhalb welchem der view
                                                                // noch als direkter Nachfolger-view betrachtet wird
   private static final double ZOOM_MIN = 1;                    // minimaler Zoom-Faktor
   private static final double ZOOM_MAX = 4;                    // maximaler Zoom-Faktor
   private static final double ZOOM_HALFWAY = 1;                // "Halbwertszeit" des Zooms in Sekunden
   private static final int MEMORY = 9;                         // Anzahl der View-Speicherplätze

   private static final double SCOREHIGHLIGHTING_EXP_CONSTANT = 0.004;

   private static final int FLAGS = Layout.RESOLUTION;   // should be multiple of 4 (now: 60)
   private static final Vector[] flags = new Vector[3*FLAGS/2];// "/2": paint mirror-pairs!

   static {
      for (int i=0; i<FLAGS/2; ++i) {     // init 3D-flags
         Complex c = Complex.expi(Math.PI*2*i/FLAGS);
         flags[i+0*FLAGS/2] = new Vector(0,c.x,c.y);
         flags[i+1*FLAGS/2] = new Vector(c.y,0,c.x);
         flags[i+2*FLAGS/2] = new Vector(c.x,c.y,0);
      }
   }

   // ----- un-static

   private final Screen screen = this;            // Reference for Inner Classes

   private boolean antialiasing = true;           // Antialiasing, Text_Antialiasing
   private boolean fullscreen = false;            // Full Screen Mode
   private double fullscreen_scale = 1.0;         // Enlarge annotations! (will dynamically be set)

   private Screen.Listener listener = null;

   private Address server = null;                 // Opening Parameters
   private Hashtable channels = new Hashtable();
   private double frequency = 0;
   private double duration = 0;
   private Sphere sphere = null;

   public Hashtable views = new Hashtable();      // View Memory (accessed by Controller)

   private boolean opened = false;                // Game State
   private boolean paused = false;
   private View view = null;

   private BufferedImage offscreen = null;        // Graphics
   private Graphics2D graphics = null;
   private boolean updated = false;

   private double scale = 0;                      // Internal Rendering
   private double zoom = 1;                       // Zoom-Factor
   private int zooming = 0;                       // Zooming? (-1,0,+1)
   private Matrix base = new Matrix();            // Globe-Base

   private Follower follower = new Follower();    // Follow Mode
   private DiskAbsolute follow = null;            // Disk to follow

   private int shotballs = 0;
   private int teamballs = 0;

   private static final int DEBUG_LIMIT = 1000;   // Debug Memory Size
   private Hashtable debuglist = new Hashtable(); // Channel -> Vector of Debugs
   private Hashtable debugging;                   // Unit -> Boolean

   private int scorehighlighting_t = 0;
   private long scorehighlighting_hitTimeMillis = 0;

   private final DragAndDrop dragAndDrop;         // Drag And Drop

   // ---- Constructor --------------------------------------------------------

   public Screen (boolean shortcuts, Hashtable debugging, Screen.Listener listener)
   {
      this.debugging = debugging;
      this.listener = listener;

      base = base.mul(Matrix.rotX(Math.PI*Math.random())); /*immutable*/
      base = base.mul(Matrix.rotY(Math.PI*Math.random())); /*immutable*/
      base = base.mul(Matrix.rotZ(Math.PI*Math.random())); /*immutable*/

      // ---- Properties to make fast...

      setOpaque(true);           // true by default
      setDoubleBuffered(false);  // I maintain double buffer myself!

      // ---- Mouse Listener

      dragAndDrop = new DragAndDrop(shortcuts);  // will be invoked by MouseScroller!

      MouseScroller scroller = new MouseScroller();
      addMouseListener(scroller);
      addMouseMotionListener(scroller);

      // ---- Keyboard Listeners

      if (shortcuts) addKeyListener(new ControlKeys(this));
      addKeyListener(new ViewKeys(this));

      // ---- Human Keyboard Listener

      addKeyListener(Human.listener);
      addFocusListener(new FocusAdapter() {
         public void focusLost(FocusEvent e) { Human.untype(); }
      });
      addFocusListener(new FocusAdapter() {  // Display Human activated!
         public void focusGained(FocusEvent e) { ((KeyboardListener)Human.listener).screen_active(true);  }
         public void focusLost  (FocusEvent e) { ((KeyboardListener)Human.listener).screen_active(false); }
      });

      // ---- Mouse Key Input Requester

      addMouseListener(new MouseAdapter() {
         public void mousePressed (MouseEvent e) { requestFocus(); }
      });
   }

   public synchronized void update () { dragAndDrop.update(); updated = false; repaint(); }

   public synchronized boolean complete () { return view != null && view.time == duration; }


   // ---- Screen's public methods --------------------------------------------

   public String getTitle () { return "Screen"; }

   public synchronized void setAntialiasing (boolean antialiasing)
   {
      this.antialiasing = antialiasing;
      if (graphics!=null) applyAntialiasing(graphics,antialiasing);
   }

   // ---- Full Screen Mode
   JFrame FRAME = null;
   Container FRAME_pane = null;
   Container VIEWER_pane = null;
   // ----

   public synchronized boolean getFullScreen () { return this.fullscreen; }
   
   public synchronized void setFullScreen (boolean fullscreen)
   {
      //if (Applet.applet != null) return;          // running as Applet!
      if (this.fullscreen == fullscreen) return;  // nothing to manage.

      GraphicsDevice DEVICE = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();
      if(!DEVICE.isFullScreenSupported()) return;

      if (FRAME==null) { // Full Screen Frame erzeugen
         FRAME = new JFrame(DEVICE.getDefaultConfiguration());
         FRAME.setTitle("Full Screen");
         if (Layout.icon_png != null) FRAME.setIconImage(new ImageIcon(Layout.icon_png).getImage());
         FRAME.setUndecorated(true);
         FRAME.setResizable(false);
         FRAME_pane = new Container(); // FRAME_pane Schachtelung wegen getForeground() und getBackground()
         FRAME_pane.setLayout(new BorderLayout());
         FRAME.getContentPane().setLayout(new BorderLayout());
         FRAME.getContentPane().add(FRAME_pane);

         VIEWER_pane = this.getParent();
      }

      if (fullscreen)
      {
         FRAME_pane.setForeground(getParent().getForeground());
         FRAME_pane.setBackground(getParent().getBackground());

         VIEWER_SIZE = VIEWER_pane.getParent().getParent().getSize(); // UGLY BUG of Swing

         Application.desktop.setVisible(false);
         VIEWER_pane.remove(this);
         FRAME_pane.add(this);
         DEVICE.setFullScreenWindow(FRAME);

         setVisible(false);   // yeah! (Größe anpassen)
         FRAME.revalidate();  // ..
         setVisible(true);    // ..

         this.requestFocus();
         this.fullscreen = true;
         this.fullscreen_scale = 1.5
            * FRAME.getWidth() / Toolkit.getDefaultToolkit().getScreenSize().width;
         offscreen = null; // neu zeichnen!
      }
      else
      {
         DEVICE.setFullScreenWindow(null);
         FRAME_pane.remove(this);
         VIEWER_pane.add(this);

         if (VIEWER_SIZE != null) { 
        	VIEWER_pane.getParent().getParent().setSize(VIEWER_SIZE); // UGLY BUG of Swing
        	VIEWER_pane.getParent().getParent().revalidate();
         }
         
         Application.desktop.setVisible(true);

         FRAME.dispose(); // Jedesmal neu erzeugen! (ist momentan besser...)
         FRAME = null;
         FRAME_pane = null;
         VIEWER_pane = null;

         this.requestFocus();
         this.fullscreen = false;
         offscreen = null; // neu zeichnen!
      }
   }

   Dimension VIEWER_SIZE = null; // UGLY BUG of Swing
   

   public synchronized void open (Address server, Channel[] channels, double frequency, double duration, double sphere)
   {
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      setBackground(getParent().getBackground());  // fetch Background Color!
         }
      });

      for (int i=0; i<channels.length; ++i) {  // setup channel table:
         Channel channel = channels[i];        // unit-channel
         if (channel == null) continue;
         this.channels.put(""+channel.t+" "+channel.n,channel);
         updateTeamChannel(channel.t);         // team-channel
      }

      this.server = server;
      this.frequency = frequency;
      this.duration = duration;
      this.sphere = new Sphere(sphere);
      opened = true;
      paused = true;

      if (paused || complete()) follower.start();
      /*if (followDefaultUnit != null)*/ followDefault();
   }

   public synchronized void remove (int t, int n)
   {
      Channel channel = (Channel)channels.get(""+t+" "+n);
      if (channel == null) return;

      debuglist.remove(channel);
      channels.remove(""+t+" "+n);

      updateTeamChannel(t);
      update();
   }

   private void updateTeamChannel (int t)
   {
      channels.remove(""+t+" "+0);

      int n = 0; String team = null;
      int r = 0; int g = 0; int b = 0;
      for (Enumeration<Channel> e = channels.elements(); e.hasMoreElements(); ) {
         Channel channel = (Channel)e.nextElement();
         if (channel.t != t) continue;
         r += new Color(channel.color).getRed();
         g += new Color(channel.color).getGreen();
         b += new Color(channel.color).getBlue();
         ++ n;
         team = channel.team;
      }

      if (n > 0) {
         int color = new Color(r/n,g/n,b/n).getRGB();
         channels.put(""+t+" "+0,new Channel(t,0,team,null,color,null));
      }
   }

   public synchronized void view (View view)
   {
      dragAndDrop.release(false);

      boolean firsttime = (this.view == null);
      double progress = firsttime? 0 : (view.time - this.view.time);
      if (progress > FOLLOW_TOLERANCE) progress = 1/frequency;

      if (!firsttime) { // scored? compare new and old view.
         for (int i=0; i<view.teams.length; ++i) {
            int j = 0;
            for (int k=0; k<this.view.teams.length; ++k)
               if (view.teams[i].t == this.view.teams[k].t) j = k;

            if (view.teams[i].score != this.view.teams[j].score) {
               scorehighlighting_t = view.teams[i].t;
               scorehighlighting_hitTimeMillis = System.currentTimeMillis();
            }
         }
      }

      this.view = view;

      if (firsttime) {
         int t = 0;     // count balls
         for (int i=0; i<view.disks.length; ++i) {
            DiskAbsolute disk = view.disks[i];
            if (disk.t != 0 && t == 0) t = disk.t;
            if (disk.what == disk.BALL && disk.t == 0)           ++ shotballs;
            if (disk.what == disk.BALL && disk.t == t && t != 0) ++ teamballs;
         }
      }

      if (follow != null) {
         DiskAbsolute disk = getDisk(follow.what,follow.t,follow.n);
         if (disk == null) follow_leave();  // doesn's exist any more? (possible only if firsttime)
         else target(getDisk(follow.what,follow.t,follow.n));
      }
      if (complete())       // final view...
         follower.start();
      else
         if (!firsttime) follower.step(progress);

      update();
   }

   public synchronized void pause (boolean paused)
   {
      if (!complete())
         if (paused) follower.start();
         else follower.stop();

      dragAndDrop.release(false);
      this.paused = paused;
      update();
   }

   public synchronized void close ()
   {
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      setBackground(Layout.colorBackground);
         }
      });

      if (paused || complete()) follower.stop();
      if (follow == null) target(null);

      dragAndDrop.release(false);
      debuglist.clear();
      server = null;
      channels.clear();
      opened = false;
      view = null;
      shotballs = 0;
      teamballs = 0;
      follower.suspend = false;
      scale = 0;  // don't: offscreen = null;!
      update();
   }


   // ---- Secure Calculation -------------------------------------------------

   private static Matrix secure_mul (Matrix A, Matrix B)  // Secure-ONB-Multiplication
   {
      Vector a = Vector.mul(A.a,B).norm();
      Vector b = Vector.mul(A.b,B);
      Vector c = Vector.vec(a,b).norm();
      b = Vector.vec(c,a);
      return new Matrix(a,b,c);
   }
/*
   private static double secure_acos (double r)           // Secure-arccos-Funktion
   {
      if (r >= 1) return 0;
      if (r <=-1) return Math.PI;
      return Math.acos(r);
   }
*/
   private static double angle (Vector a, Vector b, Vector orientation)  // copied from Vector!
   {
      double dot = Vector.dot(a,b)/a.abs()/b.abs();
      if (dot > 1) dot = 1;
      if (dot <-1) dot =-1;
      if (orientation == null || Vector.dot(Vector.vec(a,b),orientation) > 0)
         return  Math.acos(dot);
      else
         return -Math.acos(dot);
   }

   private static Vector secure_axis (Vector a, Vector b, Vector secure)
   {
      Vector axis = Vector.vec(a,b);
      if (axis.zero()) axis = new Vector(secure);  // secure
      return axis.norm();
   }


   // ---- Get Anything from View ---------------------------------------------

   private java.util.Vector getDisks ()      // returns *sorted* array of DiskAbsolutes
   {
      java.util.Vector disks = new java.util.Vector(view.disks.length);

      int n = 0;
      for (int k=0; k<view.disks.length; ++k) {
         DiskAbsolute disk = view.disks[k];

         int i = n;
         do {                                // search right place
            if (i == 0) break;
            DiskAbsolute read = (DiskAbsolute)disks.elementAt(i-1);
            if ((read.t << 16) + ((read.what == read.BALL? 0 : 1) << 8) + read.n <
                (disk.t << 16) + ((disk.what == disk.BALL? 0 : 1) << 8) + disk.n) break;
            -- i;
         } while(true);

         disks.insertElementAt(disk,i);     // insert
         ++ n;
      }
      return disks;
   }

   private DiskAbsolute getDisk (int what, int t, int n)
   {
      for (int k=0; k < view.disks.length; ++k) {
         DiskAbsolute disk = view.disks[k];
         if (disk.what == what && disk.t == t && disk.n == n) return disk;
      }
      return null;
   }


   // ---- Get Anything from Channels -----------------------------------------

   private String getName (int what, int t, int n)
   {
      if (what == DiskAbsolute.BALL && t == 0) return "Ball"+(shotballs==1? "" : " "+n);

      Channel channel = (Channel)channels.get(""+t+" "+(what==DiskAbsolute.BALL? 0 : n));
      if (channel == null) channel = new Channel(t,n,"???","???",0,null);

      if (n == 0) return channel.team;
      if (what == DiskAbsolute.BALL) return "Ball"+(teamballs==1? "" : " "+n);
      return channel.name;
   }

   private Color getColor (int what, int t, int n)
   {
      if (t == 0) return Color.white;

      Channel channel = (Channel)channels.get(""+t+" "+(what == DiskAbsolute.UNIT? n : 0));
      if (channel != null) return new Color(channel.color);

      return null;  // happens if dead players
   }


   // ---- Painting Utilities (Java2D) ----------------------------------------

   private void drawUnit (Sphere sphere, Matrix base, double radius, Color color, Color globe)
   {
      double r = radius/sphere.rad;
      double r0 = r - 0.1/sphere.rad; // inner radius

      Ellipse2D circ = new Ellipse2D.Double(-r,-r,2*r,2*r);
      Ellipse2D circ0 = new Ellipse2D.Double(-r0,-r0,2*r0,2*r0);
      Arc2D pie = new Arc2D.Double(-r,-r,2*r,2*r,-30,60,Arc2D.PIE);

      Graphics2D graphics2 = (Graphics2D)graphics.create();
      graphics2.scale(scale,scale);

      graphics2.transform(Layout.transform(base));

      graphics2.setColor(Layout.color(Color.black,Layout.fade(Color.black,globe,0.75),base.c.z)); // or just: Color.black
      graphics2.fill(circ);
      graphics2.setColor(Layout.color(color,globe,base.c.z));
      graphics2.fill(circ0);
      graphics2.setColor(Layout.color(Color.black,Layout.fade(Color.black,globe,0.75),base.c.z)); // or just: Color.black
      graphics2.fill(pie);
   }

   private void drawBall (Sphere sphere, Matrix base, double radius, Color color, Color globe)
   {
      double r = radius/sphere.rad;
      double r0 = r - 0.1/sphere.rad; // inner radius

      Ellipse2D circ = new Ellipse2D.Double(-r,-r,2*r,2*r);
      Ellipse2D circ0 = new Ellipse2D.Double(-r0,-r0,2*r0,2*r0);
      Arc2D pie1 = new Arc2D.Double(-r,-r,2*r,2*r,0,90,Arc2D.PIE);
      Arc2D pie2 = new Arc2D.Double(-r,-r,2*r,2*r,180,90,Arc2D.PIE);

      Graphics2D graphics2 = (Graphics2D)graphics.create();
      graphics2.scale(scale,scale);
      //graphics2.setStroke(new BasicStroke((float)(0.1/sphere.rad)));

      graphics2.transform(Layout.transform(base));

      graphics2.setColor(Layout.color(Color.black,Layout.fade(Color.black,globe,0.75),base.c.z)); // or just: Color.black
      graphics2.fill(circ);
      graphics2.setColor(Layout.color(color,globe,base.c.z));
      graphics2.fill(circ0);
      graphics2.setColor(Layout.color(Color.black,Layout.fade(Color.black,globe,0.75),base.c.z)); // or just: Color.black
      graphics2.fill(pie1);
      graphics2.fill(pie2);
   }


   // ---- Paint --------------------------------------------------------------

   private void applyAntialiasing(Graphics2D graphics, boolean antialiasing)
   {
      graphics.setRenderingHint(
            RenderingHints.KEY_RENDERING,
            RenderingHints.VALUE_RENDER_SPEED);
      graphics.setRenderingHint(
            RenderingHints.KEY_TEXT_ANTIALIASING,
            RenderingHints.VALUE_TEXT_ANTIALIAS_ON); // Text Antialiasing immer!
      if(antialiasing) {
         graphics.setRenderingHint(
               RenderingHints.KEY_ANTIALIASING,
               RenderingHints.VALUE_ANTIALIAS_ON);
      } else {
         graphics.setRenderingHint(
               RenderingHints.KEY_ANTIALIASING,
               RenderingHints.VALUE_ANTIALIAS_OFF);
      }
      template = null; // neu zeichnen!
      template_space = null;
      template_globe = null;
      template_zoom = 0;
      template_count = 0;
   }

   public synchronized void paint (Graphics g)
   {
      Dimension size = getSize();
      if (size.width <= 0 || size.height <= 0) return;    // not yet shown.

      // Setup Offscreen

      if (offscreen == null || offscreen.getWidth() != size.width ||
                               offscreen.getHeight() != size.height) {  // create offscreen!
         offscreen = getGraphicsConfiguration().createCompatibleImage(size.width,size.height);
         graphics = offscreen.createGraphics();
         applyAntialiasing(graphics,antialiasing);
         graphics.setBackground(getBackground());
         graphics.clearRect(0,0,size.width,size.height);
         graphics.setFont(Layout.fontViewer);
         graphics.setClip(0,0,size.width,size.height);
         graphics.translate(size.width/2,size.height/2);

         template = null;
         template_space = null;
         template_globe = null;
         template_zoom = 0;
         template_count = 0;

         scale = 0;  // reset
         updated = false;
      }

      // Paint

      if (view == null) {   // ---- closed?
         g.setColor(getBackground());
         Rectangle rect = g.getClipBounds();
         g.fillRect(rect.x,rect.y,rect.width,rect.height);

         return;
      }

      if (!updated)
      {
         int width = Math.max(1,Math.min(size.width,size.height));
         scale = zoom * SPHERE_RATIO * width/2;

         paintSphere();        // ---- paint
         paintTime();
         paintMemory();
         paintScore();
         paintFollow();

         updated = true;       // ---- now updated.
      }
      g.drawImage(offscreen,0,0,null);
   }


   private BufferedImage template = null;
   private Color template_space = null;
   private Color template_globe = null;
   private double template_zoom = 0;
   private int template_count = 0;
   private final int TEMPLATE_COUNT_CREATE = 10;

   private void paintSphere ()  // ---- paint Sphere --------------------------
   {
      Matrix base_ = Matrix.inv(base);

      // calculate bases

      Matrix[] bases = new Matrix[view.disks.length];
      for (int i=0; i<view.disks.length; ++i)
         bases[i] = Sphere.euler(view.disks[i].x).mul(base_);

      // calculate flags

      final Vector[] flags = new Vector[this.flags.length];
      for (int i=0; i<flags.length; ++i)
         flags[i] = Vector.mul(new Vector(this.flags[i]),base_);

      // calculate colors for score highlighting

      Color space_ = getParent().getBackground(); // base colors
      Color globe_ = getParent().getForeground();

      long difference = System.currentTimeMillis() - scorehighlighting_hitTimeMillis;
      double ratio = Math.exp(-SCOREHIGHLIGHTING_EXP_CONSTANT * difference);
      Color highlight = getColor(DiskAbsolute.BALL,scorehighlighting_t,0);
      Color space = Layout.fade(space_,highlight,ratio); // highlighted colors
      Color globe = Layout.fade(globe_,highlight,ratio);

      // count or discard

      if (space_.equals(template_space) &&
          globe_.equals(template_globe) &&
          zoom == template_zoom)
      {
         template_count ++;
      }
      else
      {
         template = null;
         template_space = space_;
         template_globe = globe_;
         template_zoom = zoom;
         template_count = 0;
      }

      // draw space & globe

      if (template != null &&             // use template if possible
          space.equals(template_space) && // (gains 1/4 of performance speed!)
          globe.equals(template_globe))
      {
         graphics.drawImage(template,-template.getWidth()/2,-template.getHeight()/2,template.getWidth(),template.getHeight(),null);
      }
      else // draw directly
      {
         // space
         Rectangle rect = graphics.getClipBounds();
         graphics.setColor(space);
         graphics.fillRect(rect.x,rect.y,rect.width,rect.height);

         // globe
         int radius = (int)scale;
         graphics.setColor(globe);
         graphics.fillOval(-radius,-radius,2*radius,2*radius);

         if (space_.equals(space) && // and save as template
             globe_.equals(globe))   // (if no score highlighting...)
         { // then template == null
            if(template_count >= TEMPLATE_COUNT_CREATE)
            {                        // (...and COUNT_CREATE reached)
               template = new BufferedImage(offscreen.getWidth(),offscreen.getHeight(),BufferedImage.TYPE_INT_ARGB);
               template_space = space;
               template_globe = globe;
               template_zoom = zoom;
               template.getGraphics().drawImage(offscreen,0,0,offscreen.getWidth(),offscreen.getHeight(),null);
            }
         }
      }

      // draw back-side

      for (int i=0; i<view.disks.length; ++i) {  // disks...
         Matrix base = bases[i];
         if (base.c.z > 0) continue;
         DiskAbsolute body = view.disks[i];
         Color color = Layout.fade(globe,getColor(body.what,body.t,body.n),py2ratio(body.py));

         if (body.what == body.UNIT)
            drawUnit(sphere,base,body.r,color,globe);
         else if (body.what == body.BALL)
            drawBall(sphere,base,body.r,color,globe);
      }

      for (int i=0; i<flags.length; ++i) {       // flags... (paint mirror-pair)
         double r = 1.4; // [!] 0.7;
         if (flags[i].z < 0) {
             Point2D p = Layout.scale2(scale,flags[i]);
             Ellipse2D P = new Ellipse2D.Double(p.getX()-r,p.getY()-r,2*r,2*r);
             graphics.setColor(Layout.color(Color.white,globe, flags[i].z));
             graphics.fill(P);
         } else {
             Point2D q = Layout.scale2(scale,flags[i].neg());
             Ellipse2D Q = new Ellipse2D.Double(q.getX()-r,q.getY()-r,2*r,2*r);
             graphics.setColor(Layout.color(Color.white,globe,-flags[i].z));
             graphics.fill(Q);
         }
      }

      paintDebugs(globe,false);                  // debugs...

      // draw front-side

      for (int i=0; i<flags.length; ++i) {       // flags... (paint mirror-pair)
          double r = 1.4; // [!] 0.7;
          if (flags[i].z > 0) {
              Point2D p = Layout.scale2(scale,flags[i]);
              Ellipse2D P = new Ellipse2D.Double(p.getX()-r,p.getY()-r,2*r,2*r);
              graphics.setColor(Layout.color(Color.white,globe, flags[i].z));
              graphics.fill(P);
          } else {
              Point2D q = Layout.scale2(scale,flags[i].neg());
              Ellipse2D Q = new Ellipse2D.Double(q.getX()-r,q.getY()-r,2*r,2*r);
              graphics.setColor(Layout.color(Color.white,globe,-flags[i].z));
              graphics.fill(Q);
          }
      }

      for (int i=0; i<view.disks.length; ++i) {  // disks...
         Matrix base = bases[i];
         if (base.c.z < 0) continue;
         DiskAbsolute body = view.disks[i];
         Color color = Layout.fade(globe,getColor(body.what,body.t,body.n),py2ratio(body.py));

         if (body.what == body.UNIT)
            drawUnit(sphere,base,body.r,color,globe);
         else if (body.what == body.BALL)
            drawBall(sphere,base,body.r,color,globe);
      }

      paintDebugs(globe,true);                   // debugs...
   }

   private static double py2ratio (double py)    // oder: 0.67 + 0.33 * Math.cos(body.py * Math.PI));
   {
	  if (py > 1) return 0;
	  else if (py > 0) return 1-py;
	  else return 1;
   }

   private int inset = 14;     // ---- Layout Constants for Time & Score ------
   private int padx = 14;
   private int pady = 6;

   private FontMetrics metrics = getFontMetrics(Layout.fontViewer);
   private int fontHeight = metrics.getAscent();
   private int fontWidth000 = metrics.stringWidth("000");

   private void paintTime ()  // ---- paint Time ------------------------------
   {
      Graphics2D graphics = (Graphics2D)(this.graphics).create();
      if (fullscreen) graphics.scale(fullscreen_scale,fullscreen_scale);

      String time = Layout.getTimeString(view.time);
      if (complete())  time += "     Complete";
      else if (paused) time += "     Paused";

      Rectangle rect = graphics.getClipBounds();
      graphics.setColor(Color.white);
      graphics.drawString(time,rect.x+inset,rect.y+inset+fontHeight);
   }


   private void paintMemory ()  // ---- paint Memory --------------------------
   {
      Graphics2D graphics = (Graphics2D)(this.graphics).create();
      graphics.translate(-getWidth(),0);
      if (fullscreen) graphics.scale(fullscreen_scale,fullscreen_scale);
      graphics.translate(+getWidth(),0);

      if (views.isEmpty()) return;

      Rectangle rect = graphics.getClipBounds();
      graphics.setColor(Color.white);

      for (int i=1; i<=MEMORY; ++i) {
         String memory = (views.get(""+i) == null)? "-" : ""+i;
         int width = metrics.stringWidth(memory);
         graphics.drawString(memory,rect.x+rect.width-inset-width-(MEMORY-i)*15,rect.y+inset+fontHeight);
      }
   }


   private void paintScore ()  // ---- paint Score ----------------------------
   {
      Graphics2D graphics = (Graphics2D)(this.graphics).create();
      graphics.translate(0,-getHeight());
      if (fullscreen) graphics.scale(fullscreen_scale,fullscreen_scale);
      graphics.translate(0,+getHeight());

      java.util.Vector sorted = new java.util.Vector();  // sort scores by t
      for (int i=0; i<view.teams.length; ++i) {
         hoverball.serial.Team team = view.teams[i];
         int j = sorted.size();
         while(j>0 && ((hoverball.serial.Team)sorted.elementAt(j-1)).t > team.t) --j;
         sorted.insertElementAt(team,j);
      }
      hoverball.serial.Team[] teams = new hoverball.serial.Team[sorted.size()];
      sorted.copyInto(teams);

      Rectangle rect = graphics.getClipBounds();
      int width = 0;  // paint color-panels and team names
      for (int i=0; i<teams.length; ++i) {
         int y = rect.y+rect.height-inset-(teams.length-1-i)*(fontHeight+pady);

         Channel channel = (Channel)channels.get(""+(teams[i].t)+" "+0);
         if (channel == null) channel = new Channel(teams[i].t,0,"???","???",0,null);

         Color color = getColor(DiskAbsolute.BALL,channel.t,0);
         if (color == null) color = Layout.colorGlobe;
         graphics.setColor(color);
         graphics.fillRect(rect.x+inset,y-fontHeight+2,20,fontHeight-2);
         graphics.setColor(Color.black);
         graphics.drawRect(rect.x+inset,y-fontHeight+2,20,fontHeight-2);
         graphics.setColor(Color.white);
         graphics.drawString(channel.team,rect.x+inset+20+padx,y);

         width = Math.max(width,metrics.stringWidth(channel.team));
      }

      graphics.setColor(Color.white);  // paint scores
      for (int i=0; i<teams.length; ++i) {
         int y = rect.y+rect.height-inset-(teams.length-1-i)*(fontHeight+pady);
         int xx = rect.x+inset+20+padx+width+padx+fontWidth000;

         String score = ""+teams[i].score;
         graphics.drawString(score,xx-metrics.stringWidth(score),y);

         java.util.Vector penalties = new java.util.Vector();  // sort penalties
         for (int j=0; j<view.disks.length; ++j)
            if (view.disks[j].what == DiskInfo.UNIT &&
                view.disks[j].t == teams[i].t &&
                view.disks[j].py > 0)
         {
            double py = view.disks[j].py;
            int k = penalties.size();
            while(k>0 && ((Double)penalties.elementAt(k-1)).doubleValue() > py) --k;
            penalties.insertElementAt(new Double(py),k);
         }
         Double[] pys = new Double[penalties.size()];
         penalties.copyInto(pys);

         for (int j=0; j<pys.length; ++j) {
            double py = Math.ceil(pys[j].doubleValue()); // aufrunden
            graphics.drawString(Layout.getTimeStringInteger(py),xx+(pys.length-j)*(fontWidth000+2*padx)-padx, y);
         }
      }
   }

   private void paintFollow ()  // ---- paint Follow --------------------------
   {
      Graphics2D graphics = (Graphics2D)(this.graphics).create();
      graphics.translate(-getWidth(),-getHeight());
      if (fullscreen) graphics.scale(fullscreen_scale,fullscreen_scale);
      graphics.translate(+getWidth(),+getHeight());

      Rectangle rect = graphics.getClipBounds();

      int y = rect.y+rect.height-inset;

      String zoom = Layout.getZoomString(this.zoom);  // ---- Zoom
      int widthz = metrics.stringWidth(zoom);

      graphics.setColor(Color.white);
      graphics.drawString(zoom,rect.x+rect.width-inset-widthz,y-pady-fontHeight);

      if (follow == null) return;                     // ---- Follow

      Color color = getColor(follow.what,follow.t,follow.n);
      if (color == null) color = Layout.colorGlobe;

      graphics.setColor(color);
      graphics.fillRect(rect.x+rect.width-20-inset,y-fontHeight+2,20,fontHeight-2);
      graphics.setColor(Color.black);
      graphics.drawRect(rect.x+rect.width-20-inset,y-fontHeight+2,20,fontHeight-2);

      String string = getName(follow.what,follow.t,follow.n);
      int width = metrics.stringWidth(string);

      graphics.setColor(Color.white);
      graphics.drawString(string,rect.x+rect.width-inset-20-padx-width,y);
   }


   // ---- Zoom ---------------------------------------------------------------

   public synchronized void zoom (double zoom)  // Accessed by Controller
   {
      if (zoom > ZOOM_MAX) zoom = ZOOM_MAX;
      if (zoom < ZOOM_MIN) zoom = ZOOM_MIN;
      this.zoom = zoom;
      update();
   }

   private void zooming (int zooming)  // Accessed by [+] and [-] keys
   {
      if (zooming != 0 && this.zooming != 0) return;
      this.zooming = zooming;
   }


   // ---- Follow Mode --------------------------------------------------------

   private void target (DiskAbsolute disk)
   {
      if (disk == null)
         follower.target(null,false);
      else {
         Matrix target = Sphere.euler(disk.x);
         boolean unit = (disk.what == disk.UNIT);
         if (unit) target = target.mulL(Matrix.rotY(FOLLOW_FRONT)); /*immutable*/
         follower.target(target,unit);
      }
   }

   // ---- Follow Default Unit

   Unit followDefaultUnit = null;  // Accessed by Controller
   Team followDefaultTeam = null;
   int  followDefaultBall = 0;

   public synchronized void followDefault ()
   {
      if (followDefaultUnit == null &&
          followDefaultTeam == null &&
          followDefaultBall == 0) {  // no default? leave!
         if (follow == null) return;
         followed = follow;
         follow = null; // target bleibt!

         if (listener != null) listener.listen(new Boolean(false));
      } else {                          // else: search channel und enter!
         if (!opened) return;  // will be called again.

         if (followDefaultTeam == null && followDefaultBall != 0) {  // follow = SHOT-BALL_n and do it!
            follow = new DiskAbsolute(DiskAbsolute.BALL,0,followDefaultBall,null,null,0,0,0,0,0,0,0,null);
         } else if (followDefaultTeam != null) {  // follow = search Channel... quit? TEAM-BALL_n?
            Channel found = null;
            for (Enumeration<Channel> e = channels.elements(); e.hasMoreElements(); ) {
               Channel channel = (Channel)e.nextElement();
               if (followDefaultTeam.owns(server,channel)) { found = channel; break; }
            }
            if (found == null) return;

            follow = new DiskAbsolute(DiskAbsolute.BALL,found.t,followDefaultBall,null,null,0,0,0,0,0,0,0,null);
         } else if (followDefaultUnit != null) {  // follow = search Channel... quit? do it?
            Channel found = null;
            for (Enumeration<Channel> e = channels.elements(); e.hasMoreElements(); ) {
               Channel channel = (Channel)e.nextElement();
               if (followDefaultUnit.owns(server,channel)) { found = channel; break; }
            }
            if (found == null) return;

            follow = new DiskAbsolute(DiskAbsolute.UNIT,found.t,found.n,null,null,0,0,0,0,0,0,0,null);
         }
         if (view != null) target(follow = getDisk(follow.what,follow.t,follow.n));  // even view? target!

         if (listener != null) listener.listen(new Boolean(true));  // follow_enter!
      }
   }

   // ---- Follow List Handling

   private DiskAbsolute followed = null;

   public synchronized void follow_enter ()
   {
      if (view == null) return;

      if (follow == null && followed != null) target(follow = getDisk(followed.what,followed.t,followed.n));
      if (follow == null && view.disks.length > 0) target(follow = new DiskAbsolute(view.disks[0]));
      followed = null;

      if (listener != null) listener.listen(new Boolean(follow != null));
   }

   public synchronized void follow_up ()
   {
      if (view == null) return;

      if (follow == null) follow_enter();
      else {
         java.util.Vector disks = getDisks();
         DiskAbsolute disk = getDisk(follow.what,follow.t,follow.n);
         int i = disks.indexOf(disk);
         int n = disks.size();
         target(follow = (DiskAbsolute)disks.elementAt((n+i-1)%n));
      }
      update();
   }

   public synchronized void follow_down ()
   {
      if (view == null) return;

      if (follow == null) follow_enter();
      else {
         java.util.Vector disks = getDisks();
         DiskAbsolute disk = getDisk(follow.what,follow.t,follow.n);
         int i = disks.indexOf(disk);
         int n = disks.size();
         target(follow = (DiskAbsolute)disks.elementAt((i+1)%n));
      }
      update();
   }

   public synchronized void follow_leave ()
   {
      if (view == null) return;

      if (follow == null) return;
      followed = follow;
      follow = null; // target bleibt!
      update();

      if (listener != null) listener.listen(new Boolean(false));
   }

   public synchronized void follow_home ()
   {
      if (view == null) return;
      DiskAbsolute ball = getDisk(DiskAbsolute.BALL,0,1);
      if (ball == null) return;

      if (follow == null) follow_enter();
      target(follow = ball);
   }


   // ---- class Follower -----------------------------------------------------

   private class Follower
      implements Runnable
   {
      public boolean suspend = false;

      private Thread runner = null;
      private Matrix target = null;
      private boolean rotate = false;

      // shares Screen.view;
      // shares Screen.globe;
      // shares Screen.frequency;
      // shares Screen.update();

      public void target (Matrix target, boolean rotate)
      {
         this.target = (target == null)? null : new Matrix(target);
         this.rotate = rotate;
      }

      public void start ()
      {
         if (runner != null) runner.interrupt();
//APPLET THREAD BUG//
         runner = new Thread(this,"Hoverball Follower"); //new Thread(Hovlet.group,this,"Follower");
         runner.start();
      }
      public void stop ()
      {
         if (runner == null) return;
         try { if (System.getSecurityManager() != null) System.getSecurityManager().checkAccess(runner);
            runner.interrupt();
         } catch(SecurityException e) {}  // nicht schlimm, nur sleep.
         runner = null;
      }
      public void run ()
      {
         Thread current = Thread.currentThread();
         while (current == runner) {
            try {
               Thread.sleep((int)Math.round(1000/frequency)); }
            catch(InterruptedException e) { break; }
            if (current != runner) break;  // Applet interruption;

            step(1/frequency);
            /*if (!suspend)*/ update();  // "if (!suspend)" nicht nötig. update() harmlos und wichtig für zoom.
         }
      }

      // ---- Follow Calculations

      public void step (double time)
      {
         // Zoom:

         double zoom_ratio = Math.exp(Math.log(2)/ZOOM_HALFWAY*time);

         double z = (zooming > 0)? zoom * zoom_ratio :
                    (zooming < 0)? zoom / zoom_ratio : zoom;
         if (z > ZOOM_MAX) z = ZOOM_MAX;
         if (z < ZOOM_MIN) z = ZOOM_MIN;
         zoom = z;

         // Follower:

         if (suspend || target == null) return;

         synchronized(screen)
         {
         double ratio = 1 - Math.exp(-Math.log(2)/FOLLOW_HALFWAY*time);

         // Calculate Axis Vector

         Vector axis = secure_axis(base.c,target.c,base.b);

         // Translation

         double alpha = angle(base.c,target.c,null);
         Matrix move = Matrix.rot(axis,alpha*ratio);
         base = secure_mul(base,move);

         // Rotation

         if (!rotate) return;

         double beta = angle(axis,target.a,target.c) - angle(axis,base.a,base.c);
         while (beta > Math.PI) beta -= 2*Math.PI;        // fit into -pi..+pi!
         while (beta <-Math.PI) beta += 2*Math.PI;
         Matrix turn = Matrix.rot(target.c,beta*ratio);
         base = secure_mul(base,turn);
         }
      }
   }


   // ---- Graphical Debugging ------------------------------------------------

   private void paintDebugs (Color globe, boolean front)  // ---- Debug Painting
   {
      Matrix base_ = Matrix.inv(base);

      Graphics2D graphics = (Graphics2D)this.graphics.create();
      graphics.setFont(Layout.fontDebug);

      /*
      graphics.setRenderingHint( // No Antialiasing for Debugs!
            RenderingHints.KEY_ANTIALIASING,
            RenderingHints.VALUE_ANTIALIAS_OFF);
      graphics.setRenderingHint(
            RenderingHints.KEY_TEXT_ANTIALIASING,
            RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
      */
      applyAntialiasing(graphics,antialiasing);  // [!] Debug Antialiasing by Menu, too ?

      for (Enumeration<Channel> e = debuglist.keys(); e.hasMoreElements(); ) {
         Channel channel = (Channel)e.nextElement();
         java.util.Vector debugs = (java.util.Vector)debuglist.get(channel);
         if (debugs.isEmpty()) return;

         // ---- Check: Unit Debugging?

         Boolean bool = null;  // Unit debugs?
         for (Enumeration<Unit> u = debugging.keys(); u.hasMoreElements(); ) {
            Unit unit = (Unit)u.nextElement();
            if (unit.owns(server,channel)) { bool = (Boolean)debugging.get(unit); break; }
         }
         boolean debugging = (bool != null && bool.booleanValue());

         // if (session empty) paint last session completely.
         // if (session open)  paint till last painted.

         // example: [ l=last, p=painted, o=orig ]
         // open session: O...P.P.O..  O..P.O...O..   empty session: O.P..O...O
         //               o     p l         p   l [o=p]                   l [p=l]

         // ---- Search Indexes & Setup Origin

         Origin origin = null;
         Matrix origin_base = null;

         int last = searchOrigin(debugs,debugs.size()-1);   // search last Origin [-1: ignore empty session!]
         int painted = last;
         if (!(debugs.lastElement() instanceof Origin)) {   // session open?
            painted = searchPainted(debugs,painted);        // search last painted (or Origin?)
            int orig = searchOrigin(debugs,painted+1);      // and it's Origin...  (itself?)

            Debug debug = (Debug)debugs.elementAt(orig);
            if (debug instanceof Origin) {                  // ...to be the first origin!
               origin = (Origin)debug;
               origin_base = Sphere.euler(origin.disk.x).mul(base_);
            }
         }

         // ---- Paint Debugs

         if (debugging)                                     // paints only if Unit debugs...
         for (int i=painted; i<debugs.size(); ++i) {        // start with index painted
            Debug debug = (Debug)debugs.elementAt(i);

            if (debug instanceof Origin) {
               origin = (Origin)debug;                            // set origin
               origin_base = Sphere.euler(origin.disk.x).mul(base_);
               if (view.arrival < origin.created) {               // and update if earlier!
                  DiskAbsolute disk = getDisk(DiskAbsolute.UNIT,origin.disk.t,origin.disk.n);
                  if (disk != null) origin.disk = new DiskAbsolute(disk);
               }
            } else if (debug instanceof Painted) ;                // ignore
            else if (origin != null) { // not null if player debugs correctly!
               paintDebug(debug,new Color(channel.color),globe,   // paint debugs recursively
                          graphics,front,scale,sphere,origin_base);
            }
         }

         // ---- Remove Debugs from List

         for (int i=0; i<last; ++i) debugs.removeElementAt(0);  // remove till last

         // ---- Add "Painted"-Debug

         if (!(debugs.lastElement() instanceof Origin) &&
             !(debugs.lastElement() instanceof Painted))  // session open?
            debugs.addElement(new Painted());
      }
   }

   private static void paintDebug (Debug debug, Color color, Color globe, // recursive paint
                                   Graphics2D g, boolean front, double scale, Sphere sphere, Matrix base)
   {
      if (debug.color != null) color = debug.color;  // get debug's color

      g.setColor(color);                             // paint debug
      debug.paint(g,color,globe,scale,sphere,base,front);

      if (debug.debugs == null) return;              // "final" debug?

      for (int i=0; i<debug.debugs.length; ++i)      // loop for sub-debugs
         paintDebug(debug.debugs[i],color,globe,g,front,scale,sphere,base);
   }

   private static int searchOrigin (java.util.Vector debugs, int start)
   {
      int index = start;
      while (index > 0) {
         -- index;
         Debug debug = (Debug)debugs.elementAt(index);
         if (debug instanceof Origin) break;
      }
      return index;
   }

   private static int searchPainted (java.util.Vector debugs, int start)
   {
      int index = start;
      while (index > 0) {
         -- index;
         Debug debug = (Debug)debugs.elementAt(index);
         if (debug instanceof Painted ||
             debug instanceof Origin) break;
      }
      return index;
   }

   private static class Origin extends Debug  // ---- Debug Handling ----------
   {
      public long created = 0;
      public DiskAbsolute disk = null;

      public Origin (DiskAbsolute disk, long created)
      {
         this.created = created;
         this.disk = disk;
      }
   }

   private static class Painted extends Debug {}


   synchronized void addDebug (int t, int n, Debug debug, Color color)
   {
      Channel channel = (Channel)channels.get(""+t+" "+n);
      if (channel == null || view == null) return;

      if (debug == null) {         // Setup new Debug Origin
         DiskAbsolute disk = getDisk(DiskAbsolute.UNIT,t,n);
         if (disk == null) return;
         addDebug(t,n,new Origin(new DiskAbsolute(disk),System.currentTimeMillis()),color);
      } else if (color != null) {  // Wrap Debug with Colored empty Debug
         Debug wrapper = new Debug(){};
         wrapper.color = (color == null)? Color.black : color;
         wrapper.debugs = new Debug[]{ debug };
         addDebug(t,n,wrapper,null);
      } else {                     // Add Debug
         java.util.Vector debugs = (java.util.Vector)debuglist.get(channel);
         if (debugs != null) {
            if (debug instanceof Origin && debugs.size() > 0 &&
                debugs.lastElement() instanceof Origin) debugs.removeElementAt(debugs.size()-1);
            if (debugs.size() < DEBUG_LIMIT) debugs.addElement(debug);
         } else {
            debugs = new java.util.Vector();
            debugs.addElement(debug);
            debuglist.put(channel,debugs);
         }
         if (paused || complete()) update();
      }
   }


   // ---- Keyboard Listeners ... [Controls] ----------------------------------

   private class ViewKeys extends KeyAdapter
   {
      private Screen screen;
      public ViewKeys (Screen screen) { this.screen = screen; }

      public void keyPressed (KeyEvent e) {
         synchronized(screen) {   // Follow
            if      (e.getKeyCode() == KeyEvent.VK_PAGE_UP)   follow_up();
            else if (e.getKeyCode() == KeyEvent.VK_PAGE_DOWN) follow_down();
            else if (e.getKeyCode() == KeyEvent.VK_END)       follow_leave();
            else if (e.getKeyCode() == KeyEvent.VK_HOME)      follow_home();
                                  // Zoom
            else if (e.getKeyCode() == KeyEvent.VK_ADD)       zooming(+1);
            else if (e.getKeyCode() == KeyEvent.VK_SUBTRACT)  zooming(-1);
                                  // Full Screen
            else if (e.getKeyCode() == KeyEvent.VK_ENTER)     screen.setFullScreen(!screen.fullscreen);
            else if (e.getKeyCode() == KeyEvent.VK_F11)       screen.setFullScreen(!screen.fullscreen);
            else if (e.getKeyCode() == KeyEvent.VK_ESCAPE)    screen.setFullScreen(false);
         }
      }

      public void keyReleased (KeyEvent e) {
         synchronized(screen) {   // Zoom
            if      (e.getKeyCode() == KeyEvent.VK_ADD)       zooming(0);
            else if (e.getKeyCode() == KeyEvent.VK_SUBTRACT)  zooming(0);
         }
      }
   }

   private class ControlKeys extends KeyAdapter
   {
      private Screen screen;
      public ControlKeys (Screen screen) { this.screen = screen; }

      public void keyPressed (KeyEvent e) {
         synchronized(screen) {   // Controls
        	if (false) ;
        	// Stelle Control-Keys ab für ein ungestörtes Human-Spiel... :-)
//        	else if (e.getKeyCode() == KeyEvent.VK_DELETE) control(1);
//          else if (e.getKeyCode() == KeyEvent.VK_INSERT) control(2);
//          else if (e.getKeyCode() == KeyEvent.VK_SPACE)  control(3);
//          else if (e.getKeyCode() == KeyEvent.VK_PAUSE) { if (opened) state(paused? 2 : 1); }
                                  // Memory
            else if (e.getKeyCode() >= KeyEvent.VK_1 && e.getKeyCode() <= (KeyEvent.VK_0+MEMORY)) {
               int key = e.getKeyCode()-KeyEvent.VK_0;
               if (e.isControlDown()) { if (view != null) views.put(""+key,view); }  // store
               else { if (listener != null) listener.listen(new Integer(key)); }   // send!
            }
// ############################################################################
//                        SAVE SCREENSHOTS (SHIFT + num)
// ############################################################################
/*
if (e.getKeyCode() >= KeyEvent.VK_1 && e.getKeyCode() <= (KeyEvent.VK_0+MEMORY) &&
    e.isShiftDown() && view != null) { // Screenshot!
      try {
         int key = e.getKeyCode()-KeyEvent.VK_0;
         String file = "screenshot-"+key+".gif";
         java.io.FileOutputStream fileoutput = new java.io.FileOutputStream(file);
         build.GIFOutputStream output = new build.GIFOutputStream(fileoutput);
         output.write(offscreen);  // no transparency
         output.flush();
         output.close();
         System.out.println(""+file+" saved.");
      } catch(Exception ee) { System.out.println(ee); }
}
*/
// ############################################################################
         }
      }

      private void control (int what)
      {
         int state = (!opened? 0 : paused? 1 : 2);

         if (what == 1) {
            if (state == 0) {}
            else if (state == 1) { state(0); }
            else if (state == 2) { state(1); state(0); }
         } else if (what == 2) {
            if (state == 0)      { state(1); }
            else if (state == 1) { state(0); state(1); }
            else if (state == 2) { state(1); state(0); state(1); }
         } else if (what == 3 && !complete()) {
            if (state == 0)      { state(1); state(2); }
            else if (state == 1) { state(2); }
            else if (state == 2) { state(1); }
         }
      }
      private void state (int state)
      {
         if (listener != null) listener.listen(new CS.State(state));
      }
   }


   // ---- Mouse Listeners ... [Mouse Scroller] -------------------------------

   private class MouseScroller
      implements MouseListener, MouseMotionListener
   {
      private Point mouse = null;

      public void mousePressed (MouseEvent e)
      {
         dragAndDrop.mousePressed(e); if (dragAndDrop.active()) return;       // Drag and Drop!
         synchronized(screen) {
            follower.suspend = true;  mouse = new Point(e.getX(),e.getY());
         }
      }

      public void mouseReleased (MouseEvent e)
      {
         if (dragAndDrop.active()) { dragAndDrop.mouseReleased(e); return; }  // Drag and Drop!
         if (mouse == null) return;

         synchronized(screen) {
            follower.suspend = false; mouse = null;
         }
      }

      public void mouseDragged (MouseEvent e)
      {
         if (dragAndDrop.active()) { dragAndDrop.mouseDragged(e); return; }  // Drag and Drop!
         if (mouse == null) return;

         synchronized(screen) {
            if (view == null) { follower.suspend = false; mouse = null; return; }

            int dx =   mouse.x - e.getX();  mouse.x = e.getX();
            int dy = -(mouse.y - e.getY()); mouse.y = e.getY();

            Matrix X = Matrix.rotX(Math.PI*2*dx/MOUSE_SENSITIVE);  // Rotation-Matrix
            Matrix Y = Matrix.rotY(Math.PI*2*dy/MOUSE_SENSITIVE);
            base = secure_mul(Matrix.mul(X,Y),base);

            if (follow == null) target(null);
            update();
         }
      }

      public void mouseClicked (MouseEvent e) {}
      public void mouseEntered (MouseEvent e) {}
      public void mouseExited (MouseEvent e) {}
      public void mouseMoved (MouseEvent e) {}
   }


   // ---- Mouse Listeners ... [Drag and Drop] ---------------------------------

   private class DragAndDrop
      implements MouseListener, MouseMotionListener
   {
      private final boolean enabled;
      private Point mouse = null;       // last mouse pointer (AWT)
      private DiskAbsolute disk = null; // moving disk
      private Matrix disk_base = null;  // disk's base at start
      private Vector anchor = null;     // mouse anchor at start
      private boolean moving = false;   // move or turn?

      public DragAndDrop (boolean enabled) { this.enabled = enabled; }

      public boolean active () { return (disk != null); }

      public void mousePressed (MouseEvent e)
      {
         if (!enabled) return;

         synchronized(screen) {
            if (view == null || !paused || complete()) return;

            this.mouse = new Point(e.getX(),e.getY());
            Vector mouse = where(this.mouse,0);
            if (mouse == null) return;                        // outside sphere.

            disk = null;
            for (int k=0; k<view.disks.length; ++k) {         // search object...
               DiskAbsolute d = (DiskAbsolute)view.disks[k];
               if (Vector.sub(mouse,Sphere.euler(d.x).c).abs() < d.r/sphere.rad) { disk = d; break; }
            }
            if (disk == null) { this.mouse = null; return; }  // clicked in pampa: return to Scroller.

            // now active!

            view.arrival = System.currentTimeMillis();        // mark view as "new" (Graphical Debugging)
            disk.v = new Vector();                            // disk.speed = 0

            disk_base = Sphere.euler(disk.x);
            anchor = new Vector(mouse);

            // check turn or move (clicked in black area?)

            Complex offset = new Complex(Vector.sub(mouse,disk_base.c).mul(Matrix.inv(disk_base)));

            moving = (disk.what == disk.BALL                // balls always move!
                   || offset.zero()                         // exactly center? move!
                   || Math.abs(offset.arg()) > Math.PI/6);  // not black area: move!

            screen.update();  // includes this.update()
         }
      }

      public void mouseDragged (MouseEvent e)
      {
         if (disk == null) return;

         this.mouse = new Point(e.getX(),e.getY());
         screen.update();  // includes this.update()
      }

      public void mouseReleased (MouseEvent e)
      {
         if (disk == null) return;

         release(true);
      }

      public void mouseClicked (MouseEvent e) {}
      public void mouseEntered (MouseEvent e) {}
      public void mouseExited (MouseEvent e) {}
      public void mouseMoved (MouseEvent e) {}

      public void update ()
      {
         if (disk == null) return;

         synchronized(screen) {
            if (moving) {  // move
               Vector mouse = where(this.mouse,-1);
               Vector axis = secure_axis(anchor,mouse,disk_base.b);
               double alpha = angle(anchor,mouse,null);
               disk.x = Sphere.euler(Matrix.mul(disk_base,Matrix.rot(axis,alpha)));
            } else {       // turn
               Vector mouse = where(this.mouse,+1);
               Vector axis = secure_axis(disk_base.c,mouse,disk_base.b);
               double alpha = angle(disk_base.b,axis,disk_base.c);
               disk.x = Sphere.euler(Matrix.mul(Matrix.rotZ(alpha),disk_base));
            }
         }
      }

      public void release (boolean send)
      {
         if (disk == null) return;

         synchronized(screen) {
            if (send && listener != null) listener.listen(new View(Double.NaN,new DiskAbsolute[]{disk},null));

            disk = null;
            disk_base = null;
            anchor = null;
            this.mouse = null;
         }
      }

      private Vector where (Point point, int extend) // 0: null outside, 1: rests at pi/2, -1: goes backside
      {
         Point P = new Point(point.x,point.y);
         P.x -= getSize().width/2;
         P.y -= getSize().height/2;

         Complex p = new Complex(-P.y,-P.x).mul(1/scale);
         if (p.abs2() > 1)
            if (extend == 0) return null;       // outside => null
            else if (extend > 0) p = p.norm();  // outside => norm  /*immutable*/

         Vector axis = new Vector(-p.y,p.x,0);
         Vector v = new Vector(0,0,1);
         if (axis.abs2() != 0) {
            double length = p.abs();
            while (length > 1) { length -= 2; v = new Vector(v.x,v.y,-v.z); } /*immutable*/ // v.z *= -1; }  // outside => modulo
            double alpha = Math.asin(length);
            v = v.mul(Matrix.rot(axis,alpha)); /*immutable*/
         }
         return v.mul(base);
      }
   }
}
