/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.layout.Constraints;
import hoverball.layout.Layout;  // Layout.center(...)
import hoverball.net.Address;    // PORT-Symbol

import java.awt.*;               // Default Size...
import java.awt.event.*;         // Window Closing...

import javax.swing.*;
import javax.swing.event.*;


// ---- Hoverball Session -----------------------------------------------------


/**
 * The class Session allows to run all required
 * Hoverball components simultaneously and to configure them
 * by little programming code.
 * <p>
 * The Hoverball Session assumes the following services:
 * <menu>
 * <li>Easy opening of Simulator and Controller,</li>
 * <li>Automatic connection of all hovlets on execution,</li>
 * <li>Useful window coordination,</li>
 * <li>Regard to the command line parameters
 * [[<i>host</i>]<tt>:</tt><i>port</i>],</li>
 * <li>Simple registration of units for Graphical Debugging by extended
 * <a href="#add(hoverball.Unit, boolean)">add(...)</a>-methods,</li>
 * <li>Automatic setting of the follow mode for units on showing their hovlets,</li>
 * <li>Termination by window closing.</li>
 * </menu>
 * Moreover, the Hoverball Session can be extended at will
 * by the full Hoverball command pool.
 * <p>
 * <b>Note:</b> The Session's hovlet tree must be displayed for a correct
 * working of the session!
 * <p>
 * <hr>
 * <p>
 * <b><u>Example</u>:</b>
 * This code creates a Hoverball Session which opens the team
 * <code>MyTeam</code>, sets the game duration to one minute and
 * runs the simulation.
 *
 *<pre>
 *import hoverball.*;
 *&nbsp;
 *public class MySession extends Session
 *{
 *   public static void main (String[] args) { new MySession(args); }
 *&nbsp;
 *   public MySession (String[] args)
 *   {
 *      super("My Session",args);                     // open Hoverball-Session
 *&nbsp;
 *      add(new MyTeam(),true);                       // open MyTeam with Graphical Debugging
 *&nbsp;
 *      controller.show();                            // show browser with Controller
 *      controller.viewer.show();                     // show screen
 *&nbsp;
 *      if (simulator == null) return;                // external Simulator? return!
 *      simulator.set("game.duration",60);            // set game duration to 1 minute
 *      simulator.state(1);                           // check in
 *      simulator.state(2);                           // start!
 *   }
 *}
 *</pre>
 *
 * <p>
 * <hr>
 * <p>
 * The class Session can also be run as an application:
 * <p>
 * <tt>java hoverball.Session </tt>[<i>host</i>][<tt>:</tt><i>port</i>][<tt>#</tt><i>hash</i>]
 * <p>
 * <menu>
 * <li>(without) - Opens a Simulator and a Controller.</li>
 * <li><tt>:</tt><i>port</i> - Opens a Simulator at the specified port and a Controller.</li>
 * <li><i>host</i><tt>:</tt><i>port</i><tt>#</tt><i>hash</i> - Opens a Controller and connects to the Simulator
 * at the specified address.</li>
 * </menu>
 */
public class Session extends Hovlet
{
   // ---- static Launcher ----------------------------------------------------

   /**
    * Runs a standard session (application!).
    */
   public static void main (String[] args)
   {
	  controller_frame = false; // :-)
	   
	  Session hoverball = new Session(null,args);
//      hoverball.add(new hoverball.team.FantasticThree(),true);
//      hoverball.add(new hoverball.team.TrioGaulois());
      hoverball.show();
        
      if (hoverball.simulator != null) {  // local simulation
         hoverball.controller.show();
         hoverball.controller.viewer.show();
//         hoverball.controller.state(1);
//         hoverball.controller.state(2);
      } else {                            // network connection!
         hoverball.controller.show();
         hoverball.controller.viewer.show();
//         hoverball.controller.state(1);
//         hoverball.controller.state(2);
      }
   }


   private class Banner extends JLabel  // ------------------------[Banner]----
   {
      Banner () { setPreferredSize(new Dimension(0,Hovlet.BANNER+Layout.pady)); }
   }

   private final JPanel main_;
   private static final int SUB_DISPLAY_MAX = 5;

   private class Sub_ extends JPanel  // ---------------------------[Sub_]----
   {
      final Hovlet sub;

      final Dimension icon = Team.iconsize;
      final int padx = Team.padx;
      final int tabx = Team.tabx;
      final int sepy = Team.sepy;

      public Sub_ (Hovlet sub)
      {
         this.sub = sub;

         final Hovlet sub_ = this.sub;
         addMouseListener(new MouseAdapter() {
            public void mouseEntered(MouseEvent e) {
               setBackground(Layout.colorBackgroundActive);
               setForeground(Layout.colorForegroundActive);
               setOpaque(true);
            }
            public void mouseExited(MouseEvent e) {
               setBackground(getParent().getParent().getBackground());
               setForeground(getParent().getParent().getForeground());
               setOpaque(false);
            }
            public void mouseClicked(MouseEvent e) { sub_.show(); }
         });

         setPreferredSize(new Dimension(0,sepy + icon.height-1 + sepy));
         setFont(Layout.fontMain);
         if(sub instanceof Controller)
            setFont(getFont().deriveFont(Font.BOLD));
         setOpaque(false);
      }

      public void paintComponent (Graphics g) { super.paintComponent(g);
         if (sub instanceof Controller)
            Layout.drawIcon(g,padx,sepy,icon.width,icon.height); // transparent
         else if (sub instanceof Team)
            Layout.drawIcon(g,padx,sepy,icon.width,icon.height,((Team)sub).color);
         else if (sub instanceof Unit)
            Layout.drawIcon(g,padx,sepy,icon.width,icon.height,new Color(((Unit)sub).connect.color));

         int y = sepy+icon.height-1;
         g.drawString(sub.title,padx+tabx,y);
      }
   }


   // ---- Hoverball Session --------------------------------------------------

   /**
    * Executed Simulator
    */
   public final Simulator simulator;

   /**
    * Executed Controller
    */
   public final Controller controller;

   /**
    * Required browser size
    */
   public Dimension size = null;


   private static final String NAME = "Hoverball";
   private static final String SERVER = "" + Address.PORT + Simulator.DEFAULT_PORT;

   private static final int SPACING_X = Layout.SPACING_X;
   private static final int SPACING_Y = Layout.SPACING_Y;

   private String server_;

   private final boolean sf;   // simulator_frame
   private final boolean cf;   // controller_frame

   static boolean simulator_frame = true;  // default values of former parameter concept 
   static boolean controller_frame = true; //
   //             controller_frame = false; if hoverball.Session is called via "static void main(String[] args)" :-)
	  

   // ---- Constructors -------------------------------------------------------


   /**
    * Opens a Hoverball session.
    */
   public Session ()
   {
      this(null);
   }

   /**
    * Opens a Hoverball session.
    *
    * @param name name of the Hoverball session
    */
   public Session (String name)
   {
      this(name,SERVER);
   }

   /**
    * Opens a Hoverball session.
    * <p>
    * You can specify the address of an external Simulator
    * (as a string in the form of <code>"<i>host</i>:<i>port</i>#<i>hash</i>"</code>)
    * or the required port
    * (as a string in the form of <code>":<i>port</i>"</code>).
    * The handoff of a string array is also valid so that the
    * argument array of a main(...)-method can directly be passed.
    * In this case the string with index 0 is read out.
    *
    * @param name name of the Hoverball session
    * @param server address or port of the Simulator
    */
   public Session (String name, Object server)
   {
      // ---- Constructor

      super((name == null)? NAME : name,false,true);
      Application.setTitle(title);
      if (Application.getBounds() == null) Application.setBounds(Layout.getApplicationBounds());

      main_ = new JPanel(new GridBagLayout()) {
         public Dimension getPreferredSize() {
            Dimension size = new Dimension();
            for (int i=0; i<getComponentCount(); ++i)
               size.height += getComponent(i).getPreferredSize().height;
            return size;
         }
      };
      main_.add(new Banner(),
                new Constraints(0,0,1,1,1.0,0.0,null));
      main_.setOpaque(false); // makes main's banner visible!

      main.setLayout(new GridBagLayout());
      main.add(main_,
               new Constraints(0,0,1,1,1.0,1.0,null,Constraints.NORTH,Constraints.HORIZONTAL));

      // ---- Setup Server Address (parse Server Argument)

      if (server instanceof String) {
         server_ = (String)server;
      } else if (server instanceof String[]) {
         String[] args = (String[])server;
         String arg = hoverball.layout.About.arg(args);
         server_ = (args.length == 0)? SERVER : arg;
      } else if (server == null) {
         server_ = null;
      } else {
         server_ = null;
      }

      // ---- Setup Simulator & Controller

      this.sf = simulator_frame;
      this.cf = controller_frame;

      simulator = (server_ != null && server_.startsWith(""+hoverball.net.Address.PORT)?
         new Simulator(server_,sf) :
         null);
      if (simulator != null)
         server_ = simulator.address;

      controller = new Controller();
      if (!cf) add(controller);
   }


   // -------------------------------------------------------------------------


   /**
    * [Implementation]
    */
   public void browse (JInternalFrame browser)
   {
      Rectangle desktop = Application.desktop.getBounds();

      int sh = 0;
      int ch = 0;

      // ---- Simulator

      if (simulator != null) {
         simulator.locate(SPACING_X,SPACING_Y);
//[!]         if (sf) simulator.frame.show();
         if (sf) sh = Layout.SIMULATOR_HEIGHT + SPACING_Y;

         if (simulator.frame != null) simulator.frame.setVisible(true); // SHOW SIMULATUR
      }

      // ---- Controller

      int screen_width = Layout.getApplicationBounds().width - Layout.BROWSER_WIDTH-3*SPACING_X;
      int screen_height = Layout.getApplicationBounds().height - 5*SPACING_Y; // why 5?
      Rectangle screen_area =
         new Rectangle(Layout.BROWSER_WIDTH+2*SPACING_X,SPACING_Y,
                       screen_width,screen_height);
//      
//    controller.viewer.setBounds(screen_area);
//
// "UGLY BUG of Swing" (?):
//
// Falls controller.fullscreen(true) VOR controller.show() ausgeführt wird,
// wird die Screen-Size über controller.viewer.setSize(...) (in der Methode browse())
// mit verändert, OBWOHL die JComponent screen NICHT mehr im JContainer viewer ist!
//
      controller.viewer.setLocation(screen_area.x, screen_area.y);
      controller.viewer_setSize(screen_area.width, screen_area.height); // UGLY BUG of Swing!
//
      
      if (cf) {
          controller.locate(SPACING_X,SPACING_Y+sh);
          ch = Layout.CONTROLLER_HEIGHT + SPACING_Y;
      }
    	  
      
      // ---- Session

      Dimension size = (this.size != null)? new Dimension(this.size) : browser.getSize();
      Rectangle browser_area = new Rectangle(SPACING_X,SPACING_Y+sh+ch,size.width,size.height);

      browser.addInternalFrameListener(new InternalFrameAdapter() {
         public void internalFrameClosed(InternalFrameEvent e) { Application.exit(); }
      });

      
SwingUtilities.invokeLater(new Runnable() {  // !! WORKAROUND: JAVA 1.8 NEEDS THIS
   public void run () {                      // !!      
      browser.setBounds(browser_area);
      browser.show();
      
      if (cf) {
    	  controller.show();
    	  controller.frame.setClosable(false);
      }
   }                                         // !!
});                                          // !!

      // ---- Connection

      connect(server_);                // Connect Hovlets to Server...
      if (controller.parent_ == null)
         controller.connect(server_);  // extra-connect() if standalone!
      
   }

   /**
    * Adds a hovlet to this session.
    *
    * @param hovlet hovlet
    * @see <a href="Hovlet.html#add(hoverball.Hovlet)">Hovlet.add(...)</a>
    */
   public boolean add (Hovlet hovlet)
   {
      boolean success = super.add(hovlet);

      if (success && hovlets.length <= SUB_DISPLAY_MAX)
         main_.add(new Sub_(hovlet), // add to Hovlet's Layout
                   new Constraints(0,2+hovlets.length,1,1,1.0,0.0,null));

      return success;
   }
   
   /**
    * Adds a unit to the session.
    *
    * @param unit unit
    */
   public boolean add (Unit unit)
   {
      follow(unit);
      return this.add((Hovlet)unit);
   }

   /**
    * Adds all units of a team to the session.
    *
    * @param team team
    */
   public boolean add (Team team)
   {
      for (int i=0; i<team.hovlets.length; ++i) {
         follow(team);
         if (team.hovlets[i] instanceof Unit)
            follow((Unit)team.hovlets[i]);
      }
      return this.add((Hovlet)team);
   }

   /**
    * Adds a unit to the session and registers it at the Controller
    * for Graphical Debugging.
    *
    * @param unit unit
    * @param debug graphical output enabled?
    */
   public boolean add (Unit unit, boolean debug)
   {
      controller.debug(unit,debug);
      follow(unit);
      return this.add((Hovlet)unit);
   }

   /**
    * Adds all units of a team to the session and registers them at the Controller
    * for Graphical Debugging.
    *
    * @param team team
    * @param debug graphical output enabled?
    */
   public boolean add (Team team, boolean debug)
   {
      controller.debug(team,debug);
      for (int i=0; i<team.hovlets.length; ++i) {
         follow(team);
         if (team.hovlets[i] instanceof Unit)
            follow((Unit)team.hovlets[i]);
      }
      return this.add((Hovlet)team);
   }

   /**
    * Comfortable method to start the session. 
    * This method executes:
    * 
    *<pre>
    *   controller.show();               
    *   controller.viewer.show();        
    *   simulator.state(1);
    *   simulator.state(2);              
    *</pre>
    */
   public void start ()
   {
      start(null);
   }

   /**
    * Comfortable method to start the session. 
    * This method executes:
    * 
    *<pre>
    *   unit.show();                     
    *   controller.viewer.show();        
    *   simulator.state(1);
    *   simulator.state(2);
    *</pre>
    *
    * @param unit unit to show and follow
    */
   public void start (Unit unit)
   {
      if (unit != null) unit.show();
      else controller.show();
      
SwingUtilities.invokeLater(new Runnable() {  // !! SONST KEIN FOCUS FÜR HUMAN CONTROLS
   public void run () {                      // !!
      controller.viewer.show();
   }                                         // !!
});                                          // !!
      
      if (simulator == null) return;
      simulator.state(1);
      simulator.state(2);
   }

   
   // ---- FollowListener -----------------------------------------------------

   private void follow (Unit unit)
   {
      unit.icon.addComponentListener(new FollowListener(unit));
   }

   private void follow (Team team)
   {
      team.icon.addComponentListener(new FollowListener(team));
   }

   private class FollowListener extends ComponentAdapter  // erase "//"-lines to disable turn-off
   {
      private Unit unit = null;
      private Team team = null;
      private boolean following = false; //

      public FollowListener (Unit unit) { this.unit = unit; }
      public FollowListener (Team team) { this.team = team; }

      public void componentShown (ComponentEvent e)
      {
         following = true; //
         if (unit != null) controller.follow(unit);
         if (team != null) controller.follow(team,1);
      }
      public void componentHidden (ComponentEvent e)
      {
         if (!following) return; //
         controller.follow(null,1); //
         following = false; //
      }
   }
}
