/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import java.net.InetAddress;

import javax.swing.*;

import hoverball.net.Address;
import hoverball.simulator.Constants;


// ---- Hoverball Simulator ---------------------------------- [Simulator] ----


/**
 * The simulator is the core piece of the Hoverball program.
 * It creates a virtual Hoverball world in which connected players
 * can move - according to the laws of this world.
 * <p>
 * As the Simulator is entirely directed through the network,
 * this class only offers a few elementary methods for its configuration.
 * <p>
 * Please refer to the <b>Hoverball Manual</b> for a detailed description of the Simulator.
 * <p>
 * <hr>
 * <p>
 * The class Simulator can also be run as an application:
 * <p>
 * <tt>java hoverball.Simulator </tt>[<tt>:</tt><i>port</i>]
 * <p>
 * <menu>
 * <li>(ohne) - Opens a Simulator at the default port.</li>
 * <li><tt>:</tt><i>port</i> - Opens a Simulator at the specified port.</li>
 * </menu>
 */
public final class Simulator
{
   static Simulator headless_simulator = null; // wird bei Application.exit() extra geschlossen :-)  (AWEFUL BAD WORKAROUND)
	   
   // ---- static Launcher ----------------------------------------------------

   private static boolean VISIBLE = true;  // can globally be turned off for testing purposes

   /**
    * Runs the Simulator (application!).
    */
   public static void main (String[] args)
   {
	  VISIBLE = ! java.awt.GraphicsEnvironment.isHeadless();

	  if (VISIBLE) {
		  Application.setTitle("Hoverball");
		  Application.setBounds();
	  }

      String arg = hoverball.layout.About.arg(args);
      Simulator simulator = new Simulator((args.length == 0)? null : arg,VISIBLE);

      Runtime.getRuntime().addShutdownHook(new Thread("Hoverball Bye Hook") {
         public void run() {
            simulator.close();
            if (!VISIBLE) System.out.println("\nBye.");
         }
      });

      if (!VISIBLE) { // headless
          String address = simulator.address;
          String canonical = ""; try { canonical = InetAddress.getLocalHost().getCanonicalHostName(); } catch (java.net.UnknownHostException e) {}
          String address_ = address.replace(new Address(address).getHostName(),canonical);

          System.out.println("Hoverball Simulator at '" + address_  + "' at " + address + "");
          System.out.println("Version "+hoverball.layout.About.VERSION_COMPLETE);
          System.out.println();
          System.out.println("Press 'Ctrl-C' to quit.");
      }
      
      if (simulator.frame != null) simulator.frame.setVisible(true); // SHOW SIMULATUR
   }


   // ---- class Simulator ----------------------------------------------------

   hoverball.simulator.Simulator simulator = null; // access by Server

   /**
    * Default port of the Simulator
    */
   public static final int DEFAULT_PORT = hoverball.layout.About.DEFAULT_PORT;


   /**
    * Network address of the Simulator
    */
   public final String address;

   /**
    * Frame of the Simulator
    */
/*private*/ final JInternalFrame frame;


   /**
    * Opens a Simulator.
    */
   public Simulator () { this(-1,true); }

   /**
    * Opens a Simulator at the specified port.
    *
    * @param port port
    */
   public Simulator (int port) { this(port,true); }

   /**
    * Opens a Simulator visibly or invisibly ("headless").
    *
    * @param visible visible?
    */
   public Simulator (boolean visible) { this(-1,visible); }

   /**
    * Opens a Simulator at the specified port visibly or invisibly.
    *
    * @param port port
    * @param visible visible?
    */
   public Simulator (int port, boolean visible)
   {
      this(""+port,visible);
      if (frame != null) frame.setVisible(visible);
   }

   /*
    * package-private contructor for "my" main()-methods -- does not 'show' simulator.frame
    * to locate it correctly before showing it... :-)
    */
   Simulator (String port, boolean visible)  // parse ":nnnn"-Strings (not public)
   {
	  int port_ = parse(port);
	  if (port_ < 0)
         simulator = new hoverball.simulator.Simulator(visible);
      else
         simulator = new hoverball.simulator.Simulator(visible,port_);

      address = simulator.address;
      frame = simulator.frame;
      if (frame == null) headless_simulator = this;
   }

   private static int parse (String port)
   {
      if (port != null && port.startsWith(""+hoverball.net.Address.PORT)) try {
         return Integer.parseInt(port.substring(1));
      } catch(NumberFormatException e) {
         return -1;
      } else if (port != null) try {
         return Integer.parseInt(port);
      } catch(NumberFormatException e) {
         return -1;
      } else
         return -1;
   }


   // ---- Simulator Properties (accessed by Server) --------------------------

   int getNumberOfChannels () { return simulator.getNumberOfChannels(); }
   int getNumberOfControls () { return simulator.getNumberOfControls(); }
   int getState()             { return simulator.getState(); }
   double getTime()           { return simulator.getTime(); }
   double getDuration()       { return simulator.getDuration(); }
   double getHertz()          { return simulator.getHertz(); }


   // ----

   /**
    * Sets a Simulator variable.
    * <p>
    * If <code>value</code> equals <b><code>null</code></b> the variable is deleted.
    *
    * @param key Simulator variable
    * @param value value
    */
   public void set (String key, String value)
   {
      simulator.set(key,value);
   }

   /**
    * Sets a Simulator variable.
    * <p>
    * The <code>double</code> value is converted into a <code>string</code>.
    *
    * @param key Simulator variable
    * @param value value
    */
   public void set (String key, double value)
   {
      simulator.set(key,(value==Math.round(value))? ""+(int)value : ""+value);
   }

   /**
    * Returns the value of a Simulator variable.
    *
    * @param key Simulator variable
    * @return the value
    */
   public String get (String key)
   {
      return simulator.get(key);
   }

   /**
    * Returns the value of a Hoverball option.
    * <p>
    * If <code>key</code> is a Simulator variable but no option, it
    * returns 0.
    *
    * @param key Hoverball option
    * @return the value
    */
   public double option (String key)
   {
      return simulator.option(key);
   }

   /**
    * Changes the Simulator's state.
    *
    * @param state new state
    */
   public void state (int state)
   {
      simulator.state(state);
   }

   /**
    * Closes the Simulator.
    */
   public void close ()
   {
      simulator.close();
   }

   /**
    * Relocates the Simulator window.
    *
    * @param x <i>x</i>-coordinate of the window position
    * @param y <i>y</i>-coordinate of the window position
    */
   public void locate (int x, int y)
   {
      if (frame != null) frame.setLocation(x,y);
   }
   
   // ----

   /**
    * Returns the Simulator state.
    *
    * @return state
    */
   public int state () 
   { 
	  return simulator.state(); 
   }
   
   /**
    * Returns the game time in the Simulator.
    *
    * @return time
    */
   public double time () 
   { 
	  return simulator.time(); 
   }
   
   /**
    * Returns if the game has finished ("Complete").
    *
    * @return complete
    */
   public boolean complete () 
   { 
	  return simulator.complete(); 
   }
   
}
