/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.layout.*;
import hoverball.net.Address;    // owns
import hoverball.serial.Channel; // owns

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


// ---- Team ------------------------------------------------------ [Team] ----


/**
 * The class Team allows to pool Java units together in a team.
 * <p>
 * For this, we use the interlacing hovlet structure:
 * A team is a trivial hovlet that players have been added to as
 * sub-hovlets.
 * With a simple call of the team method connect(...) all players can connect
 * simultaneously to the Simulator.
 * <p>
 * Beside this pooling function the class Team has three other attributes
 * <i>Version</i>, <i>Author</i> and <i>Comment</i> by means of which its
 * implementation can be characterized on the predefined hovlet display.
 * <p>
 * <hr>
 * <p>
 * <b><u>Example of a team</u>:</b>
 *
 *<pre>
 *import hoverball.*;
 *&nbsp;
 *public class MyTeam extends Team
 *{
 *   public MyTeam ()
 *   {
 *      super("My Team");
 *&nbsp;
 *      add(new MyUnit("One",0xFF8800));
 *      add(new MyUnit("Two",0xFF4400));
 *      add(new MyUnit("Three",0xFF0000));
 *   }
 *}
 *</pre>
 */
public class Team extends Hovlet
{
   // ---- Hovlet Design... ---------------------------------------------------

   static final Dimension iconsize = new Dimension(20,11);
   static final int padx = Layout.padx;
   static final int tabx = Layout.tabx;
   static final int sepy = 4;
   private static final int UNIT_DISPLAY_MAX = 5;

   private class Banner extends JLabel  // ------------------------[Banner]----
   {
      Banner () { setPreferredSize(new Dimension(0,Hovlet.BANNER+Layout.pady)); }
   }

   private class Team_ extends JPanel  // --------------------------[Team_]----
   {
      final String VERSION = "Version:";
      final String AUTHOR  = "Author:";
      final String COMMENT = "Comment:";

      final boolean something;

      final Dimension icon = Team.iconsize;
      final int padx = Team.padx;
      final int tabx = Team.tabx;
      final int tab2;
      final int pady = 16;
      final int taby = 16;
      final int sepy = Team.sepy/2;
      final int tabysuppl = 8;

      public Team_ ()
      {
         something = (version != null || author != null || comment != null);

         FontMetrics mb = main.getFontMetrics(Layout.fontMainBold);
         FontMetrics mp = main.getFontMetrics(Layout.fontMain);

         int h = 0; int w = 0;

         h += Team.sepy+icon.height-1;
         if (version != null) { h += taby; w = mb.stringWidth(VERSION); }
         if (author  != null) { h += taby; w = mb.stringWidth(AUTHOR); }
         if (comment != null) { h += taby+(Layout.lines(comment).length-1)*mp.getHeight();
                                w = mb.stringWidth(COMMENT); }
         if (something) h += tabysuppl - taby;  // ein taby zu viel
         h += Team.sepy;

         tab2 = w+14;
         setPreferredSize(new Dimension(0,h));
         setOpaque(false);
      }

      public void paintComponent (Graphics g) { super.paintComponent(g);
         Layout.drawIcon(g,padx,Team.sepy,icon.width,icon.height,color);

         Font FB = Layout.fontMainBold;
         Font FP = Layout.fontMain;

         int y = Team.sepy+icon.height-1;
         if (!something) { g.setFont(FP); g.drawString(title,padx+tabx,y); }
         else {
            if (version != null) { g.setFont(FB); g.drawString(VERSION,padx+tabx,y);
                                   g.setFont(FP); g.drawString(version,padx+tabx+tab2,y); y += taby; }
            if (author  != null) { g.setFont(FB); g.drawString(AUTHOR ,padx+tabx,y);
                                   g.setFont(FP); g.drawString(author ,padx+tabx+tab2,y); y += taby; }
            if (comment != null) { g.setFont(FB); g.drawString(COMMENT,padx+tabx,y);
                                   g.setFont(FP); Layout.drawText(g,comment,padx+tabx+tab2,y); y += taby; }
         }
      }
   }

   private class Unit_ extends JPanel  // --------------------------[Unit_]----
   {
      final Unit unit;

      final Dimension icon = Team.iconsize;
      final int padx = Team.padx;
      final int tabx = Team.tabx;
      final int sepy = Team.sepy;

      public Unit_ (Unit unit)
      {
         this.unit = unit;

         final Unit unit_ = this.unit;
         addMouseListener(new MouseAdapter() {
            public void mouseEntered(MouseEvent e) {
               setBackground(Layout.colorBackgroundActive);
               setForeground(Layout.colorForegroundActive);
               setOpaque(true);
            }
            public void mouseExited(MouseEvent e) {
               setBackground(getParent().getBackground());
               setForeground(getParent().getForeground());
               setOpaque(false);
            }
            public void mouseClicked(MouseEvent e) { unit_.show(); }
         });

         setPreferredSize(new Dimension(0,sepy + icon.height-1 + sepy));
         setFont(Layout.fontMain);
         setOpaque(false);
      }

      public void paintComponent (Graphics g) { super.paintComponent(g);
         Layout.drawIcon(g,padx+tabx,sepy,icon.width,icon.height,new Color(unit.connect.color));

         int y = sepy+icon.height-1;
         g.drawString(unit.title,padx+tabx+tabx,y);
      }
   }


   // ---- class Team ---------------------------------------------------------


   /**
    * Team version
    */
   public final String version;

   /**
    * Team author
    */
   public final String author;

   /**
    * Team comment
    */
   public final String comment;


   Color color = null;

   private final JPanel icon_;
   private final JPanel main_;


   /**
    * Opens a team.
    *
    * @param name team name
    */
   public Team (String name)
   {
      this(name,null,null,null);
   }

   /**
    * Opens a team.
    * <p>
    * The display of the attributes <i>Version</i>, <i>Author</i>
    * and <i>Comment</i> can be disabled by the value <b><code>null</code></b>.
    *
    * @param name team name
    * @param version version
    * @param author author
    * @param comment comment
    */
   public Team (String name, String version, String author, String comment)
   {
      super(name,true,true);

      this.version = version;
      this.author = author;
      this.comment = comment;

      icon.add(icon_ = new JPanel() {
         public void paintComponent (Graphics g) { Layout.drawIcon(g,null,getSize(),color); }
      },"Center");
      if(Layout.TOOLTIPS) icon_.setToolTipText("Debug "+name);
      icon_.addMouseListener(new MouseAdapter() {
         public void mouseClicked (MouseEvent e) {
            boolean total = true;
            for (int i=0; i<hovlets.length; ++i) if (hovlets[i] instanceof Unit) {
               Unit unit = (Unit)hovlets[i];
               Boolean state = unit.getDebugState();
               if (state == null) continue;
               total &= state.booleanValue();
            }
            for (int i=0; i<hovlets.length; ++i) if (hovlets[i] instanceof Unit) {
               Unit unit = (Unit)hovlets[i];
               Boolean state = unit.getDebugState();
               if (state == null) continue;
               unit.setDebugState(!total);
            }
         }
      });

      main_ = new JPanel(new GridBagLayout()) {
         public Dimension getPreferredSize() {
            Dimension size = new Dimension();
            for (int i=0; i<getComponentCount(); ++i)
               size.height += getComponent(i).getPreferredSize().height;
            return size;
         }
      };
      main_.add(new Banner(),
                new Constraints(0,0,1,1,1.0,0.0,null));
      main_.add(new Team_(),
                new Constraints(0,1,1,1,1.0,0.0,null));
      main_.setOpaque(false); // makes main's banner visible!

      main.setLayout(new GridBagLayout());
      main.add(main_,
               new Constraints(0,0,1,1,1.0,1.0,null,Constraints.NORTH,Constraints.HORIZONTAL));
   }

   /**
    * Adds a hovlet to the team.
    * <p>
    * If the hovlet is a unit, its team name is set to the title of
    * the team hovlet.
    *
    * @param hovlet hovlet
    * @return <b><code>true</code></b>, if connection succeeded
    */
   public boolean add (Hovlet hovlet)
   {
      boolean result = super.add(hovlet);

      if (!(hovlet instanceof Unit)) return result;

      Unit unit = (Unit)hovlet;      // link team name
      unit.connect.team = ""+title;

      int n = 0;                     // update team color
      int r = 0; int g = 0; int b = 0;
      for (int i=0; i<hovlets_.length; ++i) if (hovlets_[i] instanceof Unit) {
         Unit u = (Unit)hovlets_[i];
         r += new Color(u.connect.color).getRed();
         g += new Color(u.connect.color).getGreen();
         b += new Color(u.connect.color).getBlue();
         ++ n;
      }
      if (n > 0) color = new Color(r/n,g/n,b/n);

      if (n <= UNIT_DISPLAY_MAX)
         main_.add(new Unit_(unit),     // add to Hovlet's Layout
                   new Constraints(0,2+hovlets.length,1,1,1.0,0.0,null));

      return result;
   }


   // ---- Owns... (FollowDefaultTeam!)

   boolean owns (Address server, Channel channel)  // ex. Unit owns?
   {
      for (int i=0; i<hovlets_.length; ++i) {
         if (!(hovlets_[i] instanceof Unit)) continue;
         Unit unit = (Unit)hovlets_[i];
         if (unit.owns(server,channel)) return true;
      }
      return false;
   }
}

