/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.simulator.Constants;
import hoverball.debug.*;

import hoverball.net.*;
import hoverball.serial.*;
import hoverball.layout.*;
import hoverball.math.*;

import java.awt.*;
import java.awt.event.*;
import java.io.IOException;

import javax.swing.*;
import java.util.Hashtable;


// ---- Unit ------------------------------------------------------ [Unit] ----


/**
 * The class Unit, based on the class Hovlet, represents the base class
 * for all Java units. The Unit connects as Unit Client to the Simulator
 * and facilitates the creation of your own players
 * by offering a great deal of integrated functions.
 * <p>
 * After the definition of the constructor, where you can specify team and player
 * name, you only have to implement the thinking loop
 * <a href="#loop()">loop()</a>.
 * Please consider that
 *
 * <menu>
 * <li>You should exit the looping method when the game has been interrupted,
 *     for it is called up again when the game continues.
 *     This can be achieved by the correct handling of the method
 *     <a href="#look()">look()</a>
 *     waiting for the next "look" of the Simulator.
 *     If the game has been interrupted, look() returns the value <b><code>false</code></b>,
 *     else <b><code>true</code></b>. This means that the looping structure
 *<pre>
 *public void loop()
 *{
 *   while(look())
 *   {
 *      ...
 *   }
 *}
 *</pre>
 *     answers the purpose.
 *     <p></li>
 * <li>Before looping all check in parameters are known,
 *     and inside the loop - after calling look() -
 *     the current object positions
 *     which can be accessed in the array
 *     <a href="#pucks">pucks</a>.
 *     <p></li>
 * <li>The action should be sent at the end of the loop by
 *     <a href="#action(double, double, double)">action(...)</a>.
 *     <p></li>
 * <li>Debug objects can always be set by
 *     <a href="#debug(hoverball.debug.Debug)">debug(...)</a>.
 *     <p></li>
 * </menu>
 *
 * The main panel of a unit hovlet is unused. You can use it to display
 * your own analyses.
 * <p>
 * <hr>
 * <p>
 * <b><u>Example of a trivial Java unit</u>:</b>
 *
 *<pre>
 *import hoverball.*;
 *import hoverball.math.*;
 *&nbsp;
 *public class Clumsy extends Unit
 *{
 *   public Clumsy ()
 *   {
 *      super(null,"Clumsy",0x8888FF);  // name "Clumsy", color light blue
 *   }
 *&nbsp;
 *   public void loop ()
 *   {
 *      Sphere sphere = new Sphere(option("world.radius"));  // read out parameters...
 *      double Qmax = option("unit.charge.max");
 *      double Fmax = option("unit.engine.max");
 *&nbsp;
 *      while (look())
 *      {
 *         Puck ball = puck(BALL,0,1);       // that's the ball
 *         Puck goal = puck(BALL,self.t,1);  // that's the goal
 *&nbsp;
 *         if (ball == null) action(0,-Fmax,Fmax);     // does not see ball? turn!
 *         else {                                      // does see ball:
 *            Vector a = (goal != null)? Vector.vec(goal.X.c,ball.X.c) :  // calculate
 *                                       Vector.vec(ball.X.c,self.X.c);   // axis
 *            Matrix X = Matrix.mul(ball.X,Matrix.rot(a,(self.r+ball.r)/sphere.rad));
 *            Complex x = sphere.warp(X.c);            // X is shot position
 *&nbsp;
 *            double l = -x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // turn and
 *            double r =  x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // go to X
 *            double q = (x.abs() < ball.r)? 0.5 : 0;  // if at X: shoot!
 *&nbsp;
 *            action(q*Qmax,l*Fmax,r*Fmax);
 *         }
 *      }
 *   }
 *}
 *</pre>
 */
public abstract class Unit extends Hovlet
{
   // ---- Puck.WHATs ---------------------------------------------------------

   /**
    * Type constant: node
    * &nbsp;(cf class <a href="Puck.html#NODE">Puck</a>)
    */
   protected static final int NODE = Puck.NODE;

   /**
    * Type constant: ball
    * &nbsp;(cf class <a href="Puck.html#BALL">Puck</a>)
    */
   protected static final int BALL = Puck.BALL;

   /**
    * Type constant: unit
    * &nbsp;(cf class <a href="Puck.html#UNIT">Puck</a>)
    */
   protected static final int UNIT = Puck.UNIT;


   // ---- returned by (connect...) -------------------------------------------


   /**
    * Team number of the channel
    */
   protected int channel_t = 0;

   /**
    * Player number of the channel
    */
   protected int channel_n = 0;


   // ---- returned by (checkin...) -------------------------------------------

   private final Hashtable options = new Hashtable();

   private double _simulator_time = 0;
   private double _unit_radius = 0;
   private double _unit_mass   = 0;
   private double _ball_radius = 0;
   private double _ball_mass   = 0;


   // ---- returned by (look...) ----------------------------------------------

   // eigentlich 'protected'" -- nur wegen Jython (!) muss alles 'public' sein!
   
   /**
    * Date of the last look
    */
   public double time = 0;          // -> current.time

   /**
    * Energy of the last look
    */
   public double energy = 0;

   /**
    * Penalty of the last look
    */
   public double penalty = 0;

   /**
    * Array of objects of the last look
    * which were seen or heard
    */
   public Puck[] pucks = null;      // -> current.pucks

   /**
    * Pointer to itself inside the object array of the last look
    */
   public Puck self = null;         // -> pucks[i]


   // ---- Thread Group

//APPLET THREAD BUG//
   static ThreadGroup group = Thread.currentThread().getThreadGroup(); //new ThreadGroup("units");
//   static { group.setMaxPriority(Thread.NORM_PRIORITY); }

   // ----

   private NetAgent client = null;
   private Address host = null;        // Remember host (Graphical Debugging!)

   US.Connect connect = null;          // Stores (connect...)-Infos

   private int t_ = 0;
   private int n_ = 0;

   private final JPanel icon_;

   private static final int LOOK_EVERY = 500;

   // ---- Constructors -------------------------------------------------------

   static final int hashColor (String name)
   {
      return (-name.hashCode()) & 0xFFFFFF;
   }

   /**
    * Opens a unit.
    *
    * @param name unit name
    */
   public Unit (String name)
   {
      this(name,name);
   }

   /**
    * Opens a unit.
    *
    * @param team team name
    * @param name unit name
    */
   public Unit (String team, String name)
   {
      this(team,name,hashColor(name));
   }

   /**
    * Opens a unit.
    *
    * @param name unit name
    * @param color color
    */
   public Unit (String name, int color)
   {
      this(name,name,color);
   }

   /**
    * Opens a unit.
    *
    * @param team team name
    * @param name unit name
    * @param color color
    */
   public Unit (String team, String name, int color)
   {
      super(name,true,false);

      if (team == null) team = ""+name;

      connect = new US.Connect("",team,name,color);  // #hash will be overwritten in 'connecting'

      icon.add(icon_ = new JPanel() {
         public void paintComponent (Graphics g) { Layout.drawIcon(g,null,getSize(),new Color(connect.color)); }
      },"Center");
      if(Layout.TOOLTIPS) icon_.setToolTipText("Debug "+name);
      icon_.addMouseListener(new MouseAdapter() {  // install Debug-Listener on icon
         public void mouseClicked (MouseEvent e) {
            Boolean state = getDebugState();
            if (state == null) return;
            setDebugState(!state.booleanValue());
         }
      });
   }



   // ---- Search Pucks -------------------------------------------------------

   /**
    * Returns the puck with the specified identity within the puck array.
    *
    * @param what object type
    * @param t team number
    * @param n puck number
    * @return the puck (or <b><code>null</code></b>, if not found)
    */
   protected Puck puck (int what, int t, int n)
   {
      if (pucks == null) return null;
      for (int i=0; i<pucks.length; ++i) if (pucks[i].equals(what,t,n)) return pucks[i];
      return null;
   }

   /**
    * Returns the puck with the specified identity within the puck array.
    *
    * @param id Identity string
    * @return the puck (or <b><code>null</code></b>, if not found)
    */
   protected Puck puck (String id)
   {
      if (pucks == null) return null;
      for (int i=0; i<pucks.length; ++i) if (pucks[i].equals(id)) return pucks[i];
      return null;
   }


   // ---- Options ------------------------------------------------------------

   /**
    * Returns the value of a Simulator option that has been set on check in.
    * <p>
    * Not every Simulator option is sent to the units.
    *
    * @param key Simulator option
    * @return the value
    */
   protected double option (String key)
   {
      try {
         String value = (String)options.get(key);
         return Double.valueOf(value).doubleValue();
      } catch(Exception e) {  // NullPointerException / NumberFormatException
         return Double.NaN;
      }
   }


   // ---- Loop Handling... ---------------------------------------------------

   private static class TimeLook extends SU.Look
   {
      double time = 0;

      TimeLook (double time, SU.Look look) { super(look.E,look.py,look.disks); this.time = time; }
   }

   private final Object LOOP = new Object();  // loop monitor (only one Thread loops!)

   private class Loop extends Thread
   {
      long base = 0;
      TimeLook current  = null;
      TimeLook received = null;

//APPLET THREAD BUG//
//      public Loop (String name) { super(Unit.group,name); }
      public Loop (String name) { super("Hoverball Unit ["+name+"]"); }

      public void run () { synchronized(LOOP) {
////System.out.println("LOOP STARTS in "+this_.title);
    	  loop();
////System.out.println("LOOP ENDS in "+this_.title);
      } }  // only one Thread loops!
   }

   private Loop loop = null;                  // Active Loop Object


   // ---- Network Listener... ------------------------------------------------

   private synchronized void look_ (long currentTimeMillis, SU.Look look)
   {
      if (client != null) client.send(new US.Look());  // reply (look) immediately!

      if (loop == null) {  // first look after break
         loop = new Loop(connect.name);
         //loop.setPriority(Thread.MIN_PRIORITY);
         loop.base = currentTimeMillis;
         loop.received = new TimeLook(0,look);
         loop.start();
      } else synchronized(loop) {
         double time = ((double)(currentTimeMillis-loop.base)) / 1000;
         loop.received = new TimeLook(time,look);  // look.next will be ignored.
         loop.notify();
      }
   }

   private synchronized void break_ ()
   {
      if (loop == null) return;  // double-(break)
      synchronized(loop) {
         loop.received = null;   // null = interrupted!
         loop.notify();
      }
      loop = null;
   }


   // ---- Unit Commands ------------------------------------------------------

   /**
    * Requests a new look.
    * <p>
    * This method should be called in the unit's thinking loop.
    * It updates the variables <code>time</code>, <code>pucks</code> and
    * <code>self</code>.
    * <p>
    * The thread is stopped until the unit gets a new look from
    * the Simulator. If the game is interrupted the method returns
    * <b><code>false</code></b> and the loop should be quit.
    *
    * @return <b><code>true</code></b>, if the game continues
    */
   protected boolean look ()
   {
      debug(new Debug(){});  // add null-Debug to ensure a non-empty Session
      debug();               // Close current Debug Session

      int t_ = this.t_;  // save t_
      int n_ = this.n_;  // save n_  ("disconnect" may disturb!)

      Thread.yield();

      Loop loop = (Loop)Thread.currentThread();  // local loop = real running Loop

      try {
         synchronized(loop) {
            if (loop.received != null) while (true) { // loop.received == null: (break)
            	loop.wait(LOOK_EVERY);                           // will be notified
         		if (loop.current != loop.received) break;        // new look received? quit.
            	if (client != null) client.send(new US.Look());  // else: resend (look) and continue waiting!
////System.out.println("(look) missing! "+name()+" resends (look) after "+LOOK_EVERY+" millis.");
         	}
            loop.current = loop.received;
         }
      } catch(InterruptedException e) { return false; }  // should not happen (Security exception!)
      if (loop.current == null) { return false; }

      Thread.yield();

      debug();               // Setup new Debug Session

      Puck[] pucks = new Puck[loop.current.disks.length];
      for (int l=0; l<loop.current.disks.length; ++l) {
         DiskRelative i =(DiskRelative)loop.current.disks[l];
         pucks[l] = new Puck(i.what,i.t,i.n,                   // (what t n)
            (i.x==null)? null : Sphere.euler(i.x),             // X
            (i.what == BALL)? _ball_radius : (i.what == UNIT)? _unit_radius : 0,   // r
            (i.what == BALL)? _ball_mass   : (i.what == UNIT)? _unit_mass   : 1,   // m
            (i.message==null)? null : ""+i.message             // message
         );
      }

      this.time  = loop.current.time * _simulator_time;
      this.energy = loop.current.E;
      this.penalty = loop.current.py;
      this.pucks = pucks;
      this.self = null;
      for (int i=0; i<pucks.length; ++i)
         if (pucks[i].what == UNIT &&
             pucks[i].t == t_ &&
             pucks[i].n == n_) { this.self = pucks[i]; break; }

      if (this.self == null) return false;
      return true;
   }

   /**
    * Sends an action.
    * <p>
    * If a parameter of the action is to not be modified it can be replaced by the value
    * <code>Double.NaN</code>.
    *
    * @param Q Polarization
    * @param F_L engine power left
    * @param F_R engine power right
    */
   protected void action (double Q, double F_L, double F_R)
   {
      action (Q,F_L,F_R,null);
   }

   /**
    * Sends a message.
    *
    * @param message message
    */
   protected void action (String message)
   {
      action (Double.NaN,Double.NaN,Double.NaN,message);
   }

   /**
    * Sends an action and a message.
    * <p>
    * If a parameter of the action is to not be modified it can be replaced by the value
    * <code>Double.NaN</code>.
    *
    * @param Q Polarization
    * @param F_L engine power left
    * @param F_R engine power right
    * @param message message
    */
   protected void action (double Q, double F_L, double F_R, String message)
   {
      if (client == null) return;
      client.send(new US.Action(Q,F_L,F_R,message));
   }

   /**
    * Empty for the implementation of the thinking loop.
    */
   protected abstract void loop ();


   // ---- Debugging...

   private void debug ()                          // Setup new Debug Session!
   {
      if (client == null) return;

      if (debugger != null) debugger.addDebug(host,t_,n_,null,null);
   }

   /**
    * Sets a debug object.
    *
    * @param debug debug object
    */
   protected void debug (Debug debug)             // Add to Debug List
   {
      if (client == null) return;
      if (debug == null) return;

      if (debugger != null) debugger.addDebug(host,t_,n_,debug,null);
   }

   /**
    * Sets a debug object with the specified color.
    *
    * @param debug debug object
    * @param color color
    */
   protected void debug (Debug debug, int color)  // Add to Debug List with specified color
   {
      if (client == null) return;
      if (debug == null) return;

      if (debugger != null) debugger.addDebug(host,t_,n_,debug,new Color(color));
   }

   protected static String format (double d)
   {
      if (Complex.format == null) return ""+d;
      else return Complex.format.format(d);
   }

   protected static String format (Complex c) { return c.toString(); }
   protected static String format (Vector v)  { return v.toString(); }
   protected static String format (Matrix M)  { return M.toString(); }


   // ---- Setup Debuggers ----------------------------------------------------

   Controller debugger = null;  // Setup by Controller

   String name () { return ""+connect.name; }
   boolean owns (Address server, Channel channel)
   {
      return server.equals(host) &&                    // same Server
             channel != null && connect != null &&     // connected
             (t_ == channel.t && n_ == channel.n);     // same Channel
   }

   Boolean getDebugState ()  // by Team
   {
      if (debugger == null) return null;
      else return debugger.getDebugState(this);
   }

   void setDebugState (boolean debug)  // by Team
   {
      if (debugger == null) return;
      else debugger.debug(this,debug);
   }

   // ---- Connect & Disconnect -----------------------------------------------

   java.util.Timer pinger = null;

   /**
    * [Implementation]
    */
   protected synchronized final boolean connecting (String server)
   {
      Address address = new Address(server);
      if (!address.isValid()) return false;

      host = address;  // Remember Host before connecting! (Follow & Graphical Debugging)
      boolean success = false;
      try
      {
         client = new UnitClient(host);  // must be defined for first received serials
         connect.hash = host.getHash();
    	   success = client.connect(host,connect);
         // host's #hash will be changed externally by 'client.connect'
      }
      catch (IOException e) {}

      if (success) {
         servertext = ""+host; // ggf. mit zugewiesenem #hash vom Server
         Application.servertext_with_hash = servertext;
      }

      if (success) { // install Pinger
    	   pinger = new java.util.Timer("Hoverball Ping (Unit)");
    	   pinger.schedule(new java.util.TimerTask() {
    	      public void run () { client.send(new US.Ping()); }
    	   },0,Socket.PING_EVERY);
      }

      if (success) {
         return true;
      } else {
         if (client!=null)
         {
            client.close();
            client = null;
         }
         host = null;
         return false;
      }
   }

   /**
    * [Implementation]
    */
   protected synchronized final void disconnecting ()
   {
      break_();

      pinger.cancel(); // cancel Pinger
      pinger = null;

      client.close();
      client = null;
      host = null;
      channel_t = t_ = 0;
      channel_n = n_ = 0;
   }


   // ---- Network Protocol: Unit's Client to Server -------- [UnitClient] ----

   private final Unit this_ = this;

   private class UnitClient extends NetAgent
   {
      // ---- Bye on close() --------------------------------------------------

      private boolean bye = true;  // send bye on close()
      private boolean closed = false;

      public UnitClient(Address serverAddress) throws IOException
      {
    	  super(serverAddress);
      }

      public synchronized void close ()
      {
         if (closed) return;
         if (bye) this.send(new US.Bye());
         super.close();
         closed = true;
      }

      // ---- Receive ---------------------------------------------------------

      public void receive (Address sender, String message)
      {
//     	 System.out.println("Unit "+this+" got "+message);
         Serial serials = SU.deserialize(message);

         Serial serial_ = serials;
         while (serial_ != null) synchronized(this_) {
            Serial serial = serial_;
            serial_ = serial_.next;
            serial.next = null;
            if (client == null) continue;  // disconnected? ignore!

            // ---- (connect 3 2 ...) -----------------------------------------

            if (serial instanceof SU.Connect)
            {
               host.setHash(((SU.Connect)serial).hash);
               SU.Connect connect = (SU.Connect)serial;
               channel_t = t_ = connect.t;
               channel_n = n_ = connect.n;
            }

            // ---- (checkin...) ----------------------------------------------

            if (serial instanceof SU.Checkin)
            {
               SU.Checkin checkin = (SU.Checkin)serial;

               options.clear(); int k=0;
               for (int i=0; i<Constants.CONSTANTS; ++i) if (Constants.isForUnit(i))
                  options.put(Constants.getKey(i),checkin.options[k++]);

               _simulator_time = option(Constants.getKey(Constants.SIMULATOR_TIME));
               _unit_radius    = option(Constants.getKey(Constants.UNIT_RADIUS));
               _unit_mass      = option(Constants.getKey(Constants.UNIT_MASS));
               _ball_radius    = option(Constants.getKey(Constants.BALL_RADIUS));
               _ball_mass      = option(Constants.getKey(Constants.BALL_MASS));
            }

            // ---- (look 2.3 19 (ball ..) (unit ..)) --------------------------

            else if (serial instanceof SU.Look)
            {
               look_(System.currentTimeMillis(),(SU.Look)serial);
            }

            // ---- (break) ---------------------------------------------------

            else if (serial instanceof SU.Break)
            {
               break_();
            }

            // ---- (bye) -----------------------------------------------------

            else if (serial instanceof SU.Bye)
            {
               bye = false;  // don't send bye yourself!
               disconnect();
            }
         }
      }
   }
}
