/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.serial.*;
import hoverball.layout.*;
import hoverball.net.Address;

import java.awt.event.*;
import javax.swing.*;


// ---- Viewer - Hoverball Screen Wrapper ----------------------- [Viewer] ----


class Viewer
{
   public final JInternalFrame frame;
   public final Screen screen;

   private final int DEFAULT_WIDTH  = 512;
   private final int SUPPLEMENT_HEIGHT = 30;

   private boolean visible = false;    // viewer is visible?

   private Address server = null;
   public Channel[] channels = null;  // null: checked out, else: checked in.
   private double duration = 0;
   private double frequency = 0;
   private double sphere = 0;
   private View view = null;
   private boolean paused = false;     // true after checkin

   // ---- Constructors -------------------------------------------------------

   public Viewer (Screen screen)
   {
      this.screen = screen;
      frame = new MyInternalFrame(Layout.getTitle(screen.getTitle()),true,true,true,false);
      frame.setBackground(Layout.colorSpace);  // Default Sphere Colors...
      frame.setForeground(Layout.colorGlobe);
      frame.setSize(DEFAULT_WIDTH,DEFAULT_WIDTH+SUPPLEMENT_HEIGHT);
      frame.setDefaultCloseOperation(JInternalFrame.HIDE_ON_CLOSE);
      frame.addComponentListener(new ComponentAdapter() {
         public void componentShown  (ComponentEvent e) { show(); }
         public void componentHidden (ComponentEvent e) { hide(); }
      });

      frame.setContentPane(screen);
      screen.setBackground(Layout.colorBackground);  // Closed Screen.

      //if (Applet.applet != null) visible = true;     // Applet: pretend showing screen!
   }

   // ---- Commands

   public synchronized boolean visible () { return visible; }

   private synchronized void show ()
   {
SwingUtilities.invokeLater(new Runnable() {  // !! SONST KEIN FOCUS FÜR HUMAN CONTROLS
   public void run () {                      // !!
      screen.requestFocus();
   }                                         // !!
});                                          // !!

      if (visible) return;
      visible = true;

      if (channels == null) return;    // not checked in? out!
      screen.open(server,channels,frequency,duration,sphere);  // else...
      if (view == null) return;        // even view?
      screen.view(view);
      if (paused) return;              // even running??
      screen.pause(paused);
      
   }

   private synchronized void hide ()
   {
      if (!visible) return;
      visible = false;

      if (channels == null) return;    // not checked in? out!
      screen.close();                  // else...
   }

   public synchronized void checkin (Address server, Channel[] channels,
                                     double frequency, double duration, double sphere)
   {
      if (frame.isVisible() && !visible) show();  // prevent ekel-error! (no componentShown-event thrown)

      if (channels != null) checkout();

      this.server = new Address(server);
      this.channels = channels;
      this.frequency = frequency;
      this.duration = duration;
      this.sphere = sphere;
      this.paused = true;

      if (visible) screen.open(server,channels,frequency,duration,sphere);
   }

   public synchronized void checkout ()
   {
      if (frame.isVisible() && !visible) show();  // prevent ekel-error! (no componentShown-event thrown)
      if (channels == null) return;

      channels = null;
      duration = 0;
      view = null;
      paused = false;

      if (visible) screen.close();
   }

   public synchronized void remove (int t, int n)
   {
      if (frame.isVisible() && !visible) show();  // prevent ekel-error! (no componentShown-event thrown)
      if (channels == null) return;

      for (int i=0; i<channels.length; ++i)
         if (channels[i] != null &&
             channels[i].t == t && channels[i].n == n) channels[i] = null;

      if (visible) screen.remove(t,n);
   }

   public synchronized void view (View view)
   {
      if (frame.isVisible() && !visible) show();  // prevent ekel-error! (no componentShown-event thrown)
      if (channels == null) return;

      this.view = view;

      if (visible) screen.view(view);
   }

   public synchronized void pause (boolean paused)
   {
      if (frame.isVisible() && !visible) show();  // prevent ekel-error! (no componentShown-event thrown)
      if (channels == null) return;

      this.paused = paused;

      if (visible) screen.pause(paused);
   }


   // ---- finally!

   public synchronized void close ()
   {
////System.out.println("# Viewer closing: checkout...");
      checkout();
////System.out.println("# Viewer closing: frame.dispose...");
      frame.dispose();
////System.out.println("# Viewer closed.");
   }
}
