/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.layout;

import java.awt.*;
import java.awt.geom.*;

import javax.swing.*;

import hoverball.math.Matrix;
import hoverball.math.Vector;

import java.net.*;


// ---- static Layout ------------------------------------------- [Layout] ----


public class Layout
{
   // ---- Global 'private' properties of the Hoverball Application
	
   public static boolean auto_show = true;
   
   // ---- GUI 
   
   public static final javax.swing.plaf.metal.MetalTheme theme = 
      new javax.swing.plaf.metal.DefaultMetalTheme();

   public static final boolean TOOLTIPS = true; // Activate Tooltips
   
   // ---- Fonts

   public static final String SANSSERIF  = "Dialog";      //"Lucida Sans";
   public static final String MONOSPACED = "DialogInput"; //"Lucida Sans Typewriter";

   public static final Font fontControl = theme.getControlTextFont();
   public static final Font fontUser    = theme.getUserTextFont();

   public static final Font fontSymbol  = new Font(MONOSPACED,Font.ITALIC,fontUser.getSize());
   public static final Font fontFixed   = new Font(MONOSPACED,Font.PLAIN,fontUser.getSize());

   public static final Font fontPlain   = fontUser.deriveFont(Font.PLAIN);
   public static final Font fontBold    = fontUser.deriveFont(Font.BOLD);
   public static final Font fontItalic  = fontUser.deriveFont(Font.ITALIC);
   public static final Font fontSmall   = fontUser.deriveFont(Font.PLAIN,fontUser.getSize2D()-1);

   public static final Font fontMenu    = fontSmall;
   public static final Font fontButton  = fontControl;
   public static final Font fontCombo   = fontPlain;
   public static final Font fontViewer  = new Font (fontControl.getName(),fontControl.getStyle(),18);
   public static final Font fontDebug   = fontSmall;

   public static final Font fontBanner    = new Font(SANSSERIF,Font.BOLD+Font.ITALIC,2*fontSmall.getSize());
   public static final Font fontMain      = fontSmall;   // Hoverball's default font on Main-Panel
   public static final Font fontMainBold  = fontSmall.deriveFont(Font.BOLD);
   public static final Font fontHoverball = new Font(SANSSERIF,Font.BOLD+Font.ITALIC,3*fontSmall.getSize());
   public static final Font fontVersion   = fontSmall;   // About Window
   public static final Font fontAbout     = fontSmall;   // About Window

   // ---- Colors

   public static final Color colorBackground = theme.getControl();
   public static final Color colorForeground = theme.getControlTextColor();
   public static final Color colorBackgroundActive = theme.getPrimaryControlDarkShadow();
   public static final Color colorForegroundActive = theme.getPrimaryControlHighlight();
   // [replaced by getBackground().darker()] public static final Color colorGrayLabel  = theme.getControlShadow();
   public static final Color colorDesktop    = theme.getWindowTitleBackground();  // JDesktop's color
   public static final Color colorAboutBanner= theme.getWindowBackground();  // About Window
   public static final Color colorSpace      = new Color(0x18183C); // 0x101030
   public static final Color colorGlobe      = new Color(0x2E365A); // 0x202840

   public static int hashColor (String string)
   {
      switch(string.toLowerCase()) {
         case "white"  : return Color.white.getRGB();
         case "gray"   : return Color.gray.getRGB();
         case "black"  : return Color.black.getRGB();
         case "red"    : return Color.red.getRGB();
         case "blue"   : return Color.blue.getRGB();
         case "green"  : return Color.green.getRGB();
         case "yellow" : return Color.yellow.getRGB();
         case "magenta": return Color.magenta.getRGB();
         case "cyan"   : return Color.cyan.getRGB();
         case "orange" : return Color.orange.getRGB();
         case "pink"   : return Color.pink.getRGB();
         
         default: return (-string.hashCode()) & 0xFFFFFF;
      }
   }

   public static String firstUpperCase (String string)
   {
      if (string.length() == 0) return "";
      else return string.substring(0,1).toUpperCase() + string.substring(1);
   }
   
   // ---- Load Resources
   
   private static Image load (String file, Component who)
   {
	  URL url = ClassLoader.getSystemResource(file);  // in Groovy that doesn't work yet
	  if (url == null) return null;                   // and returns null... :-(
      Image image = Toolkit.getDefaultToolkit().getImage(url);
      if (image == null) return null;
      
      MediaTracker tracker = new MediaTracker(who);
      tracker.addImage(image,0);
      try {
         tracker.waitForID(0);
      } catch (InterruptedException e) {}
      
      return image;
   }
   
   public static Image icon_png = null;
   public static Image logo_png = null;
   
   public static void loadResources (Component who)
   {
//      if (who instanceof hoverball.Applet) return;
      
      icon_png = load("resources/icon.png",who);
      logo_png = load("resources/logo.png",who);
   }
   
   // ---- Bounds & Sizes

   public final static int BROWSER_WIDTH = 400;
   public final static int BROWSER_HEIGHT = 342;
   public final static int SIMULATOR_HEIGHT = 180;
   public final static int CONTROLLER_HEIGHT = 120;
   public final static int SPACING_X = 24;  // Session
   public final static int SPACING_Y = 16;  // Session

   private final static Insets APPLICATION_INSETS = new Insets(50,40,70,40);
   public static Rectangle getApplicationBounds ()
   {
      Dimension desktop = Toolkit.getDefaultToolkit().getScreenSize();  // locate!
      
      int height = desktop.height - APPLICATION_INSETS.top  - APPLICATION_INSETS.bottom;
      int width = height + BROWSER_WIDTH + 2*SPACING_X;
      
      Dimension application = new Dimension(width,height);
      
      return new Rectangle((desktop.width - width)/2,
                           APPLICATION_INSETS.top,
                           application.width,
                           application.height);
   }

   public static Rectangle center (Dimension area, Dimension size)
   {
      return center(new Rectangle(area),size);
   }

   public static Rectangle center (Rectangle area, Dimension size)
   {
      Point location = new Point(area.x + (area.width  - size.width) / 2,
                                 area.y + (area.height - size.height)/ 2);
      return new Rectangle(location,size);
   }

   public static Rectangle fit (Rectangle area, Rectangle rect)
   {
      if (rect.width  > area.width)  rect.width  = area.width;   // intersect size
      if (rect.height > area.height) rect.height = area.height;

      if (rect.x < area.x) rect.x = area.x;                      // move if out of area
      if (rect.y < area.y) rect.y = area.y;
      if (rect.x + rect.width  > area.x + area.width)  rect.x = area.x + area.width - rect.width;
      if (rect.y + rect.height > area.y + area.height) rect.y = area.y + area.height - rect.height;

      return rect;
   }

   // ---- ListCellRenderer

   public static JComponent finishListCell (JComponent component, JList list, boolean isSelected)
   {
      if (isSelected) {
         component.setBackground(theme.getPrimaryControl()); // component.setBackground(list.getSelectionBackground());
         component.setForeground(theme.getUserTextColor());  // component.setForeground(list.getSelectionForeground());
      } else {
         component.setBackground(list.getBackground());
         component.setForeground(list.getForeground());
      }
      component.setEnabled(list.isEnabled());
      return component;
   }

   // ---- drawBanner

   public static final int BANNER = 40;

   public static final int padx = 12;
   public static final int tabx = 30;
   public static final int pady = 12;

   public static void drawBanner (Graphics g, Dimension size,
                                  String title, String clazz, boolean line)
   {
      Rectangle rect = new Rectangle(0,0,size.width,size.height);
      final int pady =  8;

      Graphics2D g2 = (Graphics2D)g.create();

      g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

      g2.setFont(Layout.fontBanner);
      g2.drawString(title,rect.x+padx,BANNER-pady);

      g2.setFont(Layout.fontItalic);
      int width = g2.getFontMetrics(g2.getFont()).stringWidth(clazz);
      g2.drawString(clazz,rect.x+rect.width-padx-width,BANNER-pady);

      if (line) g2.drawLine(rect.y+padx,BANNER-1,rect.x+rect.width-padx,BANNER-1);
   }

   // ---- drawIcon

   public static void drawIcon (Graphics g, Point point, Dimension size, Color color)
   {
      if (point == null) point = new Point();
      if (size  == null) size = new Dimension();
      drawIcon(g,point.x,point.y,size.width,size.height,color);
   }

   public static void drawIcon (Graphics g, int x, int y, int width, int height, Color color)
   {
      if (color == null) return;  // no unit => no icon.
      Color save = g.getColor();
      g.setColor(color);
      g.fillRect(x,y,width,height);
      g.setColor(Color.black);
      g.drawRect(x,y,width-1,height-1);
      g.setColor(save);
   }

   public static void drawIcon (Graphics g, int x, int y, int width, int height)
   { // transparent
      Color save = g.getColor();
      g.setColor(Color.black);
      g.drawRect(x,y,width-1,height-1);
      g.setColor(save);
   }

   // ---- drawText

   public static void drawText (Graphics g, String text, int x, int y)
   {
      if (text == null || text.equals("")) return;

      int h = g.getFontMetrics(g.getFont()).getHeight();

      String[] lines = lines(text);
      for (int i=0; i<lines.length; ++i)
         g.drawString(lines[i],x,y+i*h);
   }

   public static String[] lines (String text)
   {
      java.util.Vector v = new java.util.Vector();

      String t = text+"\n";
      while (!t.equals("")) {
         int k = t.indexOf("\n");
         String s = t.substring(0,k);
                t = t.substring(k+1);
         v.addElement(s);
      }
      String[] lines = new String[v.size()];
      for (int i=0; i<lines.length; ++i) lines[i] = (String)v.elementAt(i);
      return lines;
   }

   // ---- Format Methods

   public static String getNumbered (int n, String object)
   {
      return ""+n+" "+object+(n==1? "" : "s");
   }

   public static String getTitle (String hovlet)
   {
      return hovlet; // +" - Hoverball "+VERSION;
   }

   public static String getTimeString (double time)
   {
      int t = ((int)Math.round(time*10))*100;
      int min = t/60000;
      int s1 = (t/10000)%6;
      int s2 = (t/1000)%10;
      int c1 = (t/100)%10;
      return ""+min+":"+s1+s2+"."+c1;
   }

   public static String getTimeStringInteger (double time)
   {
      int t = ((int)Math.round(time*10))*100;
      int min = t/60000;
      int s1 = (t/10000)%6;
      int s2 = (t/1000)%10;
      return ""+min+":"+s1+s2;
   }

   public static String getHertzString (double hertz)
   {
      return "@ "+(int)Math.round(hertz)+" Hz";
/*
      int h = (int)Math.round(hertz*10);
      int a = h/10;
      int b = h%10;
      return "@ "+a+"."+b+" Hz";
*/
   }

   public static String getZoomString (double zoom)
   {
      int z = (int)Math.round(zoom*100);
      int z1 = (z/100);
      int z2 = (z/10)%10;
      int z3 = z%10;
      return "@ "+z1+"."+z2+z3;
   }

   // ---- Screen Drawing

   public static final int RESOLUTION = 60;  // [!] mehr?

   public static Point scale (double scale, Vector v)
   {
      return new Point((int)(-v.y*scale),(int)(-v.x*scale));
   }

   public static Point2D scale2 (double scale, Vector v)
   {
      return new Point2D.Double((-v.y*scale),(-v.x*scale));
   }

   public static AffineTransform transform (Matrix M)
   {
      return new AffineTransform(-M.a.y,-M.a.x,-M.b.y,-M.b.x,-M.c.y,-M.c.x);
   }
   
   private static final double FADE_EPSILON = 0.02;

   private static final double SPHERE_COLOR_DARKEST = 0.1;
   private static final double SPHERE_COLOR_RANGE   = 0.35;  // back: 0.1..0.45 / front: 0.65..1.0
   
   public static Color fade (Color color0, Color color1, double ratio)
   {
      if (ratio <   FADE_EPSILON) ratio = 0;
      if (ratio > 1-FADE_EPSILON) ratio = 1;
      if (color0 == null) return color1;
      if (color1 == null) return color0;
      return new Color(
            (int)(color1.getRed  ()*ratio + color0.getRed  ()*(1-ratio)),
            (int)(color1.getGreen()*ratio + color0.getGreen()*(1-ratio)),
            (int)(color1.getBlue ()*ratio + color0.getBlue ()*(1-ratio)));
   }
   
   public static Color color (Color color, Color globe, double depth)  // depth = -1..1
   {
      if (color == null) return globe;
      if (depth > 1) depth = 1;
      if (depth <-1) depth =-1;

      double ratio = (depth > 0)? 1 - SPHERE_COLOR_RANGE * (1-depth) :
               SPHERE_COLOR_DARKEST + SPHERE_COLOR_RANGE * (1+depth);

      return fade(globe,color,ratio);
   }
}
