/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.math;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;


/**
 * The class Complex represents a complex number of the two dimensional
 * complex plane.
 * <p>
 * Let be: <br>
 * &nbsp; &nbsp; <i>a</i>, <i>b</i>, <i>c</i> &nbsp;-&nbsp; complex numbers<br>
 * &nbsp; &nbsp; <i>x</i>, <i>y</i>, <i>z</i> &nbsp;-&nbsp; real numbers (as coordinates)<br>
 * &nbsp; &nbsp; <i>r</i> &nbsp;-&nbsp; real number<br>
 * &nbsp; &nbsp; <i>v</i> &nbsp;-&nbsp; vector<br>
 * &nbsp; &nbsp; <b>c</b> &nbsp;-&nbsp; the complex number of the Complex object itself
 * <p>
 * Operator Overloading for <i>Groovy</i>, <i>Scala</i>, <i>Kotlin</i> and <i>Jython</i> is supported.
 */
public final class Complex
{
   /**
    * DecimalFormat object for string conversion.
    * <p>
    * This object is used for formatting all objects of this package.
    * The default format is <code>0.###</code> with <code>Locale.US</code>.
    */
   public static DecimalFormat format = new DecimalFormat("0.###",new DecimalFormatSymbols(java.util.Locale.US));

   static String format (double d)
   {
      if (format == null) return ""+d;
      else return format.format(d);
   }

   /**
    * <i>x</i>-coordinate of the complex number.
    */
   public final double x;

   /**
    * <i>y</i>-coordinate of the complex number.
    */
   public final double y;


   // ---- constructors & bare bones ------------------------------------------

   /**
    * Creates a complex number with the coordinates (0,0).
    */
   public Complex ()                   { this.x = 0;
                                         this.y = 0; }

   /**
    * Creates a complex number with the coordinates (<i>x</i>,<i>y</i>).
    *
    * @param x <i>x</i>-coordinate
    * @param y <i>y</i>-coordinate
    */
   public Complex (double x, double y) { this.x = x;
                                         this.y = y; }

   /**
    * Creates a complex number with the coordinates (<i>x</i>,0).
    *
    * @param x <i>x</i>-coordinate
    */
   public Complex (double x) { this.x = x;
                               this.y = 0; }

   /**
    * Creates a complex number equal to another.
    *
    * @param c copied complex number
    */
   public Complex (Complex c)          { this.x =(c==null)? 0 : c.x;
                                         this.y =(c==null)? 0 : c.y; }

   /**
    * Creates a complex number from a vector.
    * <p>
    * The <i>z</i>-coordinate of the vector is ignored.
    *
    * @param v vector
    */
   public Complex (Vector v)           { this.x =(v==null)? 0 : v.x;
                                         this.y =(v==null)? 0 : v.y; }

   /**
    * Returns the hash code
    * &nbsp;[
    * = <i><code>####</code></i>
    * ].
    *
    * @return the hash code
    */
   public int hashCode ()
   {
      return new Double(x+y).hashCode();
   }

   /**
    * String representation
    * &nbsp;[
    * = "..."
    * ].
    *
    * @return the string
    */
   public String toString ()
   {
      return "("+format(x)+","+format(y)+")";
   }

   /**
    * Compares with a complex number
    * &nbsp;[
    * <b>c</b> = <i>c</i> ?
    * ].
    *
    * @param c complex number
    * @return the equality
    */
   public boolean equals (Object c)
   {
      if (c == null) return false;
      if (! (c instanceof Complex)) return false;

      Complex c_ = (Complex)c;
      return (x == c_.x && y == c_.y);
   }

   /**
    * Compares with a complex number
    * &nbsp;[
    * <b>c</b> = (<i>x</i>,<i>y</i>) ?
    * ].
    *
    * @param x real number
    * @param y real number
    * @return the equality
    */
   public boolean equals (double x, double y)
   {
      return (this.x == x && this.y == y);
   }


   // ---- object methods -----------------------------------------------------

   /**
    * Sets the coordinates
    * &nbsp;[
    * <b>c</b> = (<i>x</i>,<i>y</i>)
    * ].
    *
    * @param x real number
    * @param y real number
    * @return <b>this</b>
    */
//   public final Complex set (double x, double y) { this.x=x; this.y=y; return this;}

   /**
    * Sets the coordinates
    * &nbsp;[
    * <b>c</b> = <i>c</i>
    * ].
    *
    * @param c complex number
    * @return <b>this</b>
    */
//   public final Complex set (Complex c) { x =c.x; y =c.y; return this;}

   /**
    * Adds a complex number
    * &nbsp;[
    * <b>c</b> = <b>c</b> + <i>b</i>
    * ].
    *
    * @param b complex number
    * @return <b>this</b>
    */
   public final Complex add (Complex b) { return new Complex(x+b.x,y+b.y); }

   /**
    * Substracts a complex number
    * &nbsp;[
    * <b>c</b> = <b>c</b> - <i>b</i>
    * ].
    *
    * @param b complex number
    * @return <b>this</b>
    */
   public final Complex sub (Complex b) { return new Complex(x-b.x,y-b.y); }

   /**
    * Multiplies by a real number
    * &nbsp;[
    * <b>c</b> = <b>c</b> <i>r</i>
    * ].
    *
    * @param r real number
    * @return <b>this</b>
    */
   public final Complex mul (double r)  { return new Complex(x*r,y*r); }


   /**
    * Multiplies by a complex number
    * &nbsp;[
    * <b>c</b> = <b>c</b> <i>b</i>
    * ].
    *
    * @param b complex number
    * @return <b>this</b>
    */
   public final Complex mul (Complex b)
   {
	  return new Complex(x*b.x-y*b.y,x*b.y+y*b.x);
   }

   /**
    * Divides by a real number
    * &nbsp;[
    * <b>c</b> = <b>c</b> / <i>r</i>
    * ].
    *
    * @param r real number
    * @return <b>this</b>
    */
   public final Complex div (double r)  { return new Complex(x/r,y/r); }


   /**
    * Divides by a complex number
    * &nbsp;[
    * <b>c</b> = <b>c</b> / <i>b</i>
    * ].
    *
    * @param b complex number
    * @return <b>this</b>
    */
   public final Complex div (Complex b)
   {
	  return this.mul(b.inv());
   }

   /**
    * Builds the inverse number
    * &nbsp;[
    * <b>c</b> = 1 / <b>c</b>
    * ].
    *
    * @return <b>this</b>
    */
   public final Complex inv ()
   {
      double q = x*x+y*y;
      return new Complex(x/q,-y/q);
   }

   /**
    * Builds the negated number
    * &nbsp;[
    * <b>c</b> = -<b>c</b>
    * ].
    *
    * @return <b>this</b>
    */
   public final Complex neg ()  { return new Complex(-x,-y); }

   /**
    * Builds the conjugate
    * &nbsp;[
    * <b>c</b> = <u><b>c</b></u>
    * ].
    *
    * @return <b>this</b>
    */
   public final Complex conj () { return new Complex(x,-y); }

   /**
    * Tests if zero
    * &nbsp;[
    * <b>c</b> = 0 ?
    * ].
    *
    * @return <b>true</b>, if this complex number equals (0,0)
    */
   public final boolean zero () { return abs2() == 0; }

   /**
    * Returns the argument
    * &nbsp;[
    * = &lt;(<b>c</b>)
    * ].
    *
    * If the complex number is zero, the function returns the value null.
    *
    * @return the argument
    */
   public final double arg ()   { return Math.atan2(y,x); }  // Math.atan2(0,0) == 0

   /**
    * Returns the absolute value
    * &nbsp;[
    * = | <b>c</b> |
    * ].
    *
    * @return the absolute value
    */
   public final double abs ()   { return Math.sqrt(x*x+y*y); }

   /**
    * Returns the squared absolute value
    * &nbsp;[
    * = | <b>c</b> |^2
    * ].
    *
    * @return the squared absolute value
    */
   public final double abs2 ()  { return x*x+y*y; }

   /**
    * Normalizes the absolute value to 1
    * &nbsp;[
    * <b>c</b> = <b>c</b> / | <b>c</b> |
    * ].
    *
    * @return <b>this</b>
    */
   public final Complex norm () { double r=abs(); return new Complex(x/r,y/r); }


   // ---- class methods ------------------------------------------------------

   /**
    * Sums two complex numbers
    * &nbsp;[
    * = <i>a</i> + <i>b</i>
    * ].
    *
    * @param a complex number
    * @param b complex number
    * @return the sum
    */
   public final static Complex add (Complex a, Complex b)
   {
      return new Complex(a.x+b.x,a.y+b.y);
   }

   /**
    * Substracts two complex numbers
    * &nbsp;[
    * = <i>a</i> - <i>b</i>
    * ].
    *
    * @param a complex number
    * @param b complex number
    * @return the difference
    */
   public final static Complex sub (Complex a, Complex b)
   {
      return new Complex(a.x-b.x,a.y-b.y);
   }

   /**
    * Creates the negated number
    * &nbsp;[
    * = -<i>c</i>
    * ].
    *
    * @param c complex number
    * @return the negated number
    */
   public final static Complex neg (Complex c)
   {
      return new Complex(-c.x,-c.y);
   }

   /**
    * Creates the inverse number
    * &nbsp;[
    * = 1 / <i>c</i>
    * ].
    *
    * @param c complex number
    * @return die inverse number
    */
   public final static Complex inv (Complex c)
   {
      return new Complex(c).inv();
   }

   /**
    * Creates the conjugate
    * &nbsp;[
    * = <u><i>c</i></u>
    * ].
    *
    * @param c complex number
    * @return the conjugate
    */
   public final static Complex conj (Complex c)
   {
      return new Complex(c.x,-c.y);
   }

   /**
    * Creates the normalization of the complex number to 1.
    * &nbsp;[
    * = <i>c</i> / | <i>c</i> |
    * ].
    *
    * @param c complex number
    * @return the normalization
    */
   public final static Complex norm (Complex c)
   {
      return new Complex(c).norm();
   }

   /**
    * Multiplies a complex number by a real number.
    * &nbsp;[
    * = <i>c r</i>
    * ].
    *
    * @param c complex number
    * @param r real number
    * @return the product
    */
   public final static Complex mul (Complex c, double r)  // real multiplication
   {
      return new Complex(c.x*r,c.y*r);
   }

   /**
    * Multiplies two complex numbers.
    * &nbsp;[
    * = <i>a b</i>
    * ].
    *
    * @param a complex number
    * @param b complex number
    * @return the complex product
    */
   public final static Complex mul (Complex a, Complex b)  // complex multiplication
   {
      return new Complex(a.x*b.x-a.y*b.y,a.x*b.y+a.y*b.x);
   }

   /**
    * Returns the dot product between two complex numbers.
    * &nbsp;[
    * = &lt; <i>a</i>, <i>b</i> &gt;
    * ].
    *
    * @param a complex number
    * @param b complex number
    * @return the dot product
    */
   public final static double dot (Complex a, Complex b)  // dot product
   {
      return a.x*b.x+a.y*b.y;
   }

   /**
    * Returns the three dimensional vector product between two vectors
    * &nbsp;[
    * = <i>a</i> x <i>b</i>
    * ].
    * <p>
    * Formula:
    * &nbsp;
    * (<i>a.x</i>, <i>a.y</i>, 0) x (<i>b.x</i>, <i>b.y</i>, 0) = (0, 0, <i>c.z</i>)
    *
    * @param a complex number
    * @param b complex number
    * @return the <i>z</i>-coordinate of the vector product
    */
   public final static double vec (Complex a, Complex b)  // vector product: (a.x,a.y,0) x (b.x,b.y,0) = (0,0,*)
   {
      return a.x*b.y-a.y*b.x;
   }

   /**
    * Returns the three dimensional vector product between two vectors
    * &nbsp;[
    * = <i>a</i> x <i>b</i>
    * ].
    * <p>
    * Formula:
    * &nbsp;
    * (0, 0, <i>a.z</i>) x (<i>b.x</i>, <i>b.y</i>, 0) = (<i>c.x</i>, <i>c.y</i>, 0)</i>
    *
    * @param a_z real number
    * @param b complex number
    * @return the (<i>x</i>,<i>y</i>)-coordinates of the vector product
    */
   public final static Complex vec (double a_z, Complex b)  // vector product: (0,0,a_z) x (b.x,b.y,0) = (*,*,0)
   {
      return new Complex(-a_z*b.y,a_z*b.x);
   }

   /**
    * Returns the value of the complex exponential function
    * &nbsp;[
    * = exp i<i>y</i>
    * ].
    *
    * @param y real number
    * @return the exponential value
    */
   public final static Complex expi (double y)  // exp(i*y)
   {
      return new Complex(Math.cos(y),Math.sin(y));
   }

   /**
    * Returns the value of the complex exponential function
    * &nbsp;[
    * = exp <i>a</i>
    * ].
    *
    * @param a complex number
    * @return the exponential value
    */
   public final static Complex exp (Complex a)  // exp(a)
   {
      return Complex.expi(a.y).mul(Math.exp(a.x));
   }

// [oo]
// =========================================================================

   public boolean equals (double r) { return equals(valueOf(r)); }

// =========================================================================

/*
 * java-oo
 *
 * [!] '==' / '!=' not overloaded, use c.equals(b) instead
 * [!] '+=' etc.   not overloaded, use c = c + b   instead
 * [!] no unary +
 *
 */

// public final Complex add            (Complex b) { return add(b); }                 // +
   public final Complex add            (double r)  { return add(valueOf(r)); }
   public final Complex addRev         (double r)  { return add(valueOf(r)); }
   //public final Complex positive       ()          { return this; }
   public final Complex subtract       (Complex b) { return sub(b); }                 // -
   public final Complex subtract       (double r)  { return sub(valueOf(r)); }
   public final Complex subtractRev    (double r)  { return neg().add(valueOf(r)); }
   public final Complex negate         ()          { return neg(); }
   public final Complex multiply       (Complex b) { return mul(b); }                 // *
   public final Complex multiply       (double r)  { return mul(r); }
   public final Complex multiplyRev    (double r)  { return mul(r); }
   public final Complex divide         (Complex b) { return div(b); }                 // /
   public final Complex divide         (double r)  { return div(r); }
   public final Complex divideRev      (double r)  { return inv().mul(r); }
   public final Complex not            ()          { return inv(); }                  // ~
   public static final Complex valueOf (double r)  { return new Complex(r,0); }       // c = expr

// =========================================================================

/*
 * Groovy
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 *
 */

   public final Complex plus           (Complex b) { return add(b); }               // +
   public final Complex plus           (double r)  { return add(valueOf(r)); }
   //public final Complex addRev         (double r)  { return add(valueOf(r)); }
   public final Complex positive       ()          { return this; }
   public final Complex minus          (Complex b) { return sub(b); }               // -
   public final Complex minus          (double r)  { return sub(valueOf(r)); }
   //public final Complex subtractRev    (double r)  { return neg().add(valueOf(r)); }
   public final Complex negative       ()          { return neg(); }
// public final Complex multiply       (Complex b) { return mul(b); }               // *
// public final Complex multiply       (double r)  { return mul(r); }
   //public final Complex multiplyRev    (double r)  { return mul(r); }
// public final Complex div            (Complex b) { return div(b); }               // /
// public final Complex div            (double r)  { return div(r); }
   //public final Complex divideRev      (double r)  { return inv().mul(r); }
   public final Complex bitwiseNegate  ()          { return inv(); }                // ~
   //public static final Complex valueOf (double r)  { return new Complex(r,0); }     // c = expr

// =========================================================================

/*
 * Kotlin
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 * [!] ! instead of ~
 * [!] Kotlin does not support automatic type cast from int to double
 *
 */

// public final Complex plus           (Complex b) { return add(b); }               // +
// public final Complex plus           (double r)  { return add(valueOf(r)); }
   public final Complex plus           (int r)     { return add(valueOf((double)r)); }
   //public final Complex addRev         (double r)  { return add(valueOf(r)); }
   public final Complex unaryPlus      ()          { return this; }
// public final Complex minus          (Complex b) { return sub(b); }               // -
// public final Complex minus          (double r)  { return sub(valueOf(r)); }
   public final Complex minus          (int r)     { return sub(valueOf((double)r)); }
   //public final Complex subtractRev    (double r)  { return neg().add(valueOf(r)); }
   public final Complex unaryMinus     ()          { return neg(); }
   public final Complex times          (Complex b) { return mul(b); }               // *
   public final Complex times          (double r)  { return mul(r); }
   public final Complex times          (int r)     { return mul((double)r); }
   //public final Complex multiplyRev    (double r)  { return mul(r); }
// public final Complex div            (Complex b) { return div(b); }               // /
// public final Complex div            (double r)  { return div(r); }
   //public final Complex divideRev      (double r)  { return inv().mul(r); }
// public final Complex not            ()          { return inv(); }                // ! statt ~
   //public static final Complex valueOf (double r)  { return new Complex(r,0); }     // c = expr

// =========================================================================

/*
 * Jython
 *
 * [?] no limitiations, everything's fine!
 *
 */

   public final Complex __add__        (Complex b) { return add(b); }                 // +
   public final Complex __add__        (double r)  { return add(valueOf(r)); }
   public final Complex __radd__       (double r)  { return add(valueOf(r)); }
   public final Complex __pos__        ()          { return this; }
   public final Complex __sub__        (Complex b) { return sub(b); }                 // -
   public final Complex __sub__        (double r)  { return sub(valueOf(r)); }
   public final Complex __rsub__       (double r)  { return neg().add(valueOf(r)); }
   public final Complex __neg__        ()          { return neg(); }
   public final Complex __mul__        (Complex b) { return mul(b); }                 // *
   public final Complex __mul__        (double r)  { return mul(r); }
   public final Complex __rmul__       (double r)  { return mul(r); }
   public final Complex __div__        (Complex b) { return div(b); }                 // /
   public final Complex __div__        (double r)  { return div(r); }
   public final Complex __rdiv__       (double r)  { return inv().mul(r); }
   public final Complex __invert__     ()          { return inv(); }                  // ~
   //public static final Complex valueOf (double r)  { return new Complex(r,0); }       // c = expr

// =========================================================================

/*
 * Scala
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 *
 */

   public final Complex $plus          (Complex b) { return add(b); }               // +
   public final Complex $plus          (double r)  { return add(valueOf(r)); }
   //public final Complex addRev         (double r)  { return add(valueOf(r)); }
   public final Complex unary_$plus    ()          { return this; }
   public final Complex $minus         (Complex b) { return sub(b); }               // -
   public final Complex $minus         (double r)  { return sub(valueOf(r)); }
   //public final Complex subtractRev    (double r)  { return neg().add(valueOf(r)); }
   public final Complex unary_$minus   ()          { return neg(); }
   public final Complex $times         (Complex b) { return mul(b); }               // *
   public final Complex $times         (double r)  { return mul(r); }
   //public final Complex multiplyRev    (double r)  { return mul(r); }
   public final Complex $div           (Complex b) { return div(b); }               // /
   public final Complex $div           (double r)  { return div(r); }
   //public final Complex divideRev      (double r)  { return inv().mul(r); }
   public final Complex unary_$tilde   ()          { return inv(); }                // ~
   //public static final Complex valueOf (double r)  { return new Complex(r,0); }     // c = expr

// =========================================================================

}
