/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.math;


/**
 * The class Matrix represents a real 3 x 3-Matrix.
 * <p>
 * Let be: <br>
 * &nbsp; &nbsp; <i>A</i>, <i>B</i>, <i>M</i> &nbsp;-&nbsp; matrices<br>
 * &nbsp; &nbsp; <i>a</i>, <i>b</i>, <i>c</i> &nbsp;-&nbsp; vectors (as column vectors) <br>
 * &nbsp; &nbsp; <i>x</i>, <i>y</i>, <i>z</i> &nbsp;-&nbsp; real numbers (as coordinates)<br>
 * &nbsp; &nbsp; <i>v</i> &nbsp;-&nbsp; vector<br>
 * &nbsp; &nbsp; <i>r</i> &nbsp;-&nbsp; real number<br>
 * &nbsp; &nbsp; <i>a</i> (<i>"alpha"</i>) &nbsp;-&nbsp; real number (as angle)<br>
 * &nbsp; &nbsp; <b>x</b>, <b>y</b>, <b>z</b> &nbsp;-&nbsp; the unit vectors<br>
 * &nbsp; &nbsp; <b>M</b> &nbsp;-&nbsp; the matrix of the Matrix object itself
 * <p>
 * <b>Note:</b> Unlike the standard notation of mathematics, matrix operations
 * are read <i>from left to right</i>.
 * Thus, in Hoverball we write <i>v A B</i> for a vector <i>v</i> which is multiplied by the matrix <i>A</i>
 * and afterwards by the matrix <i>B</i> (instead of <i>B A v</i>).
* <p>
 * Operator Overloading for <i>Groovy</i>, <i>Scala</i>, <i>Kotlin</i> and <i>Jython</i> is supported.
 */
public final class Matrix
{
   /**
    * First column vector of the matrix.
    */
   public final Vector a;

   /**
    * Second column vector of the matrix.
    */
   public final Vector b;

   /**
    * Third column vector of the matrix.
    */
   public final Vector c;


   // ---- constructors & bare bones ------------------------------------------

   /**
    * Creates a unit matrix.
    */
   public Matrix () { this(1); }

   /**
    * Creates a multiple of the unit matrix.
    *
    * @param r factor
    */
   public Matrix (double r)
   {
      a = new Vector(r,0,0);
      b = new Vector(0,r,0);
      c = new Vector(0,0,r);
   }

   /**
    * Creates a matrix with the coordinates ((<i>ax</i>,<i>ay</i>,<i>az</i>),
    * (<i>bx</i>,<i>by</i>,<i>bz</i>), (<i>cx</i>,<i>cy</i>,<i>cz</i>)).
    *
    * @param ax </i>x</i>-coordinate of the first column vector
    * @param ay </i>y</i>-coordinate of the first column vector
    * @param az </i>z</i>-coordinate of the first column vector
    * @param bx </i>x</i>-coordinate of the second column vector
    * @param by </i>y</i>-coordinate of the second column vector
    * @param bz </i>z</i>-coordinate of the second column vector
    * @param cx </i>x</i>-coordinate of the third column vector
    * @param cy </i>y</i>-coordinate of the third column vector
    * @param cz </i>z</i>-coordinate of the third column vector
    */
   public Matrix (double ax, double ay, double az,
                  double bx, double by, double bz,
                  double cx, double cy, double cz)
   {
      a = new Vector(ax,ay,az);
      b = new Vector(bx,by,bz);
      c = new Vector(cx,cy,cz);
   }

   /**
    * Creates a matrix with the column vectors (<i>a</i>,<i>b</i>,<i>c</i>).
    *
    * @param a first column vector
    * @param b second column vector
    * @param c third column vector
    */
   public Matrix (Vector a, Vector b, Vector c)
   {
      this.a = a;
      this.b = b;
      this.c = c;
   }

   /**
    * Creates a matrix equal to another.
    *
    * @param M copied matrix
    */
   public Matrix (Matrix M)
   {
      this.a = (M == null)? new Vector(1,0,0) : M.a;
      this.b = (M == null)? new Vector(0,1,0) : M.b;
      this.c = (M == null)? new Vector(0,0,1) : M.c;
   }

   /**
    * Returns the hash code
    * &nbsp;[
    * = <i><code>####</code></i>
    * ].
    *
    * @return the hash code
    */
   public int hashCode ()
   {
      return (a == null? 0 : a.hashCode())
           + (b == null? 0 : b.hashCode())
           + (c == null? 0 : c.hashCode());
   }

   /**
    * String representation
    * &nbsp;[
    * = "..."
    * ].
    *
    * @return the string
    */
   public String toString ()
   {
      return "("+a+","+b+","+c+")";
   }

   /**
    * Compares with another matrix
    * &nbsp;[
    * <b>M</b> = <i>M</i> ?
    * ].
    *
    * @param M matrix
    * @return the equality
    */
   public boolean equals (Object M)
   {
      if (M == null) return false;
      if (! (M instanceof Matrix)) return false;

      Matrix M_ = (Matrix)M;
      return ((a == null? M_.a == null : a.equals(M_.a)) &&
              (b == null? M_.b == null : b.equals(M_.b)) &&
              (c == null? M_.c == null : c.equals(M_.c)));
   }


   // ---- object methods -----------------------------------------------------

   /**
    * Adds a matrix
    * &nbsp;[
    * <b>M</b> = <b>M</b> + <i>B</i>
    * ].
    *
    * @param B matrix
    * @return <b>this</b>
    */
   public final Matrix add (Matrix B) { return new Matrix(a.add(B.a),b.add(B.b),c.add(B.c)); }

   /**
    * Substracts a matrix
    * &nbsp;[
    * <b>M</b> = <b>M</b> - <i>B</i>
    * ].
    *
    * @param B matrix
    * @return <b>this</b>
    */
   public final Matrix sub (Matrix B) { return new Matrix(a.sub(B.a),b.sub(B.b),c.sub(B.c)); }

   /**
    * Multiplies by a real number.
    * &nbsp;[
    * <b>M</b> = <b>M</b> <i>r</i>
    * ].
    *
    * @param r real number
    * @return <b>this</b>
    */
   public final Matrix mul (double r)  { return new Matrix(a.mul(r),b.mul(r),c.mul(r)); }

   /**
    * Pre-multiplies by a matrix
    * &nbsp;[
    * <b>M</b> = <i>A</i> <b>M</b>
    * ].
    *
    * @param A matrix
    * @return <b>this</b>
    */
   public final Matrix mulL (Matrix A)
   {
      return Matrix.mul(A,this);
   }

   /**
    * Post-multiplies by a matrix
    * &nbsp;[
    * <b>M</b> = <b>M</b> <i>B</i>
    * ].
    *
    * @param B matrix
    * @return <b>this</b>
    */
   public final Matrix mulR (Matrix B)
   {
      return Matrix.mul(this,B);
   }

   /**
    * Equal to the method mulR(Matrix).
    *
    * @param B matrix
    * @return <b>this</b>
    */
   public final Matrix mul (Matrix B)
   {
      return mulR(B);
   }

   /**
    * Conjugates by a matrix
    * &nbsp;[
    * <b>M</b> = <i>B</i>^ <b>M</b> <i>B</i>
    * ].
    *
    * @param B matrix
    * @return <b>this</b>
    */
   public final Matrix conj (Matrix B)  // Matrix conjugation [***anders als Mathe!***]
   {
	  return Matrix.inv(B).mul(this).mul(B);
//      mulL(Matrix.inv(B));
//      mulR(B);
//      return this;
   }

   /**
    * Finds the inverse of the matrix
    * &nbsp;[
    * <b>M</b> = <b>M</b>^
    * ].
    *
    * @return <b>this</b>
    */
   public final Matrix inv ()
   {
      Matrix C = new Matrix(+ b.y*c.z - b.z*c.y, - b.x*c.z + b.z*c.x, + b.x*c.y - b.y*c.x,
                            - a.y*c.z + a.z*c.y, + a.x*c.z - a.z*c.x, - a.x*c.y + a.y*c.x,
                            + a.y*b.z - a.z*b.y, - a.x*b.z + a.z*b.x, + a.x*b.y - a.y*b.x);
      return C.trans().mul(1/det());
   }

   /**
    * Negates the matrix
    * &nbsp;[
    * <b>M</b> = -<b>M</b>
    * ].
    *
    * @return <b>this</b>
    */
   public final Matrix neg ()  { return new Matrix(a.neg(),b.neg(),c.neg()); }

   /**
    * Transposes the matrix
    * &nbsp;[
    * <b>M</b> = <u><b>M</b></u>
    * ].
    *
    * @return <b>this</b>
    */
   public final Matrix trans () { return new Matrix(a.x,b.x,c.x,a.y,b.y,c.y,a.z,b.z,c.z); }

   /**
    * Tests if zero
    * &nbsp;[
    * <b>M</b> = 0 ?
    * ].
    *
    * @return <b>true</b>, if this matrix is (0)
    */
   public final boolean zero () { return a.zero() && b.zero() && c.zero(); }

   /**
    * Returns the determinant
    * &nbsp;[
    * = det <b>M</b>
    * ].
    *
    * @return the determinant
    */
   public final double det ()
   {
      return a.x*b.y*c.z - a.x*b.z*c.y
           - a.y*b.x*c.z + a.y*b.z*c.x
           + a.z*b.x*c.y - a.z*b.y*c.x;
   }


   // ---- class methods ------------------------------------------------------

   /**
    * Sums two matrices
    * &nbsp;[
    * = <i>A</i> + <i>B</i>
    * ].
    *
    * @param A matrix
    * @param B matrix
    * @return the sum
    */
   public final static Matrix add (Matrix A, Matrix B)
   {
      return new Matrix(Vector.add(A.a,B.a),Vector.add(A.b,B.b),Vector.add(A.c,B.c));
   }

   /**
    * Substracts two matrices
    * &nbsp;[
    * = <i>A</i> - <i>B</i>
    * ].
    *
    * @param A matrix
    * @param B matrix
    * @return the difference
    */
   public final static Matrix sub (Matrix A, Matrix B)
   {
      return new Matrix(Vector.sub(A.a,B.a),Vector.sub(A.b,B.b),Vector.sub(A.c,B.c));
   }

   /**
    * Creates the negated matrix
    * &nbsp;[
    * = -<i>M</i>
    * ].
    *
    * @param M matrix
    * @return the negated matrix
    */
   public final static Matrix neg (Matrix M)
   {
      return M.neg();
   }

   /**
    * Creates the inverse matrix
    * &nbsp;[
    * = <i>M</i>^
    * ].
    *
    * @param M matrix
    * @return the inverse matrix
    */
   public final static Matrix inv (Matrix M)
   {
      return M.inv();
   }

   /**
    * Creates the transposed matrix
    * &nbsp;[
    * = <u><i>M</i></u>
    * ].
    *
    * @param M matrix
    * @return the transposed matrix
    */
   public final static Matrix trans (Matrix M)
   {
      return M.trans();
   }

   /**
    * Multiplies a matrix by a real number
    * &nbsp;[
    * = <i>M r</i>
    * ].
    *
    * @param M matrix
    * @param r real number
    * @return the product
    */
   public final static Matrix mul (Matrix M, double r)  // real multiplication
   {
      return M.mul(r);
   }

   /**
    * Multiplies two matrices
    * &nbsp;[
    * = <i>A B</i>
    * ].
    *
    * @param A matrix
    * @param B matrix
    * @return the product
    */
   public final static Matrix mul (Matrix A, Matrix B)  // Matrix multiplication [***anders als Mathe!***]
   {
      return new Matrix(B.a.x*A.a.x + B.b.x*A.a.y + B.c.x*A.a.z ,
                        B.a.y*A.a.x + B.b.y*A.a.y + B.c.y*A.a.z ,
                        B.a.z*A.a.x + B.b.z*A.a.y + B.c.z*A.a.z ,

                        B.a.x*A.b.x + B.b.x*A.b.y + B.c.x*A.b.z ,
                        B.a.y*A.b.x + B.b.y*A.b.y + B.c.y*A.b.z ,
                        B.a.z*A.b.x + B.b.z*A.b.y + B.c.z*A.b.z ,

                        B.a.x*A.c.x + B.b.x*A.c.y + B.c.x*A.c.z ,
                        B.a.y*A.c.x + B.b.y*A.c.y + B.c.y*A.c.z ,
                        B.a.z*A.c.x + B.b.z*A.c.y + B.c.z*A.c.z);
   }


   /**
    * Conjugates a matrix by a matrix
    * &nbsp;[
    * = <i>B</i>^ <i>A</i> <i>B</i>
    * ].
    *
    * @param A matrix
    * @param B matrix
    * @return the conjugate matrix
    */
   public final static Matrix conj (Matrix A, Matrix B)  // Matrix conjugation
   {
      return A.conj(B);
   }


   /**
    * Creates a turn matrix around the <i>x</i>-axis
    * &nbsp;[
    * = <i>D</i> (<b>x</b>,<i>a</i>)
    * ].
    *
    * @param alpha angle
    * @return the turn matrix
    */
   public final static Matrix rotX (double alpha)
   {
      Complex c = Complex.expi(-alpha);
      return new Matrix( 1 ,  0  , 0   ,
                         0 , c.x ,-c.y ,
                         0 , c.y , c.x );
   }

   /**
    * Creates a turn matrix around the <i>y</i>-axis
    * &nbsp;[
    * = <i>D</i> (<b>y</b>,<i>a</i>)
    * ].
    *
    * @param alpha angle
    * @return the turn matrix
    */
   public final static Matrix rotY (double alpha)
   {
      Complex c = Complex.expi(-alpha);
      return new Matrix( c.x , 0 , c.y ,
                          0  , 1 ,  0  ,
                        -c.y , 0 , c.x );
   }

   /**
    * Creates a turn matrix around the <i>z</i>-axis
    * &nbsp;[
    * = <i>D</i> (<b>z</b>,<i>a</i>)
    * ].
    *
    * @param alpha angle
    * @return the turn matrix
    */
   public final static Matrix rotZ (double alpha)
   {
      Complex c = Complex.expi(-alpha);
      return new Matrix( c.x ,-c.y , 0 ,
                         c.y , c.x , 0 ,
                          0  ,  0  , 1 );
   }

   /**
    * Creates a turn matrix around an arbitrary axis
    * &nbsp;[
    * = <i>D</i> (<b>v</b>,<i>a</i>)
    * ].
    *
    * @param v vector that you turn around
    * @param alpha angle
    * @return the turn matrix
    */
   public final static Matrix rot (Vector v, double alpha)
   {
      if (v == null || v.zero()) return new Matrix();

      Vector v1 = Vector.norm(v);
      Vector v2 = Vector.norm(Vector.vec(v1,new Vector(v1.y,v1.z,-v1.x)));
      Vector v3 = Vector.norm(Vector.vec(v1,v2));
      Matrix M = new Matrix(v1,v2,v3);

      return Matrix.mul(Matrix.trans(M),Matrix.mul(Matrix.rotX(alpha),M));
   }

// [oo]
// =========================================================================

/*
 * java-oo
 *
 * [!] '==' / '!=' not overloaded, use c.equals(b) instead
 * [!] '+=' etc.   not overloaded, use c = c + b   instead
 * [!] no unary +
 *
 */

// public final Matrix add            (Matrix B) { return add(B); }                 // +
   //public final Matrix positive       ()          { return this; }
   public final Matrix subtract       (Matrix B) { return sub(B); }                 // -
   public final Matrix negate         ()         { return neg(); }
   public final Matrix multiply       (Matrix B) { return mulR(B); }                // *
   public final Matrix multiply       (double r) { return mul(r); }
   public final Matrix multiplyRev    (double r) { return mul(r); }
   public final Matrix divide         (Matrix B) { return mulR(B.inv()); }          // /
   public final Matrix divide         (double r) { return mul(1/r); }
   public final Matrix divideRev      (double r) { return inv().mul(r); }
   public final Matrix not            ()         { return inv(); }                  // ~

// =========================================================================

/*
 * Groovy
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 *
 */

   public final Matrix plus           (Matrix B) { return add(B); }                 // +
   public final Matrix positive       ()         { return this; }
   public final Matrix minus          (Matrix B) { return sub(B); }                 // -
   public final Matrix negative       ()         { return neg(); }
// public final Matrix multiply       (Matrix B) { return mulR(B); }                // *
// public final Matrix multiply       (double r) { return mul(r); }
   //public final Matrix multiplyRev    (double r) { return mul(r); }
   public final Matrix div            (Matrix B) { return mulR(B.inv()); }          // /
   public final Matrix div            (double r) { return mul(1/r); }
   //public final Matrix divideRev      (double r) { return inv().mul(r); }
   public final Matrix bitwiseNegate  ()         { return inv(); }                  // ~

// =========================================================================

/*
 * Kotlin
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 * [!] ! instead of ~
 * [!] Kotlin does not support automatic type cast from int to double
 *
 */

// public final Matrix plus           (Matrix B) { return add(B); }                 // +
   public final Matrix unaryPlus      ()         { return this; }
// public final Matrix minus          (Matrix B) { return sub(B); }                 // -
   public final Matrix unaryMinus     ()         { return neg(); }
   public final Matrix times          (Matrix B) { return mulR(B); }                // *
   public final Matrix times          (double r) { return mul(r); }
   public final Matrix times          (int r)    { return mul((double)r); }
   //public final Matrix multiplyRev    (double r) { return mul(r); }
// public final Matrix div            (Matrix B) { return mulR(B.inv()); }          // /
// public final Matrix div            (double r) { return mul(1/r); }
   public final Matrix div            (int r)    { return mul(1/(double)r); }
   //public final Matrix divideRev      (double r) { return inv().mul(r); }
// public final Matrix not            ()         { return inv(); }                  // ! statt ~

// =========================================================================

/*
 * Jython
 *
 * [?] no limitiations, everything's fine!
 *
 */

   public final Matrix __add__        (Matrix B) { return add(B); }                 // +
   public final Matrix __pos__        ()          { return this; }
   public final Matrix __sub__        (Matrix B) { return sub(B); }                 // -
   public final Matrix __neg__        ()         { return neg(); }
   public final Matrix __mul__        (Matrix B) { return mulR(B); }                // *
   public final Matrix __mul__        (double r) { return mul(r); }
   public final Matrix __rmul__       (double r) { return mul(r); }
   public final Matrix __div__        (Matrix B) { return mulR(B.inv()); }          // /
   public final Matrix __div__        (double r) { return mul(1/r); }
   public final Matrix __rdiv__       (double r) { return inv().mul(r); }
   public final Matrix __invert__     ()         { return inv(); }                  // ~

// =========================================================================

/*
 * Scala
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 *
 */

   public final Matrix $plus          (Matrix B) { return add(B); }                 // +
   public final Matrix unary_$plus    ()         { return this; }
   public final Matrix $minus         (Matrix B) { return sub(B); }                 // -
   public final Matrix unary_$minus   ()         { return neg(); }
   public final Matrix $times         (Matrix B) { return mulR(B); }                // *
   public final Matrix $times         (double r) { return mul(r); }
   //public final Matrix multiplyRev    (double r) { return mul(r); }
   public final Matrix $div           (Matrix B) { return mulR(B.inv()); }          // /
   public final Matrix $div           (double r) { return mul(1/r); }
   //public final Matrix divideRev      (double r) { return inv().mul(r); }
   public final Matrix unary_$tilde   ()         { return inv(); }                  // ~

// =========================================================================

}
