/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.math;


/**
 * The class Sphere represents the surface of a three dimensional sphere.
 * <p>
 * A sphere is only defined by its radius.
 * Unlike Matrix, Vector and Complex it is a rather stiff object.
 * You cannot "calculate" a lot, but the class Sphere offers a small library
 * of spherical operations that eases the handling of matrices, vectors and
 * complex numbers.
 */
public final class Sphere
{
   // ---- class Sphere -------------------------------------------------------

   /**
    * Sphere radius
    */
   public final double rad;

   /**
    * Sphere perimeter
    */
   public final double per;


   /**
    * Creates a sphere with the specified radius.
    *
    * @param r Radius
    */
   public Sphere (double r) { rad = r; per = 2*Math.PI*r; }



   /**
    * Returns the hash code.
    *
    * @return the hash code
    */
   public int hashCode ()
   {
      return new Double(rad).hashCode();
   }

   /**
    * String representation.
    *
    * @return the string
    */
   public String toString ()
   {
      return ""+getClass().getName()+"[rad="+Complex.format(rad)+"]";
   }

   /**
    * Compares with a sphere
    *
    * @param S sphere
    * @return the equality
    */
   public boolean equals (Object S)
   {
      if (S == null) return false;
      if (! (S instanceof Sphere)) return false;

      Sphere S_ = (Sphere)S;
      return rad == S_.rad;
   }


   /**
    * Returns the distance between two vectors <i>on</i> the sphere.
    * <p>
    * If <i>R</i> is the sphere radius, and if <i>a</i>, <i>b</i> != 0 are the
    * two vectors, this function will compute:
    * <p>
    * <i>R</i> * acos(&lt; <i>a</i>, <i>b</i> &gt; / |<i>a</i>||<i>b</i>|)
    *
    * @param a vector
    * @param b vector
    * @return the distance
    */
   public final double diff (Vector a, Vector b)
   {
      return rad * Math.acos(Vector.dot(a,b)/a.abs()/b.abs());
   }

   /**
    * Creates an orthonormal basis from a given vector.
    * <p>
    * If <i>v</i> != 0 is the vector
    * and (<i>a</i>,<i>b</i>,<i>c</i>) the orthonormal basis,
    * there is:
    * <ul>
    * <li> <i>c</i> = <i>v</i> / |<i>v</i>|
    * <li> <i>b</i> = <b>z</b> x <i>c</i> &nbsp; [ <b>z</b> = (0,0,1) ]
    * <li> <i>a</i> = <i>b</i> x <i>c</i>
    * </ul>
    *
    * @param v vector
    * @return the basis
    */
   public final Matrix base (Vector v)
   {
      if (v == null || v.zero()) return new Matrix();
      Vector c = Vector.norm(v);
      if (c.x == 0 && c.y == 0 && c.z ==  1) return new Matrix();
      if (c.x == 0 && c.y == 0 && c.z == -1) return new Matrix(-1,0,0, 0,1,0, 0,0,-1);
      Vector b = Vector.vec(new Vector(0,0,1),c).norm();
      Vector a = Vector.vec(b,c).norm();
      return new Matrix(a,b,c);
   }

   /**
    * Creates an orthonormal basis from a given vector and a turn angle.
    * <p>
    * The orthonormal basis (<i>a</i>,<i>b</i>,<i>c</i>) is first generated
    * from the vector, then turned by the turn angle around its
    * <i>c</i>-axis.
    *
    * @param v vector
    * @param alpha angle
    * @return the basis
    */
   public final Matrix base (Vector v, double alpha)
   {
      return base(v).mulL(Matrix.rotZ(alpha));
   }


   /**
    * Transforms a plane position into a spherical position.
    *
    * @param c complex number
    * @return the spherical vector
    */
   public final Vector warp (Complex c)
   {
      return warp(c,0).c;
   }

   /**
    * Transforms a plane position into a spherical position.
    * <p>
    * The plane visual angles are parametrized in a way that
    * angles of value 0 point along the ray from the origin to the position.
    *
    * @param c complex number
    * @param alpha visual angle
    * @return the spherical matrix
    */
   public final Matrix warp (Complex c, double alpha)
   {
      if (c == null || c.zero()) return Matrix.rotZ(alpha);
      else return euler(new Vector(c.arg(),c.abs()/rad,alpha));
   }

   /**
    * Transforms a spherical position into a plane position.
    *
    * @param v spherical vector
    * @return the complex number
    */
   public final Complex warp (Vector v)
   {
      if (v == null || v.zero()) return new Complex();
      else {
         Vector w = Vector.norm(v);
         if (w.x == 0 && w.y == 0) return new Complex(rad*Math.PI,0); // w.z = -1!
         else return new Complex(w.x,w.y).norm().mul(rad*Math.acos(w.z));
      }
   }


   /**
    * Transforms a spherical position into a plane position.
    * <p>
    * The plane visual angles are parametrized in a way that
    * angles of value 0 point along the ray from the origin to the position.
    * <p>
    * For practical reasons, the returned values <i>position</i> and <i>visual angle</i>
    * are packed in a Vector object and should be decomposed at once by
    * a programming code as follows:
    *<pre>
    *Vector v = Sphere.warp(M);         // "(c,a) = Sphere.warp(M);"
    *&nbsp;
    *Complex c = new Complex(cf.x,v.y);  // c = position
    *double a = v.z;                    // a = visual angle
    *</pre>
    *
    * @param M spherical matrix
    * @return the complex number and the visual angle as Vector object
    */
   public final Vector warp (Matrix M)  // returns (c,alpha) as Vector!
   {
      if (M == null) return new Vector();
      else {
         Vector v = euler(M);
         Complex c = Complex.expi(v.x).mul(v.y*rad);
         return new Vector(c.x,c.y,v.z);
      }
   }


   /**
    * Creates the matrix from an Euler vector.
    *
    * @param v Euler vector
    * @return the matrix
    */
   public final static Matrix euler (Vector v)
   {
      Matrix M = Matrix.rotZ(v.z);
      M = M.mul(Matrix.rotY(v.y)); /*immutable*/
      M = M.mul(Matrix.rotZ(v.x)); /*immutable*/
      return M;
   }

   /**
    * Creates the Euler vector from a matrix.
    *
    * @param M matrix
    * @return the Euler vector
    */
   public final static Vector euler (Matrix M)           // returns (-pi..pi,0..pi,-pi..pi)
   {
      Vector ay = Vector.vec(new Vector(0,0,1),M.c);     // get 2nd axis
      if (ay.abs() == 0) ay = new Vector(0,1,0);
      else ay = ay.norm(); /*immutable*/

      double vy = angle(new Vector(0,0,1),M.c,ay);       // 1st angle
      double vz = angle(ay,M.b,M.c);                     // 2nd angle
      Matrix A = new Matrix(M).mul(Matrix.rot(ay,-vy));  // undo 2nd
      double vx = angle(new Vector(1,0,0),A.a,A.c)-vz;   // 3rd angle
      while (vx  >  Math.PI) vx -= 2*Math.PI;            // fit into -pi..+pi!
      while (vx <= -Math.PI) vx += 2*Math.PI;

      return new Vector(vx,vy,vz);
   }

   private final static double angle (Vector a, Vector b, Vector orientation)  // angle between
   {
      double dot = Vector.dot(a,b); // /a.abs()/b.abs();
      if (dot > 1) dot = 1;
      if (dot <-1) dot =-1;
      if (orientation == null || Vector.dot(Vector.vec(a,b),orientation) > 0)
         return  Math.acos(dot);
      else
         return -Math.acos(dot);
   }
}
