/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.net;

import java.net.*;
import java.io.*;

import java.util.Hashtable;
import java.util.Vector;


// ---- Smart Network Socket ------------------------------------ [Socket] ----

public class Socket implements Runnable  // Real-Network-Receiver Thread!
{
   public static final int FULL = 0;
   public static final int SMART = 1;
   public static final int VIRTUAL = 2;
   public static int network = FULL; // !
   private static int port = 0;  // counts up ports (only real=false mode)

   public static final int PING_EVERY = 500;
   public static final int PING_TIMEOUT = 5000;

    // Local Network Buffer (Connection -> Message-Vector)
   private static Hashtable local = new Hashtable();
   private static final int KEEP_LOCAL_AFTER_CLOSE = 10 * 1000;
 //APPLET THREAD BUG//
   static ThreadGroup group = Thread.currentThread().getThreadGroup(); //new ThreadGroup("net");
   static PrintStream log = null;               // Network Log (null: no logging)
//   static PrintStream log = System.out;         // Network Log (null: no logging)

   private Address localAddress = null;         // remember own address
   private Address remoteAddress = null;        // remember remote address
   private int timeout = 0;
   private Connection connection;               // local -> remote
   private java.net.Socket socket = null;       // underlying TCP socket
   private PrintWriter out = null;              // ... and its Output PrintWriter
   private BufferedReader in = null;            // ... and its InputStream Reader
   private boolean closed = false;              // only close once, please!

   /*
   // LOGGING
   static {
	   try {
		   Socket.log = new PrintStream(new File("hoverball.log"));
	   } catch(FileNotFoundException e) {}
   }
   */

   static String thread (String name, Address address)
   {
      return ":"+address.getPort()+" "+name;
   }

   static String threadNameSuffix (int count, int port)
   {
      return "-"+count+" :"+port;
   }

   // ---- Constructors

   Socket (java.net.Socket s) throws IOException // server side, real network
   {
	   socket = s;
	   initialize();
   }

   Socket (Address local, Address remote) // server side, virtual network
   {
	   socket = null;
	   // initialize() -- nothing to do
	   remoteAddress = remote;
	   localAddress = local;
       connection = new Connection (localAddress,remoteAddress);
       putLocal(connection);
   }

   Socket (Address serverAddress) throws IOException // client side, real & virtual network
   {
	   if (Socket.network==Socket.FULL || Socket.network==Socket.SMART)
	   {
if (Socket.log != null) Socket.log.println("Connect Socket to Server: "+serverAddress);
		   socket = new java.net.Socket(serverAddress.getHost(), serverAddress.getPort());
		   initialize();
if (Socket.log != null) Socket.log.println("Connected an initialized: "+socket);
	   }
	   else // Socket.network==Socket.VIRTUAL
	   {
		   socket = null;
		   // initialize() -- nothing to do
		   remoteAddress = serverAddress;
		   localAddress = new Address(InetAddress.getLocalHost(),++ Socket.port);
		   send(new Message(serverAddress,serverAddress,localAddress.toString()));
		   connection = new Connection(localAddress,remoteAddress);
		   putLocal(connection);
	   }
   }

   private void initialize() throws IOException
   {
	   if (socket == null) return;       // ignore of no real socket is set up
	   
	   if (Socket.network==Socket.FULL || Socket.network==Socket.SMART) {
		   //socket.setKeepAlive(true); // permanently test the connection
		   socket.setTcpNoDelay(true); // send directly, don't use Nagle's algorithm
		   /*
		    * Nagle's algorithm is used to conserve bandwidth by minimizing the number of data segments that are sent.
		    * It's useful to disable the use of Nagle's algorithm (setTcpNoDelay(true)) when your communication over
		    * that socket comprises small packets and where latency is important.
		    * The classic example that comes to mind is a talk/chat application.
		    */
		   localAddress = new Address(socket.getLocalAddress(),socket.getLocalPort());
		   remoteAddress = new Address(socket.getInetAddress(),socket.getPort());
		   connection = new Connection (localAddress,remoteAddress);
		   putLocal(connection);
           out = new PrintWriter(socket.getOutputStream(),true);
           in = new BufferedReader(new InputStreamReader(socket.getInputStream()));

           // START SOCKET THREAD FOR RECEIVING MESSAGES
		   new Thread(this,"Hoverball Socket "+connection).start(); //new Thread(Socket.group,this,Socket.thread("Socket",localAddress)).start();  // Start Real-Receiver Thread

	   }
   }

   public String toString ()
   {
      return "["+localAddress+"<->"+remoteAddress+"]";
   }

   synchronized void close () throws IOException
   {
	  if (closed) return;
      closed = true;
if (Socket.log != null) Socket.log.println("Closing: "+socket);

      Vector vector = Socket.getLocal(connection);
      synchronized(vector) {
         vector.addElement(null);  // Stopper.
         vector.notifyAll();
      }

      if (socket != null)
      {
         socket.close();
      }
   }


   // ----

   Address getLocalAddress ()
   {
      if (localAddress==null) return null;
      return new Address(localAddress);
   }

   Address getRemoteAddress ()
   {
      if (remoteAddress==null) return null;
      return new Address(remoteAddress);
   }

   void setTimeout (int timeout)
   {
      this.timeout = timeout;
   }

   static Vector getLocal(Connection connection)  // accessed by this class and class Server
   {
	   synchronized(local) {
		   Vector vector = (Vector)local.get(connection);
		   if (vector!=null) return vector;
		   else if (network == VIRTUAL)
		   {
			   Vector v = new Vector();
			   local.put(connection,v);
			   return v;
		   }
		   else return null;
	   }
   }

   static void putLocal(Connection connection)  // accessed only by this class
   {
	   synchronized(local) {
		   Vector vector = (Vector)local.get(connection);
		   if (vector==null) local.put(connection,new Vector());
	   }
   }

   synchronized void send (Message message) throws IOException
   {
	  Vector vector = Socket.getLocal(new Connection(message.receiver,message.sender)); // inverted!
      if ((Socket.network==Socket.VIRTUAL) || (Socket.network==Socket.SMART && vector != null))
      {
    	  synchronized(vector) {                // put into local network.
    		  vector.addElement(message);
    		  vector.notifyAll();
          }
      }
      else if (Socket.network==Socket.FULL || Socket.network==Socket.SMART) {  // TCP
    	  try {
    		  out.println(message.data);
    	  } catch (Exception e) { }
      }
   }

   Message receive () throws IOException
   {
      Vector vector = Socket.getLocal(connection);
      if (vector == null) throw new SocketException();  // closed!
      Message message = null;
      while (true) {
          if (!vector.isEmpty()) break;  // read packet!

    	  synchronized(vector) {
             try {
                vector.wait(timeout);
             } catch (InterruptedException e) {}

             if (vector.isEmpty()) throw new InterruptedIOException();  // Timeout!
    	  }
      }
      message = (Message)vector.elementAt(0);
      vector.removeElementAt(0);

      if (message == null) throw new SocketException();  // Stopper => close!

//!!      if (Socket.log != null) Socket.log.println("["+localAddress+"]<-["+message.data+"] " + message.data);
      return message;
   }

   // ---- Real Network Socket Receiver

   public void run ()
   {
      Thread.currentThread().setPriority(Thread.NORM_PRIORITY); // ??
      Vector vector = (Vector)local.get(connection);  // "My" Vector!
      if (vector == null || in == null) return;
      try {
         String data = null;
//         
// HERE WAS "java 24% CPU" BUG after foreign Socket connection and InterruptedIOExcpetion
//
//    	 while (!closed)
//    		if ((data = in.readLine()) != null)
//
    	 while (!closed && (data = in.readLine()) != null)
    		 synchronized(vector) {  // put into local network.
    			 vector.addElement(new Message(remoteAddress,localAddress,data));
    			 vector.notifyAll();
    		 }

      } catch(Exception e) {
	     synchronized(vector) {
	    	 vector.addElement(null);  // Stopper.
	    	 vector.notifyAll();
	     }
	  }
      
      try {
    	  close();
      } catch(Exception e) {}
   }

   // ---- Get my local IP [otherwise getLocalHost().getHostAddress() will return 127.0.1.1]

   public static InetAddress getMyLocalIP() {
       try {
           java.util.Enumeration<NetworkInterface> b = NetworkInterface.getNetworkInterfaces();
           while( b.hasMoreElements()){
               for ( InterfaceAddress f : b.nextElement().getInterfaceAddresses())
                   if ( f.getAddress().isSiteLocalAddress())
                       return f.getAddress();
           }
           // not found:
           return InetAddress.getLocalHost();
       } catch (SocketException e) {
       } catch (UnknownHostException e) {
       }
       return null;
   }
}
