/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.serial;

import hoverball.net.Address;

public abstract class SU extends Protocol // Server to Unit's Client
{
   // Protocol

   public static Serial deserialize (String string)
   {
      Object[] objects = parse(string); if (objects == null) return null;
      Serial serials = new Serial(){};  // Start-Serial (will be discarded.)
      Serial last = serials;

      for (int i=0; i<objects.length; ++i)
      try {
         Object[] subobjects = (Object[])objects[i];
         Serial serial = null;
         if (((serial = SU.Connect.deserialize(subobjects)) != null) ||
             ((serial = SU.Checkin.deserialize(subobjects)) != null) ||
             ((serial = SU.Look.deserialize(subobjects)) != null) ||
             ((serial = SU.Break.deserialize(subobjects)) != null) ||
             ((serial = SU.Bye.deserialize(subobjects)) != null) ||
             ((serial = SU.Error.deserialize(subobjects)) != null)) last = last.next = serial;
         else throw new Serial.UnknownCommand();
      } catch (Exception e) {
////System.out.println("SU.deserialize failed ("+e+"): |"+string+"|");    	  
         return null;
      }
      return serials.next;
   }


   // (connect #hash "The Ultimates!" Robocop_XXL 123456 3 2)

   public static class Connect extends Serial
   {
      public int t;
      public int n;
      public String hash;
      public String team;
      public String name;
      public int color;

      public Object clone ()
      {
         return new Connect(t,n,clone(hash),clone(team),clone(name),color).next(clone(next));
      }

      public Connect (int t, int n, String hash, String team, String name, int color)
      {
         this.t = t;
         this.n = n;
         this.hash = hash;
         this.team = team;
         this.name = name;
         this.color = color;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(connect ");
         buffer.append(putInt(t));
         buffer.append(' ');
         buffer.append(putInt(n));
         buffer.append(' ');
         buffer.append(putString(Address.HASH + hash));
         buffer.append(' ');
         buffer.append(putString(team));
         buffer.append(' ');
         buffer.append(putString(name));
         buffer.append(' ');
         buffer.append(putInt(color));
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("connect")) return null;
         if (objects.length != 7) throw new IllegalNumberOfArguments();

         int t = getInt(objects[1]);
         int n = getInt(objects[2]);
         String hash = getString(objects[3]).substring(1);
         String team = getString(objects[4]);
         String name = getString(objects[5]);
         int color = getInt(objects[6]);
         return new Connect(t,n,hash,team,name,color);
      }
   }

   // (checkin 50 25 -100 100 -100 100)

   public static class Checkin extends Serial
   {
      public String[] options;

      public Object clone ()
      {
         String[] options_ = new String[options.length];
         for (int i=0; i<options.length; ++i) options_[i] = ""+options[i];
         return new Checkin(options_).next(clone(next));
      }

      public Checkin (String[] options)
      {
         this.options = options;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(checkin");
         for (int i=0; i<options.length; ++i) {
            buffer.append(' ');
            buffer.append(options[i]);
         }
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("checkin")) return null;

         String[] options = new String[objects.length-1];
         for (int i=0; i<options.length; ++i) options[i] = ""+objects[i+1];

         return new Checkin(options);
      }
   }

   // (look 3.4 19 <Disk> <Disk> ...)

   public static class Look extends Serial
   {
      public double E;
      public double py;
      public DiskRelative[] disks;

      public Object clone ()
      {
         DiskRelative[] disks2 = new DiskRelative[disks.length];
         for (int i=0; i<disks.length; ++i) disks2[i] = (DiskRelative)disks[i].clone();

         return new Look(E,py,disks2).next(clone(next));
      }

      public Look (double E, double py, DiskRelative[] disks)
      {
         this.E = E;
         this.py = py;
         this.disks = disks;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(look");

         buffer.append(' ');
         buffer.append(putDouble(E));
         buffer.append(' ');
         buffer.append(putDouble(py));

         for (int i=0; i<disks.length; ++i) {
            buffer.append(' ');
            disks[i].serialize(buffer);
         }
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("look")) return null;

         double E = getDouble(objects[1]);
         double py = getDouble(objects[2]);

         DiskRelative[] disks = new DiskRelative[objects.length-3];
         for (int i=0; i<disks.length; ++i) {
            DiskRelative disk = (DiskRelative)DiskRelative.deserialize((Object[])objects[i+3]);
            if (disk == null) throw new WrongObjectListElement();
            else disks[i] = disk;
         }

         return new Look(E,py,disks);
      }
   }

   // (break)

   public static class Break extends Serial
   {
      public Object clone () { return new Break().next(clone(next)); }

      public void serialize (StringBuffer buffer) { buffer.append("(break)"); }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("break")) return null;
         if (objects.length != 1) throw new IllegalNumberOfArguments();

         return new Break();
      }
   }

   // (bye)

   public static class Bye extends Serial
   {
      public Object clone () { return new Bye().next(clone(next)); }

      public void serialize (StringBuffer buffer) { buffer.append("(bye)"); }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("bye")) return null;
         if (objects.length != 1) throw new IllegalNumberOfArguments();

         return new Bye();
      }
   }

   // (error "(strange command ...)")

   public static class Error extends Serial
   {
      public String command;

      public Object clone ()
      {
         return new Error(clone(command)).next(clone(next));
      }

      public Error (String command)
      {
         this.command = command;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(error ");
         buffer.append(putString(command));
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("error")) return null;
         if (objects.length != 2) throw new IllegalNumberOfArguments();

         String command = getString(objects[1]);
         return new Error(command);
      }
   }
}

