/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.simulator;

import hoverball.net.*;
import hoverball.serial.*;

import java.util.Vector;
import java.util.Enumeration;


// ---- Network Protocol: Unit Client to Simulator ------------- [Channel] ----


class Channel extends NetAgent
{
   private Simulator simulator = null;

   public int t = 0;
   public int n = 0;

   public String team = null;
   public String name = null;
   public int color = 0;
   public String host = null;

   public boolean active = false;
   public boolean looking = false;
   public int version = 0;

   public String toString() { return "(CHANNEL "+t+" "+n+" "+team+" "+name+")"; } // for debugging purposes only

   public Channel (Socket socket, Simulator simulator)
   {
      super(socket);
      this.simulator = simulator;
   }


   // ---- Receive ---------------------------------------------------------

   public void receive (Address sender, String message)
   {
	  Serial serials = US.deserialize(message);
      if (serials == null) { send(new SU.Error(message)); return; }

      Serial serial_ = serials;
      while (serial_ != null) {
         Serial serial = serial_;
         serial_ = serial_.next;
         serial.next = null;

         if (serial instanceof US.Look) simulator.look(this);
         else if (serial instanceof US.Action) simulator.action(this,(US.Action)serial);
         else if (serial instanceof US.Ping) simulator.ping(this);
         else if (serial instanceof US.Bye) simulator.bye(this);
      }
   }


// ---- Channel List ----------------------------------- [Channel.Manager] ----


   static class Manager  // Channel.Manager
   {
      private Simulator simulator = null;
      private Vector vector = new Vector(5,10);

      public Manager (Simulator simulator) { this.simulator = simulator; }


      public synchronized Channel create
         (Address host, String team, String name, int color, Socket socket)
      {
         // calculate channel-id: t

         int t = 0;

         for (Enumeration<Channel> e = vector.elements(); e.hasMoreElements(); ) {  // get t
            Channel ch = (Channel)e.nextElement();
            if (ch.team.equals(team)) { t = ch.t; break; }
         }

         if (t == 0) {  // new team? search lowest new t!
            boolean found = false;
            do {
               ++ t;
               found = false;

               if (t > Constants.T_MAX) return null;  // maximum units!

               for (Enumeration<Channel> e = vector.elements(); e.hasMoreElements() && !found; ) {
                  Channel ch = (Channel)e.nextElement();
                  found = (ch.t == t);
               }
            } while(found);
         }

         // calculate channel-id: n

         int n = 0;
         if (n <= 0 || n > Constants.N_MAX) n = 0;  // bad client...
         else --n;  // anti-"++n"

         boolean found = false;
         do {           // search lowest new n >= preferred-n
            ++ n;
            found = false;

            if (n > Constants.N_MAX) return null;  // maximum units!

            for (Enumeration<Channel> e = vector.elements(); e.hasMoreElements() && !found; ) {
               Channel ch = (Channel)e.nextElement();
               found = (ch.t == t && ch.n == n);
            }
         } while(found);

         // create channel

         Channel channel = new Channel(socket,simulator);
         channel.t = t;
         channel.n = n;
         channel.team = team;
         channel.name = name;
         channel.color = color;
         channel.host = "HOST"; //""+host; //host.getHostName();
//   channel.socket.observe = ""+name;

         vector.addElement(channel);

         return channel;
      }

      public synchronized void remove (Channel channel)
      {
         vector.removeElement(channel);
      }

      public synchronized Channel get (int t, int n)
      {
         for (Enumeration<Channel> e = vector.elements(); e.hasMoreElements(); ) {
            Channel channel = (Channel)e.nextElement();
            if (channel.t == t && channel.n == n) return channel;
         }
         return null;
      }

      public synchronized Channel[] toArray ()
      {
         Channel[] array = new Channel[vector.size()];
         vector.copyInto(array);
         return array;
      }

      public synchronized Channel[] toArray (boolean active)
      {
         Vector selected = new Vector();
         for (Enumeration<Channel> e = vector.elements(); e.hasMoreElements(); ) {
            Channel channel = (Channel)e.nextElement();
            if (channel.active == active) selected.addElement(channel);
         }

         Channel[] array = new Channel[selected.size()];
         selected.copyInto(array);
         return array;
      }

      public synchronized void send (Object message)
      {
         Channel[] array = toArray();
         for (int i=0; i<array.length; ++i) array[i].send(message);
      }

      public synchronized void send (Object message, boolean active)
      {
         Channel[] array = toArray(active);
         for (int i=0; i<array.length; ++i) array[i].send(message);
      }
   }

}
