/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.simulator;

import hoverball.layout.*;
import hoverball.net.*;
import hoverball.serial.*;

import java.util.*;
import java.util.Timer;
import java.io.IOException;
import java.net.InetAddress;
import java.text.SimpleDateFormat;


// ---- Hoverball Server ---------------------------------------- [Server] ----


public class Server extends hoverball.net.Server
{
   // ---- Network Protocol ------------------------------------------------

   public void connection (Address sender, String message, Socket socket)
   {
      // ---- (connect #hash "My Team" Pur 23456 2.3) ------------ [US] ----
      Serial serial = US.deserialize(message);
      if (serial instanceof US.Connect) {
         US.Connect c = (US.Connect)serial;

         if (c.hash.equals("")) do {
            c.hash = random();
         } while(simulators.containsKey(c.hash)); // must be a new one!

         Simulator simulator = simulators.get(c.hash);
         if (simulator == null) {
            simulator = new Simulator(this);
            simulators.put(c.hash,simulator);
         }
         simulator.connect((US.Connect)serial,sender,socket);
      }

      // ---- (connect #hash MyController) ----------------------- [CS] ----

      serial = CS.deserialize(message);
      if (serial instanceof CS.Connect) {
         CS.Connect c = (CS.Connect)serial;

         if (c.hash.equals("")) do {
            c.hash = random();
         } while(simulators.containsKey(c.hash)); // must be a new one!

         Simulator simulator = simulators.get(c.hash);
         if (simulator == null) {
            simulator = new Simulator(this);
            simulators.put(c.hash,simulator);
         }
         simulator.connect((CS.Connect)serial,sender,socket);
      }

      if (!simulators.isEmpty() && timer == null) { // Timer start!
         timer = new Timer("Hoverball Server Refresh Timer");              // ...
         timer.schedule(new TimerTask() {           // ...
            private String printed = "";

            public void run()
            {
               // gc

               List<String> trash = new ArrayList<String>(recycle.keySet());
               for (String hash : trash) {
                  long trashed = recycle.get(hash).getTime();
                  long now = new Date().getTime();

                  if (now - trashed > KEEP_IN_TRASH) { // remove from memory
                     Simulator simulator = simulators.get(hash);
                	  
                     new Thread("Hoverball Server -remove- #"+hash) { // schicke Simulator ins Nirvana!
						public void run() {
                     simulator.close();
						}
                     }.start();
         
                     simulators.remove(hash);
                     recycle.remove(hash);

                     if (simulators.isEmpty() && timer != null) { // Timer stop!
                        timer.cancel();                           // ...
                        timer.purge();                            // ...
                        timer = null;                             // ...

                        refresh(); // zeigt den 'Hoverball Server Refresh Timer' noch einmal (dauerhaft) an, obwohl/weil er beendet ist  

                        // BUG: Leider schließen nicht alle Threads jedesmal sauber ab.
                        // WORKAROUND: exit, restart by 'hoverball.server.sh'-Skript!

//                        System.exit(0);
                     }
                  }
               }

               // print to console

               String date = dateFormat.format(new Date());

               if (!date.equals(printed)) refresh();

               printed = date;
            }
         }, 0, 100);

         refresh();
      }
   }

   private static String random() {
      return "" +
            (char)('a' + (int)(Math.random()*25)) +
            (char)('a' + (int)(Math.random()*25)) +
            (char)('a' + (int)(Math.random()*25)) +
            (char)('a' + (int)(Math.random()*25)) +
             "";
   }

   // ---- class Server -------------------------------------------------------

   private Hashtable<String,Simulator> simulators = new Hashtable<String,Simulator>();
   private Hashtable<String,Date> recycle = new Hashtable<String,Date>();
   private static long KEEP_IN_TRASH = 0 * 1000; // 0 seconds reactivation period before recycling

   private Timer timer = null;

   private final String canonical;

   public Server () throws IOException { this(About.DEFAULT_PORT); }

   public Server (int port) throws IOException
   {
      super((port == -1)? About.DEFAULT_PORT : port); multiple = true;

      canonical = InetAddress.getLocalHost().getCanonicalHostName();

      Runtime.getRuntime().addShutdownHook(new Thread("Hoverball Bye Hook") {
         public void run() {
            for (Simulator simulator : simulators.values()) simulator.close();
            System.out.println("\nBye.");
         }
      });

      refresh();
   }
   
   public Server (String port) throws IOException
   {
	   this(parse(port));
   }

   private static int parse (String port)
   {
      if (port != null && port.startsWith(""+hoverball.net.Address.PORT)) try {
         return Integer.parseInt(port.substring(1));
      } catch(NumberFormatException e) {
         return -1;
      } else if (port != null) try {
         return Integer.parseInt(port);
      } catch(NumberFormatException e) {
         return -1;
      } else
         return -1;
   }
   

   // ---- Console Output -----------------------------------------------------

   private boolean firsttime = true;
   private SimpleDateFormat dateFormat = new SimpleDateFormat("[yyyy.MM.dd_HH:mm:ss]");

   public synchronized void refresh ()
   {
      // print title

      if (firsttime || Ansi.terminal) {
         if (Ansi.terminal) System.out.println(Ansi.clear() + Ansi.home());

         Address server = this.getAddress();
         String address = server.toString();
         String address_ = address.replace(server.getHostName(),canonical);

         System.out.println("Hoverball Server at '" + address_  + "' at " + address + "");
         System.out.println("Version "+hoverball.layout.About.VERSION_COMPLETE);
         System.out.println();
         System.out.println("Press 'Ctrl-C' to quit.");

         firsttime = false;
      }

      // print simulator list

      List<String> keys = new ArrayList<String>(simulators.keySet());
      Collections.sort(keys);

      if (keys.isEmpty())
      {
         System.out.println();
         System.out.println("[]");
         System.out.println("Standby.");
      }
      else
      {
         System.out.println();
         System.out.println(dateFormat.format(new Date()));

         for (String hash : keys) {
            Simulator simulator = simulators.get(hash);

            int channels_length = simulator.getNumberOfChannels();
            int controls_length = simulator.getNumberOfControls();

            if (channels_length == 0 && controls_length == 0) { // move into recycle bin
               if (!recycle.containsKey(hash)) recycle.put(hash,new Date());
            } else {                                            // or reactivate it!
               if (recycle.containsKey(hash)) recycle.remove(hash);
            }

            int state = simulator.getState();
            double time = simulator.getTime();
            double duration = simulator.getDuration();
            double hertz = simulator.getHertz();

            String Layout_getTimeString_time = Layout.getTimeString(time);

            String labelState = ((state == 0? "Initialized" :
               time == duration? "Complete" :
               state == 1? "Paused" :
               state == 2? "Running" : "???") + " - ");
            String labelTime = (state == 0? "Game open" :
               Layout_getTimeString_time.substring(0, Layout_getTimeString_time.length()-2) + " " + // no 1/10 sec
               Layout.getHertzString(hertz));
            String labelChannels = (Layout.getNumbered(channels_length,"Unit Client"));
            String labelControls = (Layout.getNumbered(controls_length,"Control Client"));

            System.out.print("#" + fix(hash,12,'.') + "   "); // 16
            System.out.print(fix(labelChannels,17));
            System.out.print(fix(labelControls,20));
            System.out.print(fix(labelState + labelTime,25));
            System.out.println();
         }
      }

 /*
if (keys.isEmpty())
{
      List<String> output = new Vector<String>();
//      System.out.println();
      for (Thread thread : Thread.getAllStackTraces().keySet()) {
//         System.out.println("-------");
         output.add("" + thread
            +" state:"+thread.getState()
//            + (thread.toString().contains("Hoverball Socket") || thread.toString().contains("BLOCKED")? "\n"+trace(thread.getStackTrace()) : "")
            );
      }
      Collections.sort(output);
      for (String string : output) System.out.println(string);
}
 */
   }

   static String trace(StackTraceElement[] stack) {
      if (stack.length == 0) return "[]";
      else {
         String s = "";
         for (int i=0; i<stack.length; ++i) s += stack[i] +"\n";
         return s;
      }
   }

   private static String fix (String string, int length) { return fix(string,length,' '); }
   private static String fix (String string, int length, char fill) {
      String s = string;
      for (int i=0; i<length; i++) s += fill;
      return s.substring(0,length);
   }
}
