/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.simulator;

import hoverball.net.*;
import hoverball.serial.*;
import hoverball.layout.About;  // Version & Default Port

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;


// ---- Hoverball Simulator ---------------------------------- [Simulator] ----


public class Simulator
{
   // ---- Thread Group

//APPLET THREAD BUG//
   // static ThreadGroup group = Thread.currentThread().getThreadGroup(); //new ThreadGroup("simulator");

   // ---- Simulator

   private static final double TIME_HISTORY = 2.0;   // 2 seconds history for Hertz measuring

   private hoverball.net.Server server = null;       // Access to network
   private Channel.Manager channels = new Channel.Manager(this); // Unit Clients
   private Control.Manager controls = new Control.Manager(this); // Control Clients

   private Constants constants = new Constants();    // permanent constant table
   private Constants constants1 = null;              // in-game-constants

   private Terrain terrain = null;             // Terrain
   private Timer timer = null;                 // Simulation Timer
   private double hertz = 0;                   // Current/Last empirical frequency
   private java.util.Vector history = new java.util.Vector();  // Empirical frequency


   private double frequency;
   private double timeratio;
   private double duration;

   private int state = 0;                      // Simulator State (=0,1,2)

   private View view = null;                   // most recent View
   private int version = 0;                    // version of view
   private Hashtable timeouts = new Hashtable(); // Score Timeout (""+t -> Double)

   public final SimulatorFrame frame;          // Simulator's AWT-Frame

   public final String address;
   private ChannelGuard channelGuard = new ChannelGuard();


   // ---- Constructor --------------------------------------------------------

   public Simulator (hoverball.net.Server server) { // accessed by Server
      this.server = server;
      this.address = ""+server.getAddress();
      this.frame = null;
   }

   public Simulator (boolean visible) { this(visible,About.DEFAULT_PORT); }

   public Simulator (boolean visible, int port)
   {
      try
      {
         this.server = new Simulator.Server(port);
      }
      catch (java.io.IOException e) {}

      if (server != null)
      {
         this.address = ""+server.getAddress();

         if (visible) {
            this.frame = new SimulatorFrame(""+server.getAddress(),this);
            frame.updateChannels(channels.toArray());
            frame.updateControls(controls.toArray());
            frame.updateLabel(0,0,0);
            frame.duration = constants.getConstant(Constants.GAME_DURATION);
         } else
            frame = null;
/* Problem hier: 'headless' (visible == false) klappt dann nicht...
         this.frame = new SimulatorFrame(""+server.getAddress(),this);  // always frame != null (damit er den Simulator bei Application.exit() mit schließt)
         frame.updateChannels(channels.toArray());
         frame.updateControls(controls.toArray());
         frame.updateLabel(0,0,0);
         frame.duration = constants.getConstant(Constants.GAME_DURATION);
         frame.setVisible(visible);
*/
      }
      else
      {
         this.address = "";
         this.frame = null;
      }
   }

   private boolean closed = false;
   public void close () // war synchronized (Festhäng-Error beim Schließen)
   {
      synchronized(CLOSE_SETSTATE) {
         if (closed) return;
         if (state == 2) setState(1);
         if (state == 1) setState(0);
         closed = true;
      }

      synchronized(this) { // ab hier: harmlos.
         if (!server.multiple) server.close();

         Control[] controls = this.controls.toArray();
         for (int i=0; i<controls.length; ++i) {
            controls[i].send(new SC.Bye());
            controls[i].close();
         }

         Channel[] channels = this.channels.toArray();
         for (int i=0; i<channels.length; ++i) {
            channels[i].send(new SU.Bye());
            channels[i].close();
         }

         if (frame != null) frame.removeAll();

         // close all active Threads

         channelGuard.stop();
      }
   }

   public void locate (int x, int y) { if (frame != null) frame.setLocation(x,y); }


   // ---- Simulator Properties (accessed by Server) --------------------------

   public int getNumberOfChannels () { return channels.toArray().length; }
   public int getNumberOfControls () { return controls.toArray().length; }
   public int getState () { return state; }
   public double getTime() { return (view == null)? 0 : view.time; }
   public double getDuration() { return duration; }
   public double getHertz() { return hertz; }


   // ---- Simulator Constants ------------------------------------------------

   public synchronized void set (String key, String value)
   {
      if (value != null) {
         value = constants.put(key,value);
         controls.send(new SC.Set(key,value));
      } else { // value == null
         value = constants.remove(key);
         controls.send(new SC.Set(key,value));
      }
   }

   public synchronized String get (String key)
   {
      return (String)constants.get(key);
   }

   public synchronized double option (String key)
   {
      int i = constants.indexOf(key);
      if (i < 0) return 0;

      return constants.getConstant(i);
   }


   // ---- Simulation State Automaton -----------------------------------------

   public synchronized void state (int state) { setState(state); }
   public synchronized int state () { return state; }

   private synchronized void setState (int state)  // synchronized by state() [CS]
   {
      if (closed) return;
      if (this.state == 0 && state == 1) checkin();
      else if (this.state == 1 && state == 0) checkout();
      else if (this.state == 1 && state == 2) start();
      else if (this.state == 2 && state == 1) stop();
      else return;
   }


   // ---- Simulator Actions --------------------------------------------------

   private double hertz ()  // arithmetisches Mittel der step-Zeiten der letzten Sekunde, dann 1/.
   {
      int n = history.size()-1;  // n >= 1 !
      double diff = ((Double)history.lastElement()).doubleValue() -  // => diff > 0 !
                    ((Double)history.firstElement()).doubleValue();
      return n/diff;
   }

   private void checkin ()
   {
      constants1 = (Constants)constants.clone();

      frequency = constants1.getConstant(Constants.SIMULATOR_FREQUENCY);
      timeratio = constants1.getConstant(Constants.SIMULATOR_TIME);
      duration  = constants1.getConstant(Constants.GAME_DURATION);

      timeouts.clear();  // clear timeout-table

      if (frame != null) frame.duration = duration;

      Channel[] array = channels.toArray();
      for (int i=0; i<array.length; ++i) array[i].active = true;

      terrain = new Terrain(constants1,array);  // setup simulation environment ...

      Runnable task = new Runnable() { public void run () { simulate(); } };
      timer = new Timer(task,frequency);
      history.removeAllElements(); history.addElement(new Double(timer.get()));
      hertz = 0;

      Hashtable hash = new Hashtable();         // setup team-scores ...
      for (int i=0; i<array.length; ++i)
         hash.put(""+array[i].t,new Team(array[i].t,0));
      Team[] teams = new Team[hash.size()]; int j=0;
      for (Enumeration<Team> e = hash.elements(); e.hasMoreElements(); )
         teams[j++] = (Team)e.nextElement();

      view = new View(terrain.getTime(),terrain.getDiskAbsolutes(),teams);
      ++ version;

      state = 1; server.refresh(); // if (frame != null) frame.updateLabel(state,view.time,hertz);
      controls.send(new SC.State(state).next(new View(view.time,view.disks,view.teams)));

      String[] options = new String[Constants.IS_FOR_UNIT]; int k=0;  // setup options for Units
      for (int i=0; i<Constants.CONSTANTS; ++i) if (constants1.isForUnit(i))
         options[k++] = constants1.get(Constants.getKey(i));

      for (int i=0; i<array.length; ++i)
         array[i].send(new SU.Checkin(options));  // all Units receive (checkin...)
   }

   private void checkout ()
   {
      terrain = null;    // discard simulation environment
      timer = null;
      hertz = 0;

      view = null;

      state = 0; server.refresh(); // if (frame != null) frame.updateLabel(state,0,0);
      controls.send(new SC.State(state));

      Channel[] array = channels.toArray();
      for (int i=0; i<array.length; ++i) array[i].active = false;

      constants1 = null;
   }

   private final Object SIMULATION = new Object();
   private final Object CLOSE_SETSTATE = new Object();

   private boolean attention_stopped = false;

   private void start ()
   {
	   state = 2; server.refresh(); // if (frame != null) frame.updateLabel(state,view.time,hertz);
	   controls.send(new SC.State(state));

	   timer.start();
   }

   private void stop ()
   {
       	synchronized(SIMULATION) {
       		timer.stop();         // stop timer scheduling
       		terrain.interrupt();  // force kernel thread to terminate

       		attention_stopped = true;
       		if(simulate_still_notifies) // wait for end of simulation
       			try {
//System.out.println("stop: SIMULATION wait");
       				SIMULATION.wait(); 
//System.out.println("stop: SIMULATION wait cont.");
       			} catch(Exception e) {}
       		attention_stopped = false;
       	}

       	state = 1; server.refresh(); // if (frame != null) frame.updateLabel(state,view.time,hertz);
       	controls.send(new SC.State(state));
       	channels.send(new SU.Break(),true);  // all active units receive (break)
   }


   private boolean simulate_still_notifies = true;
   private void simulate ()  // ---- Simulation Loop --------------------------
   {
      simulate_still_notifies = true;

      // ---- start

      if (timer.first())
      {
         Channel[] array = channels.toArray();
         for (int i=0; i<array.length; ++i) array[i].looking = true;

//         state = 2; server.refresh(); // if (frame != null) frame.updateLabel(state,view.time,hertz);
//         controls.send(new SC.State(state));
      }

      // ---- simulate

      Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

      do {
         if (timer.last()) break;           // "after-stopped" timer schedule

//System.out.println("simu: SIMULATION updateLook");
         synchronized(SIMULATION) { if (attention_stopped) break;  // updateLook is synchronized!
            updateLook();                   // send (look...) to units
         }
//System.out.println("simu: SIMULATION updateLook cont.");

         Thread.yield();                    // yield

         if (terrain.getTime() == duration) break;  // Game Over?

         double real = timer.get();         // Hertz History...
         history.addElement(new Double(real));
         while (history.size() > 2 &&
                ((Double)history.firstElement()).doubleValue() < real - TIME_HISTORY)
            history.removeElementAt(0);

         hertz = hertz();                   // evaluate current Hertz

         double time = Math.min(terrain.getTime()+timeratio/hertz,duration);

         terrain.simulate(time);            // simulate!

//System.out.println("simu: SIMULATION updateView");
         synchronized(SIMULATION) { if (attention_stopped) break;  // updateView is synchronized!
            updateView();
         }
//System.out.println("simu: SIMULATION updateView cont.");

      } while(false);

      // ---- stop

      if (timer.last())
      {
         timer.put(terrain.getTime());  // adjust timer (in case of interruption)
         history.removeAllElements(); history.addElement(new Double(timer.get()));

//         state = 1; server.refresh(); // if (frame != null) frame.updateLabel(state,view.time,hertz);
//         controls.send(new SC.State(state));
//         channels.send(new SU.Break(),true);  // all active units receive (break)

         Channel[] array = channels.toArray();
         for (int i=0; i<array.length; ++i) array[i].looking = false;

//System.out.println("simu: SIMULATION still notifies");
         synchronized(SIMULATION) {
            simulate_still_notifies = false;
            SIMULATION.notifyAll(); // wake up stop()-Thread
         }
//System.out.println("simu: SIMULATION still notifies cont.");
      }
   }


   // ---- Update Look! -------------------------------------------------------

   private synchronized void updateLook ()
   {
      Channel[] actives = this.channels.toArray(true);
      for (int i=0; i<actives.length; ++i) updateLook(actives[i]);
   }

   private synchronized void updateLook (Channel channel)
   {
   	if (!channel.looking || channel.version == version) return;
      Unite self = (Unite)terrain.getDisque(Disque.UNITE,channel.t,channel.n);
      SU.Look look = new SU.Look(self.E,self.py,terrain.getDiskRelatives(self.t,self.n));
      channel.send(look);

      channel.looking = false;
      channel.version = version;
   }


   private synchronized void updateView ()
   {
      if (state == 0) return;

      view.time = terrain.getTime();
      view.disks = terrain.getDiskAbsolutes();

      Event[] events = terrain.popEvents();
      for (int i=0; i<events.length; ++i) if (events[i].what == Event.COLLISION) {  // Goals
         DiskAbsolute a = view.disks[events[i].i];
         DiskAbsolute b = view.disks[events[i].j];

         DiskAbsolute goal = null;
         // Collision shotball-teamball => goal = teamball
         if      (a.what == a.BALL && a.t == 0 && b.what == b.BALL && b.t != 0) goal = b;
         else if (b.what == b.BALL && b.t == 0 && a.what == a.BALL && a.t != 0) goal = a;
         // Collision teamball-unit     => goal = unit
         else if (a.what == a.BALL && a.t != 0 && b.what == b.UNIT) goal = b;
         else if (b.what == b.BALL && b.t != 0 && a.what == a.UNIT) goal = a;

         if (goal != null) {  // Goooooooaaaaaaaal !!!!
            Double last = (Double)timeouts.get(""+goal.what+" "+goal.t+" "+goal.n); // Timeout?
            if (last == null || view.time - last.doubleValue() >= terrain.SCORE_TIMEOUT)
               for (int j=0; j<view.teams.length; ++j)
                  if (view.teams[j].t == goal.t) {
                     if (goal.what == goal.BALL) ++ view.teams[j].score;  // ball/ball => +1
                     // else                     -- view.teams[j].score;  // ball/unit => -1
                     timeouts.put(""+goal.what+" "+goal.t+" "+goal.n,new Double(view.time)); // Lock Disk!
                     break;
                  }
         }
      }
//      view.events = events;

      ++ version;

      // server.refresh(); // if (frame != null) frame.updateLabel(state,view.time,hertz);
      if (frame != null) server.refresh(); // permanently only if SimulatorFrame!

      controls.send(new View(view.time,view.disks,view.teams));
   }


   // ---- Network Protocol: Simulator's Server ------- [Simulator.Server] ----

   private class Server extends hoverball.net.Server
   {
      public Server (int port) throws IOException { super(port); multiple = false; }

      public void connection (Address sender, String message, Socket socket)
      {
         // ---- (connect #hash "My Team" Pur 23456 2.3) ------------ [US] ----
         Serial serial = US.deserialize(message);
         if (serial instanceof US.Connect) connect((US.Connect)serial,sender,socket);

         // ---- (connect #hash MyController) ----------------------- [CS] ----

         serial = CS.deserialize(message);
         if (serial instanceof CS.Connect) connect((CS.Connect)serial,sender,socket);
      }

      public void refresh () {
         if (frame != null) frame.updateChannels(channels.toArray());
         if (frame != null) frame.updateControls(controls.toArray());
         if (frame != null) frame.updateLabel(getState(),getTime(),getHertz());
      }
   }


   // ---- Network Protocol from Unit Client (invoked by Server) ---- [US] ----

   // ---- (connect #hash "My Team" MyPur 123456 2.3)

   synchronized void connect (US.Connect connect, Address sender, Socket socket)
   {
      // create channel
      Channel channel = channels.create(sender,connect.team,connect.name,connect.color,socket);
      if (channel == null) return;  // maximum teams or units
      channelGuard.ping(channel);

      // reply to sender

      channel.send(new SU.Connect(channel.t,channel.n,
         server.multiple? connect.hash : "",
         channel.team,channel.name,channel.color));

      controls.send(new hoverball.serial.Channel(channel.t,channel.n,
         channel.team,channel.name,channel.color,channel.host));

      server.refresh(); // if (frame != null) frame.updateChannels(channels.toArray());
   }

   // ---- Network Protocol from Unit Client (invoked by Channel)

   // ---- (look)

   synchronized void look (Channel channel)
   {
      if (state == 2 && channel.active) { channel.looking = true; updateLook(channel); }
   }

   // ---- (action * * 0.7 "Let's go!")

   synchronized void action (Channel channel, US.Action action)  // action.next ignored.
   {
      if (terrain == null || !channel.active) return;

      terrain.addAction(new Action(channel.t,channel.n,
         action.polarize,action.left,action.right,action.message));
   }

   // ---- (ping)

   synchronized void ping (Channel channel)
   {
      channelGuard.ping(channel);
   }

   // ---- (bye)

   synchronized void bye (Channel channel)
   {
      action(channel,new US.Action(0,0,0,""));  // send null-action into kernel

      channelGuard.cancel(channel);
      channel.close();
      channels.remove(channel);

      server.refresh(); // if (frame != null) frame.updateChannels(channels.toArray());

      controls.send(new hoverball.serial.Channel(channel.t,channel.n));
   }



   // ---- Network Protocol from Control Client (invoked by Server) - [CS] ----

   // ---- (connect #hash MyController)

   synchronized void connect (CS.Connect connect, Address sender, Socket socket)
   {
      // create control
      Control control = controls.create(sender,socket);
      channelGuard.ping(control);

      // reply: (connect #) [(active channels)* (state 1) (view ..) [(state 2)]] (inactive channels)*

      Serial returns = new SC.Connect(
            server.multiple? connect.hash : "",
            connect.name);
      Serial last = returns;

      Constants constants_ = (state == 0)? constants : constants1;
      for (int i=0; i<Constants.CONSTANTS; ++i) {
         String key = Constants.getKey(i);
         String value = (String)constants_.get(key);  // != null
         last = last.next = new SC.Set(key,value);
      }

      if (state != 0) {
         Channel[] actives = channels.toArray(true);
         for (int i=0; i<actives.length; ++i) {
            Channel ch = actives[i];
            last = last.next = new hoverball.serial.Channel(ch.t,ch.n,ch.team,ch.name,ch.color,ch.host);
         }
         last = last.next = new SC.State(1);
         last = last.next = new View(view.time,view.disks,view.teams);  // clone ".next"!!
         if (state == 2) last = last.next = new SC.State(2);

         for (int i=0; i<Constants.CONSTANTS; ++i) {  // constants1 != null
            String key = Constants.getKey(i);
            String value1 = (String)constants1.get(key);
            String value  = (String)constants.get(key);  // != null
            if ((value1 == null && value == null) ||
                (value1 != null && value1.equals(value))) ; // ignore!
            else {                                          // send!
               last = last.next = new SC.Set(key,value);
            }
         }
      }

      Channel[] inactives = channels.toArray(false);
      for (int i=0; i<inactives.length; ++i) {
         Channel ch = inactives[i];
         last = last.next = new hoverball.serial.Channel(ch.t,ch.n,ch.team,ch.name,ch.color,ch.host);
      }
      control.send(returns);

      server.refresh(); // if (frame != null) frame.updateControls(controls.toArray());
   }

   // ---- Network Protocol from Control Client (invoked by Control)

   // ---- (set parameter_key value)

   synchronized void set (CS.Set set)               // all serial.next's will be ignored.
   {
      set(set.key,set.value);  // setting constants is public!
   }

   // ---- (channel 2 3)

   synchronized void channel (CS.Channel channel)
   {
      Channel channel_ = channels.get(channel.t,channel.n);
      if (channel_ != null) {
         channel_.send(new SU.Bye());
         bye(channel_);
      }
   }

   // ---- (state 1)

   synchronized void state (CS.State state)
   {
      setState(state.state);
   }

   // ---- (view * (unit t n 3 4 5...))  => View!

   synchronized void view (Control control, View view)
   {
      if (state != 1) return;

      boolean modified = false;

      if (!Double.isNaN(view.time)) {          // set time
         timer.put(view.time);
         terrain.setTime(timer.get());
         history.removeAllElements(); history.addElement(new Double(timer.get()));
         modified = true;
      }

      if (view.disks.length > 0) {             // set disks
         terrain.setDiskAbsolutes(view.disks);
         modified = true;
      }

      if (view.teams.length > 0) {             // set score
         for (int i=0; i<view.teams.length; ++i)
            for (int j=0; j<this.view.teams.length; ++j)
               if (this.view.teams[j].t == view.teams[i].t)
                  this.view.teams[j].score = view.teams[i].score;
         modified = true;
      }

      if (modified) updateView();
   }

   // ---- (ping)

   synchronized void ping (Control control)
   {
      channelGuard.ping(control);
   }

   // ---- (bye)

   synchronized void bye (Control control)
   {
      channelGuard.cancel(control);
      control.close();
      controls.remove(control);

      server.refresh(); // if (frame != null) frame.updateControls(controls.toArray());
   }

   private class ChannelGuard extends java.util.TimerTask
   {
       private final int LIFETIME = (int)(Socket.PING_TIMEOUT/Socket.PING_EVERY);
       private Hashtable table = new Hashtable();
       private java.util.Timer timer;
       private ChannelGuard()
       {
           timer = new java.util.Timer("Hoverball Ping !! (Simulator)");
           timer.scheduleAtFixedRate(this,0,Socket.PING_EVERY);
       }

       public synchronized void run()
       {
           Enumeration<Object> k;
//            System.out.println("client table:");
//            k = table.keys();
//            while (k.hasMoreElements()) {
//                Object key = (Object) k.nextElement();
//                System.out.println("Key " + key + "; Value " + (Integer) table.get(key));
//            }
           k = table.keys();
           while (k.hasMoreElements()) {
               Object client = (Object) k.nextElement();
               int n = ((Integer)table.get(client)).intValue();
               if (n==0) { // kick out!
                  if(client instanceof Channel) {
                     Channel channel = (Channel)client;
                     channel.send(new SU.Bye());
                     bye(channel);
                  } else if(client instanceof Control) {
                     Control control = (Control)client;
                     control.send(new SC.Bye());
                     bye(control);
                  }
               }
               else table.put(client,new Integer(n-1));
           }
       }

       synchronized void ping(Object client)
       {
          table.put(client,new Integer(LIFETIME));
       }

       synchronized void cancel(Object client)
       {
          table.remove(client);
       }

       synchronized void stop()
       {
          timer.cancel();
          timer.purge();
       }
   }

   // ----

   public double time () {
	  if (terrain == null) return 0;
	  else return terrain.getTime();
   }

   public boolean complete () {
	  if (terrain == null) return false;
	  else return terrain.getTime() == constants.getConstant(Constants.GAME_DURATION);
   }

}
