package hoverball.team.LasMaquinaTeam_;

import hoverball.Puck;
import hoverball.Unit;
import hoverball.debug.Text;
import hoverball.math.Complex;
import hoverball.math.Matrix;
import hoverball.math.Sphere;
import hoverball.math.Vector;

/**
 * Notre agent de l'équipe Las Maquina
 * Utilisation de l'architecture B.D.I.
 * @author Cyrille COIN, Florian COLIN, Thibaud COMTE
 *
 */
public class Agent extends Unit {
	
	/** 
	 * Le nom de l'agent
	 */
	public final String name;
	
	/** 
	 * L'id de l'agent 
	 */
	public final int identity;
	
	/**
	 * Les croyances de l'agent
	 */
	public Croyance croyance;
	
	/**
	 * Le désir de l'agent
	 */
	public Desir desir;
	
	/**
	 * L'intention de l'agent
	 */
	public Intention intention;
	
	/**
	 * L'état de l'agent par rapport à la balle
	 */
	public State state;
	
	/**
	 * Utilitaire d'action privé associé au joueur
	 */
	private Action action = null;
	
	/**
	 * Cible à atteindre
	 */
	public Complex target = null;	
	
	/** 
	 * Energie de polarisation : si négatif, attire la balle, si positif l'éloigne 
	 */
	public double P = 0;
	
	/** 
	 * L'énergie du moteur gauche
	 */
	public double L = 0;
	
	/** 
	 * L'énergie du moteur droit 
	 */
	public double R = 0;
	
	/** 
	 * Message passé pour l'équipe (ne doit pas dépasser 100 caractères 
	 */
	public String M = " ";
	
	/**
	 * Ordre reçu par l'un de mes co-équipiers
	 */
	public Message order = null;
	
	/**
	 * Message à envoyer à ses co-équipiers
	 */
	public Message msgToSend = null;
	

	/**
	 * Construit un joueur de l'équipe ISICO
	 * @param team Nom de l'équipe
	 * @param name Nom du joueur
	 * @param color Couleur du puck
	 */
	public Agent(String team, String name, int identity, int color) {
		super(team,name,color);
		this.name = name;
		this.identity = identity;
		this.croyance = new Croyance(this);
		this.action = new Action(this);
	}

	/**
	 * loop structure:  while (look()) { ...; action(...); }
	 */
	public void loop() {

		// Initialisation du puck
		initialize();

		/**
		 * Architecture B.D.I. (Croyance, Désir, Intention)
		 * 
		 * Les Croyances d'un agent sont les informations que l'agent possède
		 * sur l'environnement et sur d'autres agents qui existent dans le même
		 * environnement.
		 * 
		 * Les Désirs d'un agent représentent les états de l'environnement, et 
		 * parfois de lui-même, que l'agent aimerait voir réalisés.
		 * 
		 * Les Intentions d'un agent sont les désirs que l'agent a décidé d'accomplir
		 * ou les actions qu'il a décidé de faire pour accomplir ses désirs.
		 */
		while (look())
		{
			/**
			 * Obtenir des nouvelles perceptions p
			 * ex: que voit-on, qu'entendons nous, lecture des msg des co-équipiers.. 
			 */
			getWorldInfo();
			
			/**
			 * Révision des croyances suivant les nouvelles perceptions obtenues
			 * 				B <-- revBelief(B, p)
			 * 
			 * ex: je connais les nouvelles positions de mes camarades, je 
			 * peux donc connaitre qui est le plus proche de la balle ou de la cage
			 */
			reviseBelief();
			
			/** 
			 * Represente le processus de decision de l'agent prenant en 
			 * compte ses desirs et ses intentions courantes
			 * 				I <-- options(D,I)
			 * 
			 * Ex: Désir : ne pas prendre de buts
			 * Intention : prendre la place de gardien
			 * Je me place à la bonne position
			 */
			options();
			
			/**
			 * Change les desirs d'un agent si ses croyances ou intentions changent, 
			 * pour maintenir la consistance des desirs de l'agent
			 * 				D <-- des(B,D,I)
			 * 
			 * ex : alors que j'allais me placer entre les cages et la balle pour 
			 * défendre, je perçois que je suis le plus près de la balle !
			 * Mon désir devient maintenant de récupérer la balle
			 */
			des();

			/**
			 * Décide des intentions a poursuivre
			 * 				I <-- filtre(B,D,I)
			 * 
			 * ex : mon intention est maintenant d'aller vers le ballon
			 * pour le récupérer
			 */
			filtre();

			/**
			 * Planification
			 * 				PE <-- plan(B,I)
			 * 
			 * ex: Suivant ce que je sais sur l'environnement, position des points,
			 * je mets à jour P, L, R, et M. 
			 */
			plan();

			/**
			 * Execution de l'action
			 * 				Exécuter(PE)
			 * 
			 * ex: action(P,L,R,M) // envoye de l'action au serveur
			 */
			execute();
		}
	}

	/**
	 *  Initialise l'environnement pour le puck
	 */
	public void initialize() 
	{
		croyance.sphere_rad = option("world.radius");
		croyance.sphere = new Sphere(croyance.sphere_rad);
		croyance.charge_min = option("unit.charge.min");
		croyance.charge_max = option("unit.charge.max");
		croyance.engine_min = option("unit.engine.min");
		croyance.engine_max = option("unit.engine.max");
		croyance.radius = option("unit.radius");
		croyance.vision = option("unit.vision");
		croyance.distanceBallUnit = croyance.radius + option("ball.radius");

		this.desir = Desir.NONE;
		this.intention = Intention.NONE;
		this.state = State.HAVE_NO_BALL;
		for (int i=0; i<croyance.nbFriends; i++) this.croyance.friendDesir[i] = Desir.NONE;
	}

	/**
	 * Obtention des nouvelles perceptions
	 */
	private void getWorldInfo() {
		
		// Analyse les informations retournées par l'environnement
		// et met à jour les perceptions du joueur sur l'ensemble 
		// des pucks, i.e. la balle, le but à atteindre, le but des adversaires
		// la position des amis et adversaires...
		detect(pucks);
	}

	/**
	 * Fonction de révision des croyances de l'agent 
	 * lorsqu'il recoit de nouvelles perceptions sur 
	 * l'environnement
	 */
	private void reviseBelief() 
	{
		// Initialisation de certains paramètres
		order = null;
		msgToSend = null;
		M = "";		// Msg envoyé aux co-équipiers
		
		// On révise les croyances
		croyance.reviseBelief();
		
		// Lecture des messages envoyés par les autres pucks de l'équipe
		order = croyance.readMsgs();
	}

	/**
	 * Fonction qui represente le processus de decision 
	 * de l'agent prenant en compte ses desirs et 
	 * ses intentions courantes
	 */
	private void options() 
	{
		// Inutile ici
	}

	/**
	 *  Fonction qui peut changer les desirs d'un agent 
	 *  si ses croyances ou intentions changent, 
	 *  pour maintenir la consistance des desirs de l'agent
	 */
	private void des() 
	{
		/**
		 * Si un ordre nous a été envoyé par un co-équipier, on l'exécute
		 */
		if (order != null) {
			this.desir = Desir.EXECUTE_ORDER;
			return;
		}
		
		/**
		 * Tant qu'on n'a pas détecté tous les pucks, on tourne pour les trouver
		 */
		if (!croyance.pucksLocated) {
			this.desir = Desir.NONE;
			return;
		}
		
		/**
		 * Prise en compte des évènements prioritaires
		 * ex : évitement des cages
		 */
		// Si on est dangereusement proche de nos cages
		if (croyance.nearFlop) {
			this.desir = Desir.EVITE_FLOP;
			return;
		}
		// Si on est dangereusement proche des cages adverses
		if (croyance.nearGoal) {
			this.desir = Desir.EVITE_GOAL;
			return;
		}
		
		/**
		 * Stratégies de défense
		 * Elle s'applique si l'équipe adverse a la balle
		 */
		if (croyance.opponentHaveBall /*&& !croyance.friendHaveBall*/)
		{
			// Si je suis le plus proche de la balle
			if (croyance.friendBallClosest == this.identity) {
				this.desir = Desir.DEF_RECUPERATEUR;
			}

			// si je suis le plus proche de la cage
			else if (croyance.friendFlopClosest == this.identity) {
				this.desir = Desir.DEF_GOAL;
			}
			
			// si je suis dans aucune des deux situations précedentes
			else if (croyance.friendGoalClosest == this.identity) {
				this.desir = Desir.DEF_MARQUAGE;
			}
		}
		
		/**
		 * Startégies d'attaque
		 * Elle s'applique si l'équipe adverse ne l'a pas. 
		 * Si l'équipe adverse a la balle, on se replace dans une stratégie de défense
		 */
		else // if (!croyance.opponentHaveBall)
		{
			// Si je suis le plus proche de la balle
			if (croyance.friendBallClosest == this.identity) {
				this.desir = Desir.ATT_ATTAQUANT;
			}

			// si je suis le plus proche de la cage
			else if (croyance.friendFlopClosest == this.identity) {
				this.desir = Desir.DEF_GOAL;
			}
			
			// si je suis dans aucune des deux situations
			else if (croyance.friendGoalClosest == this.identity) {
				this.desir = Desir.ATT_ASSISTANT;
			}			
		}
	}

	/**
	 * Fonction la plus importante car elle decide 
	 * des intentions a poursuivre
	 */
	private void filtre()
	{
		// Si aucuns évènements prioritaires
		// Alors on décide de l'intention
		switch (desir) 
		{		
		case DEFENSEUR :  break;
			// Si on est le plus près de la balle
		case ATT_ATTAQUANT :
			switch (state)
			{
				// On n'a pas la balle, alors on va la chercher
			case HAVE_NO_BALL :
				P = 0;
				this.target = croyance.sphere.warp(croyance.ball.X.c);
				this.intention = Intention.GOTO;
				break;
				// On est près de la balle, alors on la récupère
			case NEAR_BALL :
				this.intention = Intention.GET_BALL;
				break;
				// On a la balle
			case HAVE_BALL :
				if (!croyance.canIShoot())
				{	// Il y a un obstacle, je tente de faire une passe
					this.target = croyance.sphere.warp(croyance.friend[croyance.friendGoalClosest].X.c);
					this.intention = Intention.PASS;
				}
				else 
				{	// Sinon j'essaie de tirer
					if (croyance.distanceTo(croyance.goal) < 40.0) {
						this.target = croyance.sphere.warp(croyance.goal.X.c);
						this.intention = Intention.SHOOT;
					} else {
						this.target = croyance.sphere.warp(croyance.goal.X.c);
						this.intention = Intention.GOTO;
					}
				}
				break;
			}
			break;

			// L'assistant assiste l'attaquant et se place pour fournir une solution pour marquer
		case ATT_ASSISTANT :
			// On choisit de placer l'assistant à la perpendiculaire de la droite 
			// passant par l'attaquant et le but.
			this.target = croyance.getPerpendicularPoint(croyance.friend[croyance.friendBallClosest], croyance.goal, 30.0);
			this.intention = Intention.GOTO;
			break;

			// Défenseur le plus proche de la balle
			// Son role est de se diriger vers la balle pour géner l'adversaire
		case DEF_RECUPERATEUR :
			
			switch(state)
			{
				// Si on est près de la balle, on dégage
			case NEAR_BALL :
				P = croyance.charge_max;
				break;
				// Sinon on va vers le ballon
			default :
				P = 0;
				this.target = croyance.sphere.warp(croyance.ball.X.c);
				this.intention = Intention.GOTO;
				break;
			}
			break;
			
			// Défenseur qui n'est ni gardien ni le plus proche de la balle
			// Son rôle est de perdurber l'adversaire le 2eme plus proche de la balle
		case DEF_MARQUAGE :
			this.target = croyance.sphere.warp(croyance.opponent[croyance.opponentFlopClosest].X.c);
			this.intention = Intention.GOTO;
			break;
			
			// Etre le gardien. On se place entre les cages à défendre et la balle.
		case DEF_GOAL :
			// Si on est pas statique (causé par une pénalité par exemple)
			// Alors on fait le gardien
			
			if (croyance.isStatic && croyance.distanceTo(croyance.flop) < 5.00) 
			{	// sinon qqun d'autre doit prendre notre place rapidement
				// On envoie donc un message au bon agent pour nous remplacer
//				System.out.println(name + "(" + identity + ")" + " static et envoye une requete to " + croyance.friendGoalClosest);
				msgToSend = new Message(identity, croyance.friendGoalClosest, Intention.DO_THE_GOAL, null);
				P = 0;
			}
			else 
			{
				P = 0;
				this.intention = Intention.DO_THE_GOAL;
			} 
			break;
			
			// Stratégie d'évitement de la cage
		case EVITE_FLOP :
			if (croyance.isStatic && croyance.distanceTo(croyance.flop) < 5.00) 
			{	// sinon qqun d'autre doit prendre notre place rapidement
				// On envoie donc un message au bon agent pour nous remplacer
//				System.out.println(name + "(" + identity + ")" + " static et envoye une requete to " + croyance.friendGoalClosest);
				msgToSend = new Message(identity, croyance.friendGoalClosest, Intention.DO_THE_GOAL, null);
				P = 0;
			}
			else {
				// On choisit de se placer à la perpendiculaire de la droite passant par l'agent et les cages
				this.target = croyance.getPerpendicularPoint(this.self, croyance.flop, 10.0);
				this.intention = Intention.GOTO;
			}

			// Stratégie d'évitement de la cage
		case EVITE_GOAL :
			// On choisit de se placer à la perpendiculaire de la droite passant par l'agent et les cages
			this.target = croyance.getPerpendicularPoint(this.self, croyance.goal, 10.0);
			this.intention = Intention.GOTO;
			
			// Aucun désir
		case NONE :
			this.intention = Intention.TURN;
			break;
			
			// On nous a envoyé un ordre
		case EXECUTE_ORDER :
			//TODO
//			System.out.println(name + "(" + identity + ") éxécute un ordre");
			this.target = order.getTarget();
			this.intention = order.getIntention();
		}
	}

	/**
	 * La fonction qui transforme les plans partiels en plans exécutables.
	 * Un plan est une séquence d'actions à exécuter dans le temps.
	 */
	private void plan() {

		// Action qui mette à jour P,L,R et M, c'est-à-dire
		// la polorisation, le moteur gauche et droit ainsi que le message
		// envoyé aux co-équipers

		switch (intention)
		{
		case TURN :
			action.turn(1.0);
			break;
			
		case GOTO :
			action.moveToPos(target);
			break;
			
		case GET_BALL :
			action.getBall();
			break;
			
		case SHOOT :
			action.shoot(croyance.sphere.warp(target));
			break;
			
		case PASS :
			L = 0.00;
			R = 0.00;
			action.shoot(croyance.sphere.warp(target));
			break;
			
		case DO_THE_GOAL :
			action.doTheGoal();
			break;
			
		default:
			break;
		}
		
		// Si on a un message à envoyer, on le fait
		if (msgToSend != null) {
			M = msgToSend.toString();
		}

	}

	/**
	 * Exécution du plan
	 */
	private void execute() 
	{
		String msg_debug = name + "[" + state + "][" + desir + "][" + intention + "]";
		debug (new Text(new Vector(0,0,0), msg_debug, -2));
		
		// On envoie l'action au serveur Hoverball
		action(P, L, R, M);
	}




	// Fonctions utilitaires pour B.D.I. ==============================================

	/** 
	 * analyzes the pucks[] array and sorts all information into the variables
	 * ball, goal, flop, friend[], opponent[], node[]
	 * @author sb
	 * @param pucks
	 */
	private void detect(Puck[] pucks) 
	{		
		// update known information thanks to a visible node
		Matrix m = null;
		for (int i = 0; i < 6; ++i)
			if ((m == null) && (croyance.node_[i] != null)) if (see(puck(NODE,0,i+1))) {
				m = Matrix.inv(croyance.node_[i].X).mul(puck(NODE,0,i+1).X);
				break;
			}      

		if (m != null) // should be the case for any unit.vision > 0.66
		{
			if (see(croyance.ball)) croyance.ball.X = croyance.ball.X.mul(m); /*immutable*/
			if (see(croyance.goal)) croyance.goal.X = croyance.goal.X.mul(m); /*immutable*/
			if (see(croyance.flop)) croyance.flop.X = croyance.flop.X.mul(m); /*immutable*/
			for (int k=0; k<croyance.nbFriends; ++k) if (see(croyance.friend[k])) croyance.friend[k].X = croyance.friend[k].X.mul(m); /*immutable*/
			for (int k=0; k<croyance.nbOpponents; ++k) if (see(croyance.opponent[k])) croyance.opponent[k].X = croyance.opponent[k].X.mul(m); /*immutable*/
			for (int k=0; k<6; ++k) if (see(croyance.node[k])) croyance.node[k].X = croyance.node[k].X.mul(m); /*immutable*/
		}

		// set old nodes to null
		for (int i = 0; i < 6; i++) croyance.node_[i] = null;

		// overwrite with new data  
		for (int i = 0; i < pucks.length; ++i) {
			if (pucks[i].X != null) // I have to "see" them, not only "hear" them
			{
				if (pucks[i].equals(BALL))
				{
					// it's a ball
					if (pucks[i].t == 0) { croyance.ball = new Puck(pucks[i]); }
					// it's a goal
					else if (pucks[i].t == self.t) { croyance.goal = new Puck(pucks[i]); }
					// it's a flop
					else { croyance.flop = new Puck(pucks[i]); }
				}
				else if (pucks[i].equals(UNIT))
				{ 
					// it's a unit...
					if (pucks[i].t == self.t)
					{ 
						// On récupère l'id du puck
						int idPuck = Integer.parseInt(pucks[i].id.substring(pucks[i].id.lastIndexOf(' ') + 1));
						
						// ... of my own team
						croyance.friend[idPuck-1] = new Puck(pucks[i]);
						

						// TODO Lecture des msgs ici
					}
					else
					{ 
						// On récupère l'id du puck
						int idPuck = Integer.parseInt(pucks[i].id.substring(pucks[i].id.lastIndexOf(' ') + 1));
						
						// ... of the opponent team
						croyance.opponent[idPuck-1] = new Puck(pucks[i]);

					}
				}
				else if (pucks[i].equals(NODE)) // it's a node
				{
					croyance.node[pucks[i].n-1] = new Puck(pucks[i]);
					croyance.node_[pucks[i].n-1] = new Puck(pucks[i]);
				}
			}
		}     
		// perhaps all pucks are located now?
		if (!croyance.pucksLocated) croyance.pucksLocated = croyance.pucksLocated();
	}

	

	/**
	 * Est-ce que je perçois le Puck ?
	 * @author sb
	 * @param p
	 * @return
	 */
	private boolean see(Puck p)
	{
		if (p==null) return false; // I don't perceive it
		if (p.X==null) return false; // I hear but don't see it
		return true;
	}

	/**
	 * Retourne la position actuelle de l'agent
	 * @return la position actuelle de l'agent
	 */
	public Vector getPosition()	{
		return this.self.X.c;
	}
	

}
